/***************************************************************************
*    Copyright 2000  Broadcom Corporation
*    All Rights Reserved
*    No portions of this material may be reproduced in any form without the
*    written permission of:
*             Broadcom Corporation
*             16251 Laguna Canyon Road
*             Irvine, California  92618
*    All information contained in this document is Broadcom Corporation
*    company private, proprietary, and trade secret.
*
****************************************************************************
*
*    Filename: provlib.c
*    Creation Date: 24 Aug 1999 (v0.00)
*    VSS Info:
*        $Revision: 1.2 $
*        $Date: 2004/10/04 19:51:17 $
*
****************************************************************************
*    Description:
*
*      PROVLIB Module: MTA Provisioining Library Module.
*      The provosioning module contains all neccessary functionality in
*      accordance with the "PKT-SP-PROV-I01-991201" PacketCable specification
*
****************************************************************************/

#include <globalCfg.h>     /* MTA configuration options */
#include <provCfg.h>
#include <appos.h>
#include <prepc.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
//#include <pna_mib.h>

#include <hausenv.h>    /* HH Debug definitions */
//#include <pktcMib.h>    /* PKTC MIB Related Defs&Prototypes */
#include <net.h>        /* IPADDR data type */
#include <provlib.h>    /* Provisioning Task definitions */
#include <cmmtaint.h>
#include <snmputils.h>
#include <krb5client.h> /* To work with the Krb Client Library */
#include <krbtasks.h>   /* task to handle wakeup/rekey and timeouts */
#include <MibAccessSec.h>
#include <SigSec.h>     /* Signalling Security Defs to Init Security Component */

#include "dhcputils.h"  /* DHCP client Wrapper functions definitions */
#include "SecLib.h"     /* To do decryption and to calculate hash of config file */


/*
   setCfgFnameEvt:
                  signals that ProvSever set the ConfigFile Name to the
                  "mibMtaDevConfigFileName"
*/
OS_EVENT    setCfgFnameEvt;

PRZPROVIS   *g_pProvisDescBlock = NULL;

OS_SEMID    prgsemaProvstart; /* Semaphore to wait until provisioning is OK */
OS_SEMID    prgsynSema;       /* General Synchronization Semaphore */

OS_TASKID   provTaskId;       /* Provisioning task id */

/*
** This global structre contains PacketCable related info retrieved by the CM's DHCP process.
*/
PacketCableServers   g_PktcServerInfo = NULL;

#define     TFTP_MAX_BUFLEN   0x4000 /* (16K) */
char        prgCfgBuffer[TFTP_MAX_BUFLEN];
UINT32      prgCfgBytesRead;
char        *hashBuffer, *keyBuffer;
UINT32      hashLength, keyLength;

#if 0 /* un-def this if you want to debug more */
#  define  SNMP_DIRECT_DEBUG
#endif

#define   PROV_WAIT_FOR_CFGNAME_TO  30 /* (sec) */

#define  PROV_PC_COMPLIANT    0

/*
** The following Random Generating Functions are prototyped from
** several sources:
**    1. Seed Randomizer approach is taken from "www.umoncton.ca/desliem/JavaScript/random.html"
**       To boost chancess of "randomness" I am using combination of EthAddress and current time with 10msec ticks
**    2. Random Number Generator is taken from  "www.helsbreth.org/random/rng_combo.html" and
**                                              "www.helsbreth.org/random/shuffle.html"
*/

/************************************************************************
**
** COMBO
**
**  a simple but very good combination generator.
**  It combines, by addition mod 2^32,
**          x(n)=x(n-1)*x(n-2) mod 2^32
**      and
**          y(n)=30903*y(n-1) + carry mod 2^16
**  The period of the first is 3*2^29, on odd integers, and the
**  period of the second, a multiply-with-carry generator, is
**  30903*2^15-1=1012629503, so the period of combo exceeds 2^60.
**  This generator is simple to program in Fortran or C and quite
**  fast.
**
*************************************************************************/

/*
**  Simple combo, period > 2^60.5
**  x(n)=x(n-1)*x(n-2) mod 2^32 added to
**  period of x(n)=x(n-1)*x(n-2) is 3*2^29 if both seeds are
**    odd, and one is +or-3 mod 8.
**  easy to ensure: replace seed x with 8*seed+3, and y with 2*seed+1
*/

static UINT32  combo_x = 3;
static UINT32 combo_y = 1;
static UINT32 combo_z = 1;
static UINT32 combo_v = 0;

void seed_rand_combo(UINT32 seed)
{
   combo_x = seed * 8 + 3;
   combo_y = seed * 2 + 1;
   combo_z = seed | 1;
   combo_v = 0;
}

UINT32 rand_combo(void)
{
   combo_v = combo_x * combo_y;
   combo_x = combo_y;
   combo_y = combo_v;
   combo_z = (combo_z & 65535) * 30903 + (combo_z >> 16);

   return combo_y + combo_z;
}


#define  RND_SEED_RANDOMAIZER    214783647

/*
*****************************************************************************
** FUNCTION:   mtaRndSeed
**
** PURPOSE:    Initializes the SEED number for Random Number Generator,
**
** PARAMETERS: none.
**
** RETURNS:    none
**
** NOTE:       To boost chancess of "randomness" I am using combination of
**             EthAddress and current time with 10msec ticks
*****************************************************************************
*/
void mtaRndSeed(void)
{
   unsigned char EthernetAddress[6];
   int      SeedVal;
   OS_TMDATE   date;
   OS_TMTIME   time;
   OS_TICKS    ticks;

   bcmOsGetTimeEx(&date, &time, &ticks);
   cmMiscGetMtaMacAddr (EthernetAddress);

   SeedVal = (((unsigned long) EthernetAddress[3]) << 24) |
             (((unsigned long) EthernetAddress[4]) << 16) |
             (((unsigned long) EthernetAddress[5]) << 8);

   SeedVal += date + time + ticks;
   SeedVal %= RND_SEED_RANDOMAIZER;

   seed_rand_combo(SeedVal);
}


unsigned long mtaRand(void)
{
   return rand_combo();
}




#define TYPE_TLV_MIB_OBJECT      11
#define TYPE_TLV_MIB_NEW_OBJECT  35
#define TYPE_TLV_FILE_MARK       254

#define TLV_FILE_START_VAL       1
#define TLV_FILE_END_VAL         255


/*
*****************************************************************************
** FUNCTION:   provParseConfigInfo
**
** PURPOSE:    This function parses a config info retrievd from Provis Server
**             as a binary file in SNMP TLV format..
**
** PARAMETERS: pGwConfig  - ptr to CONFIG structure which is being filled in.
**             pBuffer    - ptr to the Buffer with SNMP TLV provisi data.
**             bufferSize - size of the "pBuffer" in bytes.
**
** RETURNS:    0 if OK, 1 - otherwise;
**
** NOTE:
*****************************************************************************
*/
UINT32    provParseConfigInfo(char* pBuffer, UINT32 bufferSize)
{
   char    typeField, valField;
   int     lengthField, start = 0, tlvMarks = 0, tlvNum = 0;

   /* Keep processing TLVs until we run out of buffer space. */
   while (bufferSize > 1)
   {
      /* Pull the type field out of the buffer. */
      typeField = *pBuffer;
      pBuffer++;
      bufferSize--;

      /* Also, pull out the length field while we're here. */
      if (typeField == TYPE_TLV_MIB_NEW_OBJECT)
      {
         lengthField = (*pBuffer)<<8;
         lengthField |= (*(pBuffer+1)) & 0x00FF;

         pBuffer +=2;
         bufferSize -= 2;
      }
      else
      {
         lengthField = *pBuffer;
         pBuffer++;
         bufferSize--;
      }

      printf("provParseConfigInfo:  Parsed type=%d; length=%d\n", typeField, lengthField);

      if (typeField == TYPE_TLV_FILE_MARK)
      {
         tlvMarks++;
         /* Pull out the value field which discriminates between Start and End Marks. */
         valField = *pBuffer;
         pBuffer++;
         bufferSize--;

         if (start)
         {
            if (valField != TLV_FILE_END_VAL)
            {
               printf ("ERROR: Unknown Second (Ending) TLV FileMark Tag (%d).\n", valField);
               return 1;
            }
            else
            {
               break;   /* End of MTA Prov CfgFile processing */
            }
         }
         if (valField != TLV_FILE_START_VAL)
         {
            printf ("ERROR: Unknown First (Starting) TLV FileMark Tag (%d).\n", valField);
            return 1;
         }
         start = 1;
         continue;
      }

      /* Make sure we have enough bytes left in the buffer for the value. */
      if (lengthField > bufferSize)
      {
         printf("ERROR: provParseConfigInfo:  Not enough bytes left for the value!\n");
         return 1;
      }

      /* Here we have a data item which has a corresponding MIB OID object */
      if (typeField == TYPE_TLV_MIB_OBJECT || typeField == TYPE_TLV_MIB_NEW_OBJECT)
      {
         if (!start)
         {
            printf("ERROR: provParseConfigInfo():  CFG File - no START Marker.\n");
            return 1;
         }
         printf ("\nNext SNMP MIB object found\n");

         if (snmpProcessBerVar(pBuffer))
         {
            printf ("ERROR: provParseConfigInfo(): Cannot process the OID object.\n");
            return 1;
         }
      }
      else
      {
         printf ("ERROR: provParseConfigInfo(): Wrong TLV MIB Object Type (expected %d).\n", TYPE_TLV_MIB_OBJECT);
         return 1;
      }
      /* Skip the bytes that we used. */
      pBuffer += lengthField;
      bufferSize -= lengthField;
      tlvNum++;
   }

   /*
   ** If there is leftover data in the buffer, then we have to assume that
   ** there was a problem.
   */
   if (bufferSize > 0)
   {
      printf("ERROR: provParseConfigInfo() - Partial TLV left over in the buffer!  Other TLV may be invalid.\n");
      return 1;
   }

   if (tlvMarks != 2)
   {
      printf("ERROR: provParseConfigInfo():  CFG File - no END Marker.\n");
      return 1;
   }

   return 0;

} /* end of provParseConfigInfo() */

/* ================================================================================================  */


/*
*****************************************************************************
** FUNCTION:   setProvState()
**
** PURPOSE:    Sets the Boot state value of the "ServerBootState" MIB Object
**
** PARAMETERS: state - correponding state of the boot process
**
** RETURNS:    NONE
**
** NOTE:
**
*****************************************************************************
*/
static void setProvState(PRZPROVIS  *provBlockp, int state)
{
   if (provBlockp)
   {
      provBlockp->prxMibAccessFunc(MIBITEM(PRV, ProvState), NULL, NULL, NULL, (void*)&state, MIB_ITEM_SET);
   }

} /* end of setProvState() */

/*
*****************************************************************************
** FUNCTION:   makeUsmUserSecurityName
**
** PURPOSE:    Routine creates the USM UserSecurityName according to the
**             rules in ProvSpec-I02, 8.1.4.2.
**
** PARAMETERS: usmUserName    -  output buffer for the USM Sec User Name
**
** RETURNS:    None
**
** NOTE:
*****************************************************************************
*/
static void makeUsmUserSecurityName(char* usmUserName)
{
   char  mac[6];
   char  macAddr[32];

   if (!usmUserName)
   {
      return;
   }

   cmMiscGetMtaMacAddr (mac);
   sprintf(macAddr, "%02X:%02X:%02X:%02X:%02X:%02X", mac[0],mac[1],mac[2],mac[3],mac[4],mac[5]);
   strcpy (usmUserName, "MTA-Prov-");
   strcat(usmUserName, macAddr);

} /* end of makeUsmUserSecurityName() */



/*
*****************************************************************************
** FUNCTION:   mergeServersInfo
**
** PURPOSE:    Merges the Servers Info received from CM (stored in
**             the 1st parameter) into Provis Data Block
**
** PARAMETERS: ppktcServrInfo - PKTC Servers Info Block from CM
**             provBlockp     - Provisioining Info Block (received from
**                              CFG-0 intially)
**
** RETURNS:    0 if OK, <>0 - otherwise
**
** NOTE:       If both sources of PKTC Servers Info are presented -
**             CFG-0 and CM, then the info is OR-ed with the CFG's info
**             having a priority (2nd parameter).
*****************************************************************************
*/
static int mergeServersInfo(PacketCableServers* ppktcServrInfo, PRZPROVIS* provBlockp)
{
   UINT32   addrLength;    /* length of the DNS address */
   char     *pAddr;        /* pointer to the DNS address */
   UINT32   ipAddress;     /* IP address in numerical format */

    /* Sanity check */
   if (!ppktcServrInfo || !provBlockp)
   {
      return 1; /* Too BAD! */
   }

   /* get pointer to and length of the DHCP server address string */
   pAddr = ppktcServrInfo->dhcpServerAddress;
   addrLength = strlen(pAddr);

   /* Save the DHCP IP address */
   if (addrLength && ppktcServrInfo->foundDHCPServerAddress )
   {
      /* Do this only is CFG-0 didn't already set it */
      if (!provBlockp->prxMtaDhcp.prxSrvrIpAddr)
      {
         bcmNetIpAddrToNum(pAddr, &ipAddress);
         provBlockp->prxMtaDhcp.prxSrvrIpAddr = ipAddress;

#if 0
         if (!provBlockp->prxMtaDhcp.prxSrvrIpAddr)
         {
            printf("ERROR: DHCP IP address in config file is wrong.\n");
            return 1; /* Too BAD! */
         }
#endif
         strcpy(provBlockp->prxMtaDhcp.prxSrvrIpAddrStr, pAddr);
         bcmNetGetPortNum( pAddr, &provBlockp->prxMtaDhcp.prxSrvrPort );

         /* Also store DHCP entry in MIB */
         provBlockp->prxMibAccessFunc(MIBITEM(PRV, DHCP1), NULL, NULL, NULL,
                                             (void*)&ipAddress, MIB_ITEM_SET);
      }
   }

   /* get pointer to and length of the DHCP server address string */
   pAddr = ppktcServrInfo->dhcpServerAddress2;
   addrLength = strlen(pAddr);

   /* Save the DHCP IP address */
   if (addrLength && ppktcServrInfo->foundDHCPServerAddress2 )
   {
      /* Do this only is CFG-0 didn't already set it */
      if (!provBlockp->prxMtaDhcp2.prxSrvrIpAddr)
      {
       LOG(("Setting DHCP2 to 0x%x", ipAddress));
       bcmNetIpAddrToNum(pAddr, &ipAddress);
         provBlockp->prxMtaDhcp2.prxSrvrIpAddr = ipAddress;

         strcpy(provBlockp->prxMtaDhcp2.prxSrvrIpAddrStr, pAddr);
         bcmNetGetPortNum( pAddr, &provBlockp->prxMtaDhcp2.prxSrvrPort );

         /* Also store DHCP entry in MIB */
         provBlockp->prxMibAccessFunc(MIBITEM(PRV, DHCP2), NULL, NULL, NULL,
                                             (void*)&ipAddress, MIB_ITEM_SET);
      }
   }

   /* get pointer to and length of the SNMP server address string */
   pAddr = ppktcServrInfo->snmpEntityAddress;
   addrLength = strlen(ppktcServrInfo->snmpEntityAddress);

   /* Save the SNMP IP address */
   if (addrLength && ppktcServrInfo->foundSNMPEntityAddress)
   {
      /* Do this only is CFG-0 didn't already set it */
      if (!provBlockp->prxProvSnmp.prxSrvrIpAddr)
      {
         bcmNetIpAddrToNum(pAddr, &provBlockp->prxProvSnmp.prxSrvrIpAddr);
         strcpy(provBlockp->prxProvSnmp.prxSrvrIpAddrStr, pAddr);
         bcmNetGetPortNum( pAddr, &provBlockp->prxProvSnmp.prxSrvrPort );

         /* Also store SNMP entry in MIB */
         provBlockp->prxMibAccessFunc(MIBITEM(PRV, SNMP), NULL, NULL, &addrLength,
                                             (void*)&pAddr, MIB_ITEM_SET);
      }
   }

   /* get pointer to and length of primary DNS address string */
   pAddr = ppktcServrInfo->primaryDNSAddress;
   addrLength = strlen(pAddr);

   /* Save the Primary DNS IP address */
   if (addrLength && ppktcServrInfo->foundPrimaryDNSAddress)
   {
      /* Do this only is CFG-0 didn't already set it */
      if (!provBlockp->prxMtaDns1.prxSrvrIpAddr)
      {
         bcmNetIpAddrToNum(pAddr, &ipAddress);
         provBlockp->prxMtaDns1.prxSrvrIpAddr = ipAddress;

#if 0
         if (!provBlockp->prxMtaDns1.prxSrvrIpAddr)
         {
            printf("ERROR: DNS-1 IP address in PKTC Server Info is wrong.\n");
            return 1; /* Too BAD! */
         }
#endif
         strcpy(provBlockp->prxMtaDns1.prxSrvrIpAddrStr, pAddr);
         bcmNetGetPortNum( pAddr, &provBlockp->prxMtaDns1.prxSrvrPort );

         /* Also store primary DNS entry in MIB */
         provBlockp->prxMibAccessFunc(MIBITEM(PRV, DNS1), NULL, NULL, NULL,
                                             (void*)&ipAddress, MIB_ITEM_SET);
      }
   }

   /* get pointer to and length of secondary DNS address string */
   pAddr = ppktcServrInfo->secondaryDNSAddress;
   addrLength = strlen(pAddr);

   /* Save the Secondary DNS IP address */
   if (addrLength && ppktcServrInfo->foundSecondaryDNSAddress)
   {
      /* Do this only is CFG-0 didn't already set it */
      if (!provBlockp->prxMtaDns2.prxSrvrIpAddr)
      {
#if 0
         bcmNetIpAddrToNum(pAddr, &provBlockp->prxMtaDns2.prxSrvrIpAddr);
         if (!provBlockp->prxMtaDns2.prxSrvrIpAddr)
         {
            printf("ERROR: DNS-2 IP address in PKTC Server Info is wrong.\n");
            return 1; /* Too BAD! */
         }
#endif
         bcmNetIpAddrToNum(pAddr, &ipAddress);
         provBlockp->prxMtaDns2.prxSrvrIpAddr = ipAddress;

         strcpy(provBlockp->prxMtaDns2.prxSrvrIpAddrStr, pAddr);
         bcmNetGetPortNum( pAddr, &provBlockp->prxMtaDns2.prxSrvrPort );

         /* Also store secondary DNS entry in MIB */
         provBlockp->prxMibAccessFunc(MIBITEM(PRV, DNS2), NULL, NULL, NULL,
                                             (void*)&ipAddress, MIB_ITEM_SET);
      }
   }

   /* get pointer to and length of Kerberos realm name */
   pAddr = ppktcServrInfo->provKerbRealmName;
   addrLength = strlen(pAddr);

   /* Save the Kerb Realm name in MIB */
   if (addrLength && ppktcServrInfo->foundProvKerbRealmName)
   {
      provBlockp->prxMibAccessFunc(MIBITEM(PRV, KerbRealmName), NULL, NULL, &addrLength,
                                             (void*)&pAddr, MIB_ITEM_SET);

      strcpy(provBlockp->prxKdcServer.prxSrvrIpAddrStr, pAddr);
      bcmNetGetPortNum( pAddr, &provBlockp->prxKdcServer.prxSrvrPort );
   }

   LOG(("mergeServersInfo:  Setting tgt request to %d",ppktcServrInfo->tgtKerberosExec ));
   provBlockp->prxTgtRequest = ppktcServrInfo->tgtKerberosExec;


   if(ppktcServrInfo->foundProvMtaProvTimer)
   {     /*Save the provisionig timer into the MIB*/
      provBlockp->prxProvTimer = (ppktcServrInfo->provMtaProvTimer);
      provBlockp->prxMibAccessFunc(MIBITEM(PRV, ProvisioningTimer), NULL, NULL, NULL,
                                (void *)&provBlockp->prxProvTimer, MIB_ITEM_SET);
   }



   return 0; /* OK */

} /* end of mergeServersInfo() */

/*
*****************************************************************************
** FUNCTION:   CheckStandAlone
**
** PURPOSE:    Checks to see if we are working standalone
**             without the CM part.
**
** PARAMETERS: None
**
** RETURNS:    1 if standalone, 0 - otherwisw
**
** NOTE:       If pSOS time is NOT set then we are standalone.
*****************************************************************************
*/
int CheckStandAlone(void)
{
   g_pProvisDescBlock->prxMtaStandalone = !(cmMiscIsCableModem());

   return (int)g_pProvisDescBlock->prxMtaStandalone;

}  /* end of CheckStandAlone() */

void  mtaSetSystemTime(int yyyy, int mm, int dd, int hh, int min, int ss, int tt)
{
   UINT32    date, time;

   date = ((yyyy & 0xFFFF) << 16 ) +
          ((mm & 0xff) << 8 ) +
          (dd);
   time = ((hh & 0xFF) << 16) + ((min & 0xFF) << 8) +
         (ss & 0xFF);

   bcmOsSetTime(date, time, tt);

} /* end of mtaSetSystemTime() */

/*
*****************************************************************************
** FUNCTION:   GenerateNonce
**
** PURPOSE:    This function generates "nonce" random number according
**             to the Provis Spec (I01). It must be generated each 'time
**             an MTA is powered up.
**
** PARAMETERS: None
**
** RETURNS:    None
**
** NOTE:
*****************************************************************************
*/
void  GenerateNonce(PRZPROVIS* provBlockp)
{
   if (!provBlockp) /* Sanity check */
   {
      return;
   }

   provBlockp->prxErrCode = 0;
   mtaRndSeed();

   provBlockp->prxCorrelationId = mtaRand();
   /* set Correlation Id */
   provBlockp->prxMibAccessFunc(MIBITEM(PRV, CorrelationId), NULL, NULL, NULL,
                                 (void*)&provBlockp->prxCorrelationId, MIB_ITEM_SET);

   printf("Correlation-ID = 0x%08x.\n", provBlockp->prxCorrelationId);

} /* end of GenerateNonce() */


#define  PROV_ERROR_CRYT   0x1100


/*
*****************************************************************************
** FUNCTION:   prfTlvConfigFileDownload
**
** PURPOSE:    Main function for the TFTP download.
**
** PARAMETERS: provBlockp - pointer to  PRZPROVIS structure.
**
** RETURNS:    None
**
** NOTE:       Retrieves the TLV formatted Configuration File from
**             TFTP Server.
*****************************************************************************
*/
void prfTlvConfigFileDownload(PRZPROVIS* provBlockp)
{
   UINT32       tftpIpAddr, rc;
   char        *pcfgFileName, *decryptedBuf;
   char        ipAddrStr[16];
   int         bytesRead, i;
   char        tempBuffer[MAX_MTA_DEV_PROV_CONFIG_HASH_SIZE], tempKeyBuf[8];


   if (!provBlockp) /* Sanity check */
   {
      return;
   }

   tftpIpAddr = provBlockp->prxTlvTftp.prxSrvrIpAddr;
   pcfgFileName = provBlockp->prxtftpTlvFile;
   provBlockp->prxErrCode = 0;

   bcmNetNumToIpAddr( tftpIpAddr, ipAddrStr );
   LOG(( "TFTP Started. TFTP Server - %s, TFTP File - %s.",
                                       ipAddrStr,
                                       pcfgFileName ));

   rc = bcmOsTftpReadFile( (UINT32)tftpIpAddr, pcfgFileName, cmGetIpStackIndex(), prgCfgBuffer,
                            sizeof(prgCfgBuffer), (UINT32*)&prgCfgBytesRead, NULL );
   if (rc)
   {
      provBlockp->prxErrCode = PROV_TASK_ERR(0x1032); /* Set error code */
      LOG(("ERROR: TFTP Client failed."));
      return;
   }

   /* process hash data */
   if (hashLength)
   {
      LOG(("Processing HASH"));
      /* Calculate hash of prgCfgBuffer and compare with hashBuffer */
      if (hashLength == MD5_SIZE)
      {
         rc = secLibCalcHash(HASH_MD5_TYPE, prgCfgBuffer, prgCfgBytesRead, NULL, 0, tempBuffer, &(unsigned long)bytesRead);
      }
      else if (hashLength == SHA1_SIZE)
      {
         rc = secLibCalcHash(HASH_SHA1_TYPE, prgCfgBuffer, prgCfgBytesRead, NULL, 0, tempBuffer, &(unsigned long)bytesRead);
      }
      else
      {
         provBlockp->prxErrCode = PROV_TASK_ERR(0x1050); /* Set error code */
         LOG(("ERROR: Failed to match hash type."));
         return;
      }

      if (rc)
      {
         printf("Received:\n");
         for (rc=0; rc<hashLength; rc++)
         {
            printf("%x ", hashBuffer[rc]);
         }
         printf("\nCalculated:\n");
         for (rc=0; rc<hashLength; rc++)
         {
            printf("%x ", tempBuffer[rc]);
         }
         printf("\n");

         provBlockp->prxErrCode = PROV_TASK_ERR(0x1051); /* Set error code */
         LOG(("ERROR: Failed to calculate hash."));
         return;
      }

      /* Compare hash */
      if (hashLength == bytesRead)
      {
         rc = strncmp((char*)hashBuffer, tempBuffer, (size_t)bytesRead);
         if (rc)
         {
            provBlockp->prxErrCode = PROV_TASK_ERR(0x1052); /* Set error code */
            LOG(("ERROR: Hash does not match."));
            return;
         }
      }
      else
      {
         provBlockp->prxErrCode = PROV_TASK_ERR(0x1053); /* Set error code */
         LOG(("ERROR: Hash length does not match."));
         return;
      }

      LOG(("Hash is verified."));
   }

   /* decrypt config file */
   if (keyLength)
   {
      LOG(("Decrypting..."));

      decryptedBuf = malloc(TFTP_MAX_BUFLEN);
      /* Decryption */
      memcpy(tempKeyBuf, keyBuffer, 8);
      rc = secCryptoDes(prgCfgBuffer, decryptedBuf, (int*) &prgCfgBytesRead, tempKeyBuf, 0);

      if (!rc)
      {
         provBlockp->prxErrCode = PROV_TASK_ERR(0x1054); /* Set error code */
         LOG(("ERROR: Failed to decrypt config file."));
         free(decryptedBuf);
         return;
      }
      for (rc=0; rc<prgCfgBytesRead; rc++)
      {
         prgCfgBuffer[rc]=decryptedBuf[rc];
      }
      free(decryptedBuf);
      LOG(("Config file is decrypted successfully"));
   }

   /* Here CONFIG data is successfully retrieved!!! */
   rc = provParseConfigInfo(prgCfgBuffer, prgCfgBytesRead);
   if (rc)
   {
      provBlockp->prxErrCode = PROV_TASK_ERR(PROV_ERROR_CRYT); /* Set error code */
      LOG(("ERROR: Failed to Parse Config Data."));
      return;
   }

   /* Let's check that all MUST TLV data is in place. */
   rc = mibValidateMustTlvData();
   if (rc)
   {
      provBlockp->prxErrCode = PROV_TASK_ERR(PROV_ERROR_CRYT + 1); /* Set error code */
      LOG(("ERROR: Some Required data is missing from the TLV file."));
      return;
   }


} /* end of prfTlvConfigFileDownload() */



/*
*****************************************************************************
** FUNCTION:   prfGetTlvConfigFile
**
** PURPOSE:    Retrieves TLV configuration file from the TFTP Sever.
**
** PARAMETERS: provBlockp - pointer to PRZPROVIS block.
**
** RETURNS:    0 - if OK, otherwise - FAIL
**
** NOTE:
*****************************************************************************
*/
UINT32 prfGetTlvConfigFile(PRZPROVIS*  provBlockp)
{
   UINT32 rc;


   if (!provBlockp) /* Sanity check */
   {
      return 1;
   }
   provBlockp->prxErrCode = 0;

   if (!provBlockp->prxTlvTftp.prxSrvrIpAddr && !strlen(provBlockp->prxTlvTftp.prxSrvrIpAddrStr) )
   {
      return 0;
   }

   LOG(( "MTA TFTP Task Starting."));

   /*
   ** There is no TlvTFTP Server to download the info from, no need to do so.
   ** If TLV Config file TFTP Server IP Address has not been defined
   ** DNS might be able to resolve it.
   */

   if (!provBlockp->prxTlvTftp.prxSrvrIpAddr && strlen(provBlockp->prxTlvTftp.prxSrvrIpAddrStr) )
   {
     bcmNetIpAddrToNum( provBlockp->prxTlvTftp.prxSrvrIpAddrStr, &provBlockp->prxTlvTftp.prxSrvrIpAddr);
   }
   if (!provBlockp->prxTlvTftp.prxSrvrIpAddr)
   {
      rc = 0x1030;
      provBlockp->prxErrCode = PROV_TASK_ERR(rc); /* Set error code */
      printf("ERROR: TFTP Server IP Address (""%s"") cannot be resolved. Err=0x%x", provBlockp->prxTlvTftp.prxSrvrIpAddrStr, rc);
      return rc;
   }

   /*
   ** Here is the Code to activate the TFTP client.
   */

   /* Retrieve the Config Info from the TFTP server. */
   prfTlvConfigFileDownload( provBlockp );

   if (provBlockp->prxErrCode)
   {
      rc = provBlockp->prxErrCode;
      LOG(("ERROR: TFTP task failed. Err=0x%x", rc));
   }
   else
   {
      rc = 0;
      LOG(("TFTP Configuration file has been received."));
   }

   return rc;

} /* end of prfGetTlvConfigFile() */


/*
*****************************************************************************
** FUNCTION:   prfDhcpProv
**
** PURPOSE:    Initiates DHCP client for DHCP provisioning.
**
** PARAMETERS: provBlockp - pointer to PRZPROVIS structure.
**
** RETURNS:    0 if OK, <>0 - otherwise
**
** NOTE:
*****************************************************************************
*/
UINT32 prfDhcpProv(PRZPROVIS*  provBlockp)
{
   extern PacketCableServers  gpktcServerInfo;    /* defined in dncputils.c */
   dhcpcfg_t                  dhcp_cfg;
   UINT32                     rc, dhcpStarted = 0, failCounter = 0;

   /*
      If EMTA DHCP is disabled, there is nothing much to do here.
      The IP Address of the EMTA MUST be assigned statically.
   */
   if (FALSE == cmNvIsMtaDhcpEnabled())
   {
      return 0;
   }

   if (!provBlockp->prxMtaDhcp.prxSrvrIpAddr)
   {
      /*  ERROR: DHCP Server IP Address is NOT defined */
      rc = PROV_TASK_ERR(0x1020); /* Set error code */
      provBlockp->prxErrCode = rc;
      LOG(("ERROR: prfDhcpProv() - Missing EMTA DHCP Server from CM. Err=0x%x", rc));
      return rc;
   }

   while(1)
   {
      /* Clear error code before every attempt */
      provBlockp->prxErrCode = 0;

      LOG(( "\nDHCP Provisioning Start. Attempt=%d\n", failCounter+1));

      /* Here is the Code to activate the DHCP client. */

      bcmOsSemCreate(PROV_TASK_SEMA_GEN, &prgsynSema);
      if (!prgsynSema)
      {
         rc = PROV_TASK_ERR(0x1005); /* Set error code */
         printf("ERROR: Failure to create DHCP Sema. Err=0x%x", rc);
         goto p_fail;
      }

      dhcp_cfg.task_prio = provBlockp->prxprioDhcpTask;
      dhcp_cfg.log_msgs = 0;

      rc = cmGetIpStackIndex();
      printf("==>DBG!!! prfDhcpProv(): IPStackIdx=%d\n", rc);

      /* Get ready. Releasing is the preferred method over stopping
      ** during restarts */
      dhcpRel();
      dhcpStarted = 0;

      /* go */
      rc = dhcpStart(rc, &prgsynSema, provBlockp);

      if (rc)
      {
         rc = PROV_TASK_ERR(rc); /* Set error code */
         LOG(("ERROR: DHCP startup error: 0x%x", rc));
         goto p_fail;
      }

      dhcpStarted = 1;
      /*
         Let's always make COLD DHCP start (no "warm" start). Just to be consistent with the CM side.
      */
      rc = dhcpCold();

      if (rc)
      {
         rc = PROV_TASK_ERR(rc); /* Set error code */
         printf("ERROR: DHCP Init error: 0x%x", rc);
         goto p_fail;
      }

      /*
      ** Let's wait until we execute all DHCP related procedures and return
      ** (regardless - successfully or not).
      */
      rc  = bcmOsSemTake(prgsynSema, SEC2TICKS(PROV_TASK_DHCP_TO));
      if (rc == OS_STATUS_TIMEOUT)
      {
         rc = PROV_TASK_ERR(0x1007);
         LOG(("ERROR: DHCP procedure has timed out.\n"));
         goto p_fail;
      }

      if (provBlockp->prxErrCode)
      {
         rc = provBlockp->prxErrCode;
         printf("ERROR: DHCP procedure failed. Err=0x%x",provBlockp->prxErrCode);
         goto p_fail;
      }

      dhcpStarted = 0;
      rc = mergeServersInfo(&gpktcServerInfo, provBlockp);
      if (rc)
      {
         provBlockp->prxErrCode = rc;
         printf("ERROR: Wrong info in ServersInfo data in DHCP procedure.\n");
         goto p_fail;
      }

      LOG(("DHCP Provisioning OK."));
      provBlockp->prxDhcpActive = 1;
      rc = 0;

p_fail:
      if (dhcpStarted && rc)
      {
         dhcpStop(); /* Stop DHCP Client */
         dhcpStarted = 0;
      }
      provBlockp->prxErrCode = rc;
      bcmOsSemDelete(prgsynSema);

      if (!rc)
      {
         /* exit while(1) */
         break;
      }
      failCounter++;
      bcmOsSleep((TEN_MSEC*1001)); /* just make some time to delete the sema and DHCP task */

   }

   return rc;

} /* end of prfDhcpProv() */

/*
*****************************************************************************
** FUNCTION:   prfParseTlvFileName
**
** PURPOSE:    Parses the Config File Name being done in URL format:
**             tftp://<TFTP_IP>/<CfgFileName>
**
** PARAMETERS: ptlvCfgFname   - Config File Name in URL format
**             provBlockp     - Provisioning Data block.
**
** RETURNS:    0 - if TFTP IP and FileName have been set. <> 0  - Otherwise
**
** NOTE:
*****************************************************************************
*/
static int prfParseTlvFileName(char* ptlvCfgFname, PRZPROVIS   *provBlockp)
{
   char  *p = ptlvCfgFname, *pFQDN, *pPort = NULL, *pFname = NULL;
   int   l;

   if (!ptlvCfgFname || !provBlockp)   /* Just being causeous */
   {
      return 1;   /* Too bad */
   }

   provBlockp->prxErrCode = 0;

   l = strlen(p);
   if ( ( l > 255) || !l )
   {
      return 1;   /* Too bad */
   }

   /* Convert the string to lower case for better convenience */
   while(*p)
   {
      *p++ = (char)tolower((int)*p);
   }

   p = ptlvCfgFname;

   if (memcmp("tftp", p, 4))
   {
      return 1;   /* The format is wrong */
   }

   p += 4; /* skip 1st 4 chars */

   if (memcmp("://", p, 3))
   {
      return 1;   /* The format is wrong */
   }

   p += 3; /* skip next chars */
   pFQDN = p; /* We found an FQDN */

   for ( ; *p; p++)
   {
      if (*p == ':')
      {  /* We found the port delimiter */
         *p = 0;
         pPort = (p + 1);
      }
      if (*p == '/')
      {  /* We found the Cfg File Name delimiter */
         *p = 0;
         pFname = (p + 1);
      }
   }

   if (pFname <= pPort + 2)
   {
      return 1;   /* The format is wrong */
   }

   strcpy(provBlockp->prxTlvTftp.prxSrvrIpAddrStr, pFQDN);
   strcpy(provBlockp->prxtftpTlvFile, pFname);
   provBlockp->prxTlvTftp.prxSrvrPort = 0;
   if (pPort)
   {
      provBlockp->prxTlvTftp.prxSrvrPort = atol(pPort);
   }

   return 0;

} /* end of prfParseTlvFileName() */


/*
*****************************************************************************
** FUNCTION:   prfWaitForTlvCfgFileName
**
** PURPOSE:    Waits up to 'PROV_TASK_TLV_CFG_TO" seconds for a new TLV Cfg
**             File Name to be set in the MIB (via SNMP).
**
** PARAMETERS: provBlockp  - Provisioning Data block.
**
** RETURNS:    0 - if TFTP IP and FileName have been set. Otherwise - if FAIL.
**
** NOTE:
*****************************************************************************
*/
static int prfWaitForTlvCfgFileName(PRZPROVIS*  provBlockp)
{
   char        *ptlvCfgFname;
   OS_STATUS   stat;
   int         rc;
   static      char  cfgTlvFileName[256] = {0};
   int         tmr, done=0, hashFlag=0, configFlag=0, keyFlag=0;

   if (!provBlockp) /* Sanity check */
   {
      return 1;
   }

   provBlockp->prxErrCode = 0;
   tmr = PROV_WAIT_FOR_CFGNAME_TO; /* (sec) */
   LOG(("Waiting for TLV Configuration File...\n"));

   while (!done)
   {
      stat = bcmOsEventReceive( &setCfgFnameEvt, SEC2TICKS( tmr ) );

      switch(stat)
      {
         case OS_STATUS_OK:
            /* get pointer to mibGetMtaDevConfigFile */
            if (!configFlag)
            {
               provBlockp->prxMibAccessFunc(MIBITEM(PRV, ConfigFile), NULL, NULL, NULL, (void*)&ptlvCfgFname, MIB_ITEM_GET);
               if (strlen(ptlvCfgFname)>0)
               {
                  configFlag = 1;
                  strcpy(cfgTlvFileName, ptlvCfgFname);
                  LOG(("MTA Configuration File Name has been set by ProvServer = %s\n.", ptlvCfgFname));
                  tmr = 5;
               }
            }
            if (!hashFlag)
            {
               provBlockp->prxMibAccessFunc(MIBITEM(PRV, ConfigHash), NULL, NULL, &hashLength, (void*)&hashBuffer, MIB_ITEM_GET);
               if (hashLength!=0)
               {
                  hashFlag = 1;
                  LOG(("MTA Configuration Hash has been set by ProvServer"));
                  tmr = 5;
               }
            }
            if (!keyFlag)
            {
               provBlockp->prxMibAccessFunc(MIBITEM(PRV, ConfigKey), NULL, NULL, &keyLength, (void*)&keyBuffer, MIB_ITEM_GET);
               if (keyLength!=0)
               {
                  keyFlag = 1;
                  LOG(("MTA Configuration Key has been set by ProvServer"));
                  tmr = 5;
               }
            }
            if (keyFlag && hashFlag && configFlag)
            {
               done = 1;
            }
            break;
         case OS_STATUS_TIMEOUT:
            if (!configFlag)
            {
               LOG(("ERROR: prfWaitForTlvCfgFileName(): TIMEOUT of waiting for MTA Config File Name.\n"));
               return 1;
            }
            if (keyFlag && !hashFlag)
            {
               LOG(("ERROR: prfWaitForTlvCfgFileName(): TIMEOUT of waiting for MTA Config HASH file.\n"));
               return 1;
            }
            done = 1;
            break;
         default:
            printf("ERROR: prfWaitForTlvCfgFileName(): Failure on \"ev_receive\". Ret=0x%x\n", stat);
            return 1;
      }
   }

   rc = prfParseTlvFileName(cfgTlvFileName, provBlockp);
   if (rc)
   {
      printf("ERROR: prfWaitForTlvCfgFileName(): TLV File Name is not in URL Format\n");
      return 1;
   }

   return 0;   /* the new Tlv Cfg File Name was set */

} /* end of prfWaitForTlvCfgFileName() */


/*
*****************************************************************************
** FUNCTION:   prfSetCfgFileNameCB()
**
** PURPOSE:    Called when "mibMtaDevConfigFileName" is SET. Sets the synch
**             object for the "waitForTlvCfgFileName()" routine to wait on.
**
** PARAMETERS: lastmatch   - last index of the OID
**             compc       - number of OIDs in the object
**             compl       - ptr to OID object.
**             vbp         - ptr binding info for the OID.
**             data        - user data being passed to the callback:
**                           data[0] - ptr to OID of "mibMtaDevConfigFileName" obj
**                           data[1] - TaskID to send the event to
**
** RETURNS:    1 if the OID being passed is NOT of interest, 0 - if OK
**
** NOTE:
*****************************************************************************
*/
static int prfSetCfgFileNameCB(
                                    UINT32         lastmatch,
                                    int            compc,
                                    UINT32*        compl,
                                    SNMP_OBJ_ID*   pOid,
                                    SNMP_VAR_TYPE  dataType,
                                    UINT32         dataLength,
                                    void*          dataValue,
                                    UINT32         data[4]  /* Arbirtary data which can be passed to the callback function */
                               )
{
   SNMP_OBJ_ID*   pcfgObj = (SNMP_OBJ_ID*)data[0]; /* Object OID we are only interested in - ConfigFileName*/

   LOG(("Config_File_Name Object has been set and \"prfSetCfgFileNameCB()\" is called!\n"));
   if (!pcfgObj || !pOid)
   {              /* If this happens something is seriously ill !!! */
      return 1;   /* No vaild data to compare */
   }

   if ((!snmpIsOidEqual(pcfgObj, pOid))
    && (!snmpIsOidEqual(pcfgObj+1, pOid))
    && (!snmpIsOidEqual(pcfgObj+2, pOid)))
   {
      return 1;
   }

   /* Let's inform the waiting party we got the Config File Name */
   bcmOsEventSend( &setCfgFnameEvt );

   return 0;

} /* end of prfSetCfgFileNameCB() */

#if (PROV_CFG_SECURITY == 1)
/*
 *****************************************************************************
 ** FUNCTION:    CMSAuthBlock
 **
 ** PURPOSE:     get the authBlock for endpt
 **
 ** PARAMETERS:  endpt - endpoint ID
 **              authBlock - authBlock for the endpoint
 **
 ** RETURNS:     success or failure
 **
 ** NOTE:
 **
 *****************************************************************************
 */
static UINT32 CMSAuthBlock(
      int endpt,
      KRBCLI_AUTH_BLOCK *authBlock)
{
   UINT32               caLen;
   IPADDR               tempIpAddr;
   char                 *realmName;
   UINT32               realmNameLen;
   int                  rc;
   char                 dns_srv_query[255];
   char                 realm_list[255];
   char                 *p;
   NET_STATUS           status;


   authBlock->serviceType = APPLICATION_CMS;
   strcpy(authBlock->szMtaMac, g_pProvisDescBlock->prxMacAddrStr);
   strcpy(authBlock->szMtaFQDN, g_pProvisDescBlock->prxMtaFqdn);
   strcpy(authBlock->szProvServerFQDN, g_pProvisDescBlock->prxProvSnmp.prxSrvrIpAddrStr);

   /* get the CMS FQDN for this endpoint */
   mibAccessNcs( endpt, MIB_NCSITEMID_CaID, &caLen, authBlock->szServiceName, MIB_ACCESS_GET);
   if ( authBlock->szServiceName[0] == '\0' )
   {
      LOG(("ERROR: CMSAuthBlock() - For EndpointID %d, Could not find CMS FQDN from MIB.", endpt));
      return 1;
   }
   LOG(("INFO: CMSAuthBlock() - EndpointID = %d, CMS FQDN(authBlock->szServiceName) is %s.",
         endpt, authBlock->szServiceName));
   bcmNetIpAddrToNum(authBlock->szServiceName, &tempIpAddr);
   bcmNetNumToIpAddr(tempIpAddr, authBlock->szServiceIpAddress);

   /* Get the realm for this CMS from MIB */
   rc = mibAccessSec (
         MIBITEM(SEC, cmsRealm),
         NULL,
         authBlock->szServiceName,
         NULL,
         &realmNameLen,
         (void*)&realmName,
         MIB_ITEM_GET);
   if ( rc )
   {
      LOG(("ERROR: CMSAuthBlock() - For EndpointID %d, Could not get CMS Realm form MIB.", endpt));
      return rc;
   }
   else
   {
      LOG(("INFO: CMSAuthBlock() - For CMS %s realm is %s.",
               authBlock->szServiceName, realmName));
      strcpy(authBlock->szKrbServiceRealm, realmName);
   }

   /* DNS SRV for realm */
   strcpy(dns_srv_query, "_kerberos._udp.");
   strcat(dns_srv_query, authBlock->szKrbServiceRealm);
   bcmOsDnsGetSRVName(dns_srv_query, realm_list);
   if ( realm_list[0] == '\0' )
   {
      LOG(("ERROR: CMSAuthBlock() - NO DNS SRV list for srv record %s.", dns_srv_query));
      return 1;
   }
   LOG(("INFO: CMSAuthBlock() - For SRV record %s, first DNS SRV is %s.", dns_srv_query, realm_list));

   p = realm_list;
   while ( *p != '\0' )
   {
      status = bcmNetIpAddrToNum(p, &tempIpAddr);
      if (status == NET_STATUS_OK )
      {
         bcmNetNumToIpAddr(tempIpAddr, authBlock->szKdcIpAddress);
         LOG(("INFO: CMSAuthBlock() - DNS SRV resolved. KDC IP address %s.",  authBlock->szKdcIpAddress));
         break;
      }
      else
      {
         LOG(("WARNING: CMSAuthBlock() - Could not resolve DNS SRV entry list %s.", p));
         p += strlen(p)+1;
         LOG(("\t\tTrying next entry %s.", p));
      }
   }

   authBlock->serverNonce = 0;
   authBlock->krbReason = KRB_BOOTUP;
   g_pProvisDescBlock->prxTgtRequest = gpktcServerInfo.tgtKerberosExec;
   authBlock->tgtFlag = g_pProvisDescBlock->prxTgtRequest;

   /* pass the rekey info gathered from the rekey message */
   authBlock->rekeyInfoBlock.authTID = 0;
   authBlock->rekeyInfoBlock.encTID = 0;
   authBlock->rekeyInfoBlock.secParamLifeTime = 0;
   authBlock->rekeyInfoBlock.gracePeriod = 0;

   return 0;
} /* end of CMSAuthBlock() */

/*
 *****************************************************************************
 ** FUNCTION:    CMSAsdReq
 **
 ** PURPOSE:     get the AsdReq
 **
 ** PARAMETERS:  authBlock - authBlock for the endpoint
 **
 ** RETURNS:     success or failure
 **
 ** NOTE:
 **
 *****************************************************************************
 */
static UINT32 CMSASDReq(int endpt,
      CMS_ASD_APREQ *asdReq)
{
   asdReq->inbound_SPI = rand();
   return 0;
}


/*
*****************************************************************************
** FUNCTION:   prfCmsKrbClient()
**
** PURPOSE:    Implements the Kerberized Authentication and Key Negotiation
**             for the CMS IPSEC Provisionig
**
** PARAMETERS: provBlockp  - Provisioning Block
**
** RETURNS:    0 if OK, otherwise - FAIL
**
** NOTE:       1. Make sure the function is called AFTER the DNS is activated.
**                Otherwise the Krb Realm will not be resolved.
**             2. This function will actually do something only if Kerb Realm
**                (suboption-6) is NOT empty.
*****************************************************************************
*/
static int prfCmsKrbClient(PRZPROVIS *provBlockp)
{
   UINT32               rc=1;
   IPADDR               provIpAddr;
   KRBCLI_AUTH_BLOCK    krbCliBlock;
   char                 ipAddrStr[32];
   CMS_ASD_APREQ       asdApReq;
   CMS_ASD_APREP       asdApRep;
   CMS_CLIENT_KEYS     clientKeys;
   int                 i;
   UINT32              caLen;


#if  (PROV_PC_COMPLIANT == 0)
   /*
      EN: The following is NOT PC compliant! Remove it!
   */
   if (!strlen(provBlockp->prxKdcServer.prxSrvrIpAddrStr))
   {
      return 0;
   }
#endif

   setProvState(provBlockp, mtaProvState_waitingForProvRealmKdcAddrResponse);

   /*
      EN: TBD:
      The following line must be changed by teh proper DNS SRV Resolution
      and then FQDN->IP_Addr resolution sequence when pSOS DNS Resolver
      is ready for that.
   */
   bcmNetIpAddrToNum(provBlockp->prxKdcServer.prxSrvrIpAddrStr, &provBlockp->prxKdcServer.prxSrvrIpAddr);

   if (!provBlockp->prxKdcServer.prxSrvrIpAddr)
   {
      LOG(("ERROR: prfCmsKrbClient() - Cannot resolve the ProvKerberos Realm!"));
      return 1;
   }

   setProvState(provBlockp, mtaProvState_waitingForAsReply);


   for (i=0;i<GLOBAL_CFG_MAX_CNX;i++)
   {
      rc = CMSAuthBlock(i, &krbCliBlock);
      if (rc)
      {
         LOG(("ERROR: prfCmsKrbClient() - For Endpoint %d CANNOT build CMS authBlock!", i));
         continue;
      }

      rc = CMSASDReq(i, &asdApReq);
      if (rc)
      {
         LOG(("ERROR: prfCmsKrbClient() - For Endpoint %d CANNOT build CMS ASD Request!", i));
         continue;
      }


      /* Do AS and AP exchange */
      rc = secKrb5ClientAuthenticate(
            &krbCliBlock,
            (void*)&asdApReq,
            (void*)&asdApRep,
            (void*)&clientKeys );

   }

   return (rc) ? 1 : 0;

} /* end of prfCmsKrbClient() */


/*
 *****************************************************************************
 ** FUNCTION:    ProvAuthBlock
 **
 ** PURPOSE:     get authBlock for provisioning
 **
 ** PARAMETERS:  authBlock for provisioning
 **
 ** RETURNS:     success or failure
 **
 ** NOTE:
 **
 *****************************************************************************
 */
static UINT32 ProvAuthBlock(KRBCLI_AUTH_BLOCK *authBlock)
{
   IPADDR               tmpIpAddr;
   char                 dns_srv_query[255];
   char                 realm_list[255];
   char                 *p;
   NET_STATUS           status;

   authBlock->serviceType = APPLICATION_PROV;
   strcpy(authBlock->szMtaMac, g_pProvisDescBlock->prxMacAddrStr);
   strcpy(authBlock->szMtaFQDN, g_pProvisDescBlock->prxMtaFqdn);
   strcpy(authBlock->szKrbServiceRealm, g_pProvisDescBlock->prxKdcServer.prxSrvrIpAddrStr);
   strcpy(authBlock->szServiceName, g_pProvisDescBlock->prxProvSnmp.prxSrvrIpAddrStr);
   strcpy(authBlock->szProvServerFQDN, g_pProvisDescBlock->prxProvSnmp.prxSrvrIpAddrStr);

   bcmNetIpAddrToNum(g_pProvisDescBlock->prxProvSnmp.prxSrvrIpAddrStr, &tmpIpAddr);
   bcmNetNumToIpAddr(tmpIpAddr, authBlock->szServiceIpAddress);

   /* DNS SRV for realm */
   strcpy(authBlock->szKrbServiceRealm, g_pProvisDescBlock->prxKdcServer.prxSrvrIpAddrStr);
   strcpy(dns_srv_query, "_kerberos._udp.");
   strcat(dns_srv_query, authBlock->szKrbServiceRealm);
   bcmOsDnsGetSRVName(dns_srv_query, realm_list);
   if ( realm_list[0] == '\0' )
   {
      LOG(("NO DNS SRV list for srv record %s.", dns_srv_query));
      return 1;
   }
   LOG(("For srv record %s, first DNS SRV is %s.", dns_srv_query, realm_list));

   p = realm_list;
   while ( *p != '\0' )
   {
      status = bcmNetIpAddrToNum(p, &tmpIpAddr);
      if (status == NET_STATUS_OK )
      {
         bcmNetNumToIpAddr(tmpIpAddr, authBlock->szKdcIpAddress);
         LOG(("DNS SRV resolved. KDC IP address %s.",  authBlock->szKdcIpAddress));
         break;
      }
      else
      {
         LOG(("Could not resolve dns srv entry list %s.", p));
         p += strlen(p)+1;
         LOG(("Trying next entry %s.", p));
      }
   }

   authBlock->serverNonce = 0;
   authBlock->krbReason = KRB_BOOTUP;
   g_pProvisDescBlock->prxTgtRequest = gpktcServerInfo.tgtKerberosExec;
   authBlock->tgtFlag = g_pProvisDescBlock->prxTgtRequest;

   /* pass the rekey info gathered from the rekey message */
   authBlock->rekeyInfoBlock.authTID = 0;
   authBlock->rekeyInfoBlock.encTID = 0;
   authBlock->rekeyInfoBlock.secParamLifeTime = 0;
   authBlock->rekeyInfoBlock.gracePeriod = 0;

   return 0;
} /* end of ProvAuthBlock() */

/*
 *****************************************************************************
 ** FUNCTION:    provASDReq
 **
 ** PURPOSE:     get ASD request info for provisioning
 **
 ** PARAMETERS:  asdReq for provisioning
 **
 ** RETURNS:     success or failure
 **
 ** NOTE:
 **
 *****************************************************************************
 */
static UINT32 ProvASDReq(PROV_ASD_APREQ *asdReq)
{
   char  temp[256];
   int   eLen, eTime, eBoot, i;

   /* ASD for AP REQ */
   if (cmSnmpGetEngineParams (temp, sizeof(temp), &eLen, &eBoot, &eTime))
   {
      LOG(("ERROR: ProvASDReq() - Cannot Retrieve EngineID Parameters"));
      return 1;
   }


   memcpy(asdReq->engId, temp, eLen);
   asdReq->engIdLen = (char)eLen;
   asdReq->engBoots = eBoot;
   asdReq->engTime = eTime;
   makeUsmUserSecurityName(asdReq->userName);
   asdReq->userNameLen = (char)strlen(asdReq->userName);

   printf("INFO: ProvASDReq() - engineId(len=%d)=", eLen);
   for (i=0; i<eLen; i++)
   {
      printf("0x%02X ", asdReq->engId[i]);
   }
   printf("\n");

   LOG(("INFO: ProvASDReq() - engineBoots=%d, engineTime=%d\n", eBoot, eTime));

   return 0;

} /* end of ProvASDReq() */


/*
*****************************************************************************
** FUNCTION:   prfProvKrbClient()
**
** PURPOSE:    Implements the Kerberixed Authentication and Key Negotiation
**             for the SNMPv3 Provisionig
**
** PARAMETERS: provBlockp  - Provisioning Block
**             pClientKeys - SNMPv3 Keys
**             pAsdApRep   - App Specific Data for Provisioning (AP-REP)
**
** RETURNS:    0 if OK, otherwise - FAIL
**
** NOTE:       1. Make sure the function is called AFTER the DNS is activated.
**                Otherwise the Krb Realm will not be resolved.
**             2. This function will actually do something only if Kerb Realm
**                (suboption-6) is NOT empty.
*****************************************************************************
*/
static int prfProvKrbClient(PRZPROVIS *provBlockp, PROV_CLIENT_KEYS  *pClientKeys, PROV_ASD_APREP  *pAsdApRep)
{
extern PacketCableServers  gpktcServerInfo;    /* defined in dncputils.c */
   char                 ipAddrStr[32];
   UINT32               rc;
   IPADDR               provIpAddr;
   KRBCLI_AUTH_BLOCK    krbCliBlock;
   PROV_ASD_APREQ       asdApReq;


#if  (PROV_PC_COMPLIANT == 0)
   /*
      EN: The following is NOT PC compliant! Remove it!
   */
   if (!strlen(provBlockp->prxKdcServer.prxSrvrIpAddrStr))
   {
      LOG(("WARNING: prfProvKrbClient(): NoKDC Server Set - Simulate Prov Authentication()."));
      memset((void*)&krbCliBlock, 0, sizeof(PROV_CLIENT_KEYS));
      memset((void*)pAsdApRep, 0, sizeof(PROV_ASD_APREP));
      memset((void*)&asdApReq, 0, sizeof(PROV_ASD_APREQ));
      krbCliBlock.serviceType = APPLICATION_PROV;
      krbCliBlock.krbReason = KRB_BOOTUP;
      krbCliBlock.tgtFlag = gpktcServerInfo.tgtKerberosExec;

/* We need to call the following just to make a call to prfSetSnmpV3() fucntion to Init SNMPv3 in
   NoAuthNoPriv mode */
      rc = secKrb5ClientAuthenticate(
            &krbCliBlock,
            (void*)&asdApReq,
            (void*)pAsdApRep,
            (void*)pClientKeys);

      return (rc) ? 1 : 0;
   }
#endif

   setProvState(provBlockp, mtaProvState_waitingForProvRealmKdcAddrResponse);

   bcmNetIpAddrToNum(provBlockp->prxKdcServer.prxSrvrIpAddrStr, &provBlockp->prxKdcServer.prxSrvrIpAddr);

   if (!provBlockp->prxKdcServer.prxSrvrIpAddr)
   {
      LOG(("ERROR: prfProvKrbClient() - Cannot resolve the ProvKerberos Realm!"));
      return 1;
   }

   setProvState(provBlockp, mtaProvState_waitingForAsReply);

   /* fill in the provisioning info */
   rc = ProvAuthBlock(&krbCliBlock);
   if ( rc )
   {
      LOG(("ERROR: prfProvKrbClient() - Can not build provisioning authBlock!"));
      return 1;
   }

   /* fill up the ASD */
   rc = ProvASDReq(&asdApReq);
   if ( rc )
   {
      LOG(("ERROR: prfProvKrbClient() - Can not build provisioning ASD !"));
      return 1;
   }


   rc = secKrb5ClientAuthenticate(
         &krbCliBlock,
         (void*)&asdApReq,
         (void*)pAsdApRep,
         (void*)pClientKeys);

   return (rc) ? 1 : 0;

} /* end of prfProvKrbClient() */


static int  prfSetSnmpV3(PROV_CLIENT_KEYS  *pSnmpKeys, PROV_ASD_APREQ *pProvAsdApReq, PROV_ASD_APREP  *pProvAsdApRep)
{
   int   i;

   LOG(("INFO: prfSetSnmpV3() - called SNMPv3 Key Assosiation Function."));
   if (!pSnmpKeys || !pProvAsdApReq || !pProvAsdApRep)
   {
      LOG(("*** PANIC *** prfSetSnmpV3() - Intput Parameter(s) is a NULL pointer!"));
      return 1;
   }

   switch(pSnmpKeys->provAuthKey.tid)
   {
      case SNMPV3_HMAC_MD5_TID:
         pSnmpKeys->provAuthKey.tid = SNMP_AUTH_MD5;
         break;
      case SNMPV3_HMAC_SHA1_TID:
         pSnmpKeys->provAuthKey.tid = SNMP_AUTH_SHA1;
         break;
      default:
         pSnmpKeys->provAuthKey.tid = 0;
   }
   pSnmpKeys->provEncrKey.tid = (pSnmpKeys->provEncrKey.tid == SNMPV3_DES_TID) ? 1 : 0;

   g_pProvisDescBlock->prxSnmpMgrEngineIDLen = pProvAsdApRep->engIdLen;
   memcpy(g_pProvisDescBlock->prxSnmpMgrEngineID, pProvAsdApRep->engId, g_pProvisDescBlock->prxSnmpMgrEngineIDLen);
   g_pProvisDescBlock->prxSnmpMgrUserNameLen = pProvAsdApRep->userNameLen;
   memcpy(g_pProvisDescBlock->prxSnmpMgrUserName, pProvAsdApRep->userName, g_pProvisDescBlock->prxSnmpMgrUserNameLen);
   g_pProvisDescBlock->prxSnmpMgrEngineBoot = pProvAsdApRep->engBoots;
   g_pProvisDescBlock->prxSnmpMgrEngineTime = pProvAsdApRep->engTime;

   g_pProvisDescBlock->prxSnmpMgrUserName[g_pProvisDescBlock->prxSnmpMgrUserNameLen] = 0;
   LOG(("INFO: prfSetSnmpV3() - SNMP Mgr's Engine Patrameters:"));
   LOG(("\tSNMP Mgr's Name Length - %d", g_pProvisDescBlock->prxSnmpMgrUserNameLen));
   LOG(("\tSNMP Mgr's Name - %s", g_pProvisDescBlock->prxSnmpMgrUserName));
   LOG(("\tSNMP Mgr's EngineID Length - %d", g_pProvisDescBlock->prxSnmpMgrEngineIDLen));
   LOG(("\tSNMP Mgr's EngineID - "));
   for (i=0; i<g_pProvisDescBlock->prxSnmpMgrEngineIDLen; i++)
   {
      printf("%02x ", g_pProvisDescBlock->prxSnmpMgrEngineID[i]);
   }
   LOG((""));
   LOG(("\tSNMP Mgr's Engine Boot - %d", g_pProvisDescBlock->prxSnmpMgrEngineBoot));
   LOG(("\tSNMP Mgr's Engine Time - %d", g_pProvisDescBlock->prxSnmpMgrEngineTime));

   if (snmpEngineInit(g_pProvisDescBlock, PROV_PKTC_USM_VIEW, pSnmpKeys->provAuthKey.tid,
                     pSnmpKeys->provAuthKey.key, pSnmpKeys->provEncrKey.tid, pSnmpKeys->provEncrKey.key)
      )
   {
      g_pProvisDescBlock->prxErrCode = PROV_TASK_ERR(0x1041);
      LOG(("ERROR: prfSetSnmpV3() - Failure to initialize SNMP Engine."));
      return 1;
   }

   return 0;

} /* end if prfSetSnmpV3() */

/*
 *****************************************************************************
 ** FUNCTION:    prfGetSecInfoBlock
 **
 ** PURPOSE:     Retrieve the Security Info Block Data with the call back functions.
 **
 ** PARAMETERS:  secProvInfoBlock the shared block
 **
 ** RETURNS:
 **
 ** NOTE:
 **
 *****************************************************************************
 */
void prfGetSecInfoBlock(SEC_PROV_INFO_BLOCK *secProvInfoBlock)
{
   secProvInfoBlock->ProvAuthBlockCB = ProvAuthBlock;
   secProvInfoBlock->CmsAuthBlockCB = CMSAuthBlock;
   secProvInfoBlock->ProvASDReqCB = ProvASDReq;
   secProvInfoBlock->CmsASDReqCB = CMSASDReq;
   secProvInfoBlock->IpsecKeyInfoCB = sigSecIpsecKeyInfo;
   secProvInfoBlock->SnmpV3KeyInfoCB = prfSetSnmpV3;
} /* end of prfGetSecInfoBlock() */


#endif /* #PROV_CFG_SECURITY */


/*
*****************************************************************************
** FUNCTION:   prfResetNowCB
**
** PURPOSE:    Called when the pktcMtaDevResetNow MIB object is set
**
** PARAMETERS: unused
**
** RETURNS:    0
**
** NOTE:
*****************************************************************************
*/
static int prfResetNowCB(
                           UINT32         lastmatch,
                           int            compc,
                           UINT32*        compl,
                           SNMP_OBJ_ID*   pOid,
                           SNMP_VAR_TYPE  dataType,
                           UINT32         dataLength,
                           void*          dataValue,
                           UINT32         data[4]  /* Arbirtary data which can be passed to the callback function */
                        )
{
   UINT32         val;
   SNMP_OBJ_ID*   pcfgObj = (SNMP_OBJ_ID*)data[0]; /* Object OID we are only interested in - ResetNow*/

   if (!pcfgObj || !pOid || !dataValue)
   {              /* If this happens something is seriously ill !!! */
      LOG(("***PANIC*** prfResetNowCB(): Wrong Parameter passed into function!!!\n"));
      return 1;   /* No vaild data to compare */
   }

   /* The OID is NOT ours, it's for somebody else! */
   if (!snmpIsOidEqual(pcfgObj, pOid))
   {
      return 1;
   }

   /* YES, this OID is ours, let's react correspondingly! */
   val = *((UINT32 *)dataValue);

   /* Check if pktcMtaDevResetNow was set to TRUE(1) */
   if ( val == 1 )
   {
      LOG(("ResetNow MIB Object was set and \"prfResetNowCB()\" is called!\n"));
      if ( g_pProvisDescBlock && g_pProvisDescBlock->prxMibEventNotifyFunc )
      {
         g_pProvisDescBlock->prxMibEventNotifyFunc(MIBEVENT_MTARESETNOW);
      }
   }

   return 0;

} /* end of prfResetNowCB */

/*
*****************************************************************************
** FUNCTION:   prfRegisterResetNow
**
** PURPOSE:    Called to register the ResetNow Callback which handles reset
**
** PARAMETERS: NONE
**
** RETURNS:    NONE
**
** NOTE:
*****************************************************************************
*/
void prfRegisterResetNow(void)
{
   UINT32         data_buf[4] = {0};
   OS_TASKID      taskId;
   UINT32         rc;
   static UINT32  oid_ResetNow[] =
                     {CLABS_OID_ROOT, 2, 2, 1, 1, 1, 1, 0}; /* pktcMtaDevResetNow */

   static SNMP_OBJ_ID   oidObjs[] =
   {
      {sizeof(oid_ResetNow)/sizeof(UINT32), oid_ResetNow},
       {0, NULL}         /* MUST be the last statement */
   };

   bcmOsTaskIdentSelf(&taskId);
   data_buf[0] = (UINT32)oidObjs;
   data_buf[1] = (UINT32)taskId;
   snmpMibSetCB(prfResetNowCB, data_buf);

} /* end of prfRegisterResetNow() */

/*
*****************************************************************************
** FUNCTION:   prfUnregisterResetNow
**
** PURPOSE:    Called to unregister the ResetNow Callback
**
** PARAMETERS: NONE
**
** RETURNS:    NONE
**
** NOTE:
*****************************************************************************
*/
void prfUnregisterResetNow(void)
{
   snmpMibClearCB( prfResetNowCB );

} /* end of prfRegisterResetNow() */

/*
*****************************************************************************
** FUNCTION:   prfSnmpProv()
**
** PURPOSE:    Implements the whole cycle of SNMP dependent parts of the
**             proviosning: sending SNMP Inform trap, waiting on SNMP SET for
**             TLV Config File Name.
**
** PARAMETERS: provBlockp  - Provisioning Block
**
** RETURNS:    0 if OK, otherwise - FAIL
**
** NOTE:       1. Make sure the function is called AFTER the DNS is activated.
**                Otherwise the SNMP Mgr FQDN will not be resolved.
**             2. This function will actually do something only if SNMP IP
**                Address string is NOT empty.
*****************************************************************************
*/
static int prfSnmpProv(PRZPROVIS *provBlockp)
{
   UINT32      data_buf[4] = {0};
   OS_TASKID   taskId;
   UINT32       rc;
   static      UINT32   oid_CfgFileName[] =
                           {CLABS_OID_ROOT,2,2,1,1,2,5,0};    /* pktcMtaDevConfigFile (DispString) */
   static      UINT32   oid_CfgHash[] = {CLABS_OID_ROOT, 2, 2, 1, 1, 2, 7, 0};
   static      UINT32   oid_CfgKey[]  = {CLABS_OID_ROOT, 2, 2, 1, 1, 2, 8, 0};

   static   SNMP_OBJ_ID    oidObjs[] =
   {
      {sizeof(oid_CfgFileName)/sizeof(UINT32), oid_CfgFileName},
      {sizeof(oid_CfgHash)/sizeof(UINT32), oid_CfgHash},
      {sizeof(oid_CfgKey)/sizeof(UINT32), oid_CfgKey},
      {0, NULL}         /* MUST be the last statement */
   };

   provBlockp->prxErrCode = 0;

   if (!provBlockp->prxProvSnmp.prxSrvrIpAddr && !strlen(provBlockp->prxProvSnmp.prxSrvrIpAddrStr) )
   {
      return 0;   /* OK: just no SNMP Server has been defined */
   }

   /* Let's make sure that SNMP server IP address can be resolved with DNS (if it was 0) */
   if (!provBlockp->prxProvSnmp.prxSrvrIpAddr && strlen(provBlockp->prxProvSnmp.prxSrvrIpAddrStr) )
   {
      bcmNetIpAddrToNum(provBlockp->prxProvSnmp.prxSrvrIpAddrStr, &provBlockp->prxProvSnmp.prxSrvrIpAddr);
   }

   if (!provBlockp->prxProvSnmp.prxSrvrIpAddr)
   {
      provBlockp->prxErrCode = PROV_TASK_ERR(0x1008);
      printf("ERROR: SNMP Mgr FQDN (\"%s\") cannot be resolved.\n", provBlockp->prxProvSnmp.prxSrvrIpAddrStr);
      return provBlockp->prxErrCode;   /* Too bad! */
   }

   /* Before we send the EnrolInform Trap, lets set the "MibSet" Callback for ConfigFileName OID
      to wait for in "waitForTlvCfgFileName()".
   */
   bcmOsTaskIdentSelf(&taskId);
   data_buf[0] = (UINT32)oidObjs;
   data_buf[1] = (UINT32)taskId;
   rc = snmpMibSetCB(prfSetCfgFileNameCB, data_buf);
   if (rc)
   {
      provBlockp->prxErrCode = PROV_TASK_ERR(0x1009);
      printf("ERROR: Cannot set the MIB SET_Callback Function for CfgFileName. Err = 0x%x.\n", provBlockp->prxErrCode);
      return provBlockp->prxErrCode;   /* Too bad! */
   }

   /* Send "enroll" Trap/Inform to the SNMP Entity (if set) */
   rc = snmpSendEnrolTrap(provBlockp);
   if (rc)
   {
      provBlockp->prxErrCode = PROV_TASK_ERR(0x1010);
      printf("ERROR: Enrollment INFORM Failed. Err = 0x%x.\n", provBlockp->prxErrCode);
      return provBlockp->prxErrCode;   /* Too bad! */
   }

   /*
   ** Let's wait for the SNMP Set with the config File Name to come in.
   ** The file is int URL notation so, we'll parse it out and
   ** prepare for the TFTP-ing at the later stage.
   */
   rc = prfWaitForTlvCfgFileName(provBlockp);
   snmpMibClearCB (prfSetCfgFileNameCB); /* We do not need the "prfSetCfgFileNameCB" Callback function any more */

   if (rc)
   {
      provBlockp->prxErrCode = PROV_TASK_ERR(0x1011);
      printf("ERROR: Waiting for the Config Filie Name Failed. Err = 0x%x.\n", provBlockp->prxErrCode);
      return 1;
   }

   return 0;

} /* end of prfSnmpProv() */

/*
*****************************************************************************
** FUNCTION:   setProvisioningState()
**
** PURPOSE:    Sets the Provisioning State value of the "ProvisioningState"
**             MIB Object
**
** PARAMETERS: state - correponding state of the boot process
**
** RETURNS:    NONE
**
** NOTE:
**
*****************************************************************************
*/
static void setProvisioningState(PRZPROVIS  *provBlockp, int state)
{
   if (provBlockp)
   {
      provBlockp->prxMibAccessFunc(MIBITEM(PRV, ProvisioningState), NULL, NULL, NULL, (void*)&state, MIB_ITEM_SET);
   }

} /* end of setProvisioningState() */

/*
*****************************************************************************
*****************************************************************************
*/
static void dispProvBlock(PRZPROVIS*  provBlockp)
{
   printf("Provisioning Block Data:\n");
   printf("   DHCP1 Server - %s\n"
        "   DHCP2 Server - %s\n"
          "   Prov Server - %s\n"
          "   DNS1 Server - %s\n"
          "   DNS2 Server - %s\n"
          "   Prov KDC Server - %s\n"
          "   MTA FQDN - %s\n"
          "   MTA MAC Addr - %s\n"
          "   Last Err Code - 0x%x\n",
               provBlockp->prxMtaDhcp.prxSrvrIpAddrStr,
               provBlockp->prxMtaDhcp2.prxSrvrIpAddrStr,
               provBlockp->prxProvSnmp.prxSrvrIpAddrStr,   /* MTA Provis SNMP Entity */
               provBlockp->prxMtaDns1.prxSrvrIpAddrStr,   /* MTA Provis Primary DNS */
               provBlockp->prxMtaDns2.prxSrvrIpAddrStr,   /* MTA Provis Secondary DNS */
               provBlockp->prxKdcServer.prxSrvrIpAddrStr,  /* Provisioning KDC Server */
               provBlockp->prxMtaFqdn, /* MTA FQDN set from the DHCP (option-12) */
               provBlockp->prxMacAddrStr, /* MTA MAC Address */
               provBlockp->prxErrCode          /* Error Code to be returned to the Provis task */
         );
} /* end of dispProvBlock() */

static void updateMibInfo(PRZPROVIS*  provBlockp)
{
   char     *pAddr;
   UINT32   length;

   /* Store DHCP entry in MIB */
   provBlockp->prxMibAccessFunc(MIBITEM(PRV, DHCP1), NULL, NULL, NULL,
                                             (void*)&(provBlockp->prxMtaDhcp.prxSrvrIpAddr), MIB_ITEM_SET);
     /* Store DHCP2 entry in MIB */
   provBlockp->prxMibAccessFunc(MIBITEM(PRV, DHCP2), NULL, NULL, NULL,
                                             (void*)&(provBlockp->prxMtaDhcp2.prxSrvrIpAddr), MIB_ITEM_SET);

   /* Store primary and secondary DNS entry in MIB */
   provBlockp->prxMibAccessFunc(MIBITEM(PRV, DNS1), NULL, NULL, NULL,
                                             (void*)&(provBlockp->prxMtaDns1.prxSrvrIpAddr), MIB_ITEM_SET);
   provBlockp->prxMibAccessFunc(MIBITEM(PRV, DNS2), NULL, NULL, NULL,
                                             (void*)&(provBlockp->prxMtaDns2.prxSrvrIpAddr), MIB_ITEM_SET);
   /* Store the FQDN into the MIB */
   pAddr = provBlockp->prxMtaFqdn;
   length = strlen(pAddr);
   provBlockp->prxMibAccessFunc(MIBITEM(PRV, FQDN), NULL, NULL, (UINT32*)&length,
                                             (void*)&(pAddr), MIB_ITEM_SET);

  /* Store the Telephony Syslog server address into the MIB*/
  pAddr = provBlockp->prxMtaSysLogServer;
  length = strlen(pAddr);
  provBlockp->prxMibAccessFunc(MIBITEM(PRV, SysLog), NULL, NULL, (UINT32*)&length,
                                             (void*)&(pAddr), MIB_ITEM_SET);


} /* end of updateMibInfo */

static void updateMibProvCounter (PRZPROVIS* provBlockp, UINT32 *counter)
{

   provBlockp->prxMibAccessFunc(MIBITEM(PRV, ProvisioningCounter), NULL, NULL, NULL,
                                             (void*)counter, MIB_ITEM_SET);

} /*updateMibProvCounter*/

/*
*****************************************************************************
** FUNCTION:   prfProvTaskInitCB
**
** PURPOSE:    Task initialization function for the Provisioning task.
**
** PARAMETERS:
**
** RETURNS:    None
**
** NOTE:
*****************************************************************************
*/
void prfProvTaskInitCB(void)
{
   OS_TASKID taskId;

   bcmOsTaskIdentSelf( &taskId );
   bcmOsEventCreate( &setCfgFnameEvt, &taskId );
}

/*
*****************************************************************************
** FUNCTION:   prfProvTaskDeinitCB
**
** PURPOSE:    Task de-initialization function for the Provisioning task.
**
** PARAMETERS:
**
** RETURNS:    None
**
** NOTE:
*****************************************************************************
*/
void prfProvTaskDeinitCB(void)
{
   LOG(( "\nDe-initing PROV task ...\n"));

   bcmOsEventDelete( &setCfgFnameEvt );

   LOG(( "\nPROV task done.\n" ));

   /* Not stopping DHCP */

#if 0
   /* check for restarting/deleting */
   printf("\nIn the PROV_TASK, checking for TFTP task ... \n");

   bcmOsTaskRestart( TFTP_TASK_NAME, NULL );

   printf("\nStopping DHCP ... \n");
   dhcpStop();

   printf("\nDeleting PROV task ...\n");

   CLEANUP_TASK;
   bcmOsTaskDelete(0);
#endif
}

/*
*****************************************************************************
** FUNCTION:   prfProvTaskMain
**
** PURPOSE:    Main function for the Provisioning task.
**
** PARAMETERS: context - passes the pointer to PRZPROVIS structure.
**
** RETURNS:    None
**
** NOTE:
*****************************************************************************
*/
void prfProvTaskMain( void *context )
{
   PRZPROVIS*        provBlockp;
   UINT32            rc,failCounter, tempprovcounter;
   OS_STATUS         status;
   IPADDR            dnsSrvList[ OS_DNS_MAX_SERVERS ];
   int               nDnsSrvs;
   int               krb_cont_flag;
   OS_TASKID         taskId;
   IPADDR            ipAddr;
   PROV_CLIENT_KEYS  snmpKeys = {0};
   PROV_ASD_APREP    provAsdApRep;

   provBlockp = (PRZPROVIS*)context;

   for (failCounter=0; ; failCounter++)
   {
      if ( appOsTaskIsResetPending() )
      {
         appOsTaskResetAck();
         return;
      }

      /*
      ** (Re)starting provisioning, need to:
      **   - update counters, unregister resetMibCB, turn off EventLogging
      */
      LOG(( "\nStart Provisioning Attempt #%d...\n", failCounter + 1));
      provBlockp->prxProvCntr++;
      /* update mib counter variable */
      updateMibProvCounter(provBlockp, &provBlockp->prxProvCntr);
      prfUnregisterResetNow();
      /* disableEventLogging(); */

      /*
      *****  DHCP *******
      */

      /* Let's arrange DHCP provisioning part. If it's running it won't start again. */
      rc = prfDhcpProv(provBlockp);
      if (rc)
      {
         provBlockp->prxErrCode = PROV_TASK_ERR(0x1043);
         LOG(("ERROR: DHCP Attempt Failure. Err = 0x%x.", provBlockp->prxErrCode));

         /* Restart initialization from DHCP Provisioning (MTA1) */
         provBlockp->prxDhcpActive = 0;
         continue;
      }

      dispProvBlock(provBlockp);

      updateMibInfo(provBlockp);

      GenerateNonce(provBlockp);


      /*
      ** DNS
      */

      /* Let's start DNS sevices. If it's running it won't start again. */
      setProvState(provBlockp, mtaProvState_waitingForAsReply);

      /* Create list of DNS server addresses */
      dnsSrvList[0] = provBlockp->prxMtaDns1.prxSrvrIpAddr;
      dnsSrvList[1] = provBlockp->prxMtaDns2.prxSrvrIpAddr;

      /* Determine the number of valid DNS servers */
      nDnsSrvs = 0;
      if ( dnsSrvList[0] )
      {
         nDnsSrvs++;
      }
      if( dnsSrvList[1] )
      {
         nDnsSrvs++;
      }

      /* Start the DNS server */
      LOG(("DNS starting. DNS-1 IpAddr - 0x%x, DNS-2 IpAddr - 0x%x.\n", dnsSrvList[0], dnsSrvList[1] ));
      if (!provBlockp->prxDnsActive)
      {
         status = bcmOsStartDns( dnsSrvList, nDnsSrvs, cmGetIpStackIndex());
      }
      else
      {
         /*
            We'll need to use bcmOsModifyDns(...) API call which currently does not exist.
         */
         status = bcmOsStartDns( dnsSrvList, nDnsSrvs, cmGetIpStackIndex());
      }

      if ( status != OS_STATUS_OK )
      {
         /* Log error */
         LOGEVENT( PROVERR_01 );

         /* if error: Let's try again */
         provBlockp->prxErrCode = PROV_TASK_ERR( status );
         printf("ERROR: DNS Client Start Failed. Err = 0x%x.\n", provBlockp->prxErrCode);

         /* Restart initialization from DHCP Provisioning (MTA1) */
         provBlockp->prxDhcpActive = 0;
         continue;
      }
      else
      {  /* This is needed to possibly modify the list if DNS Srvrs! */
         provBlockp->prxDnsActive = 1;
      }

      memset((void*)&snmpKeys, 0, sizeof(snmpKeys));
      memset((void*)&provAsdApRep, 0, sizeof(provAsdApRep));


      /*
      ***** SNMP (Secure) *******
      */

      /*
         Let's Register MTA's IP Address to SNMP Agent if we are NOT DHCP-ing.
         Otherwise - Registering is done during the DHCP process.
      */
      if (!cmNvIsMtaDhcpEnabled())
      {
         bcmNetGetOwnIpAddress(cmGetIpStackIndex(), &ipAddr);
         LOG(("prfProvisionMain(): Registering to SNMP Agent in StandAlone_Mode - MyOwnIP_Addr(IPstack#=%d) = 0x%08X\n", cmGetIpStackIndex(), ipAddr));
         if (cmSnmpMtaAgentRegister(ipAddr, cmGetIpStackIndex()))
         {
            provBlockp->prxErrCode = PROV_TASK_ERR(0x1040);
            LOG(("ERROR: prfProvisionMain() - Failure to register the MTA IP Stack to SNMP Agent. Err = 0x%x.", provBlockp->prxErrCode));
            /* Restart initialization from DHCP Provisioning (MTA1) */
            provBlockp->prxDhcpActive = 0;
            continue;
         }
         LOG(("PROV: prfProvisionMain() - MTA is registered to SNMP Agent."));
      }

      makeUsmUserSecurityName(provBlockp->usmUserSecName);

#if (PROV_CFG_SECURITY == 1)

/*
   Let's perform Kerberized Authentication and SNMPv3 Key Negotiation
   with the Provisioning KDC.
   If Prov Authentication is disabled by TGT-flag=0, it returns "0" (OK).
   If we fail here, Prov Process will loop back to the very beg starting from DHCP.
*/
      rc = prfProvKrbClient(provBlockp, &snmpKeys, &provAsdApRep);
      if (rc) /* If we */
      {
         provBlockp->prxErrCode = PROV_TASK_ERR(0x1042);
         LOG(("ERROR: Provisioning Kerberos Failure. Err = 0x%x.", provBlockp->prxErrCode));

         /* Restart initialization from DHCP Provisioning (MTA1) */
         provBlockp->prxDhcpActive = 0;

         /* if kerberos fails, complete provisioning or go back to dhcp
          * depending on non-vol flag
         */
         cmNvGetMtaKrbContProv(&krb_cont_flag);
         if (  !krb_cont_flag )
         {
            bcmOsSleep((TEN_MSEC*500)); /* If we don't do that, CM's SNMP will not work!!! */
            continue;
         }
      }

#endif   /* #PROV_CFG_SECURITY */


      /*
      ** SNMP complete, register reset callback and enable EventLogging
      */
      prfRegisterResetNow();
      /* enableEventLogging(); */


      /*
      ***** Config file *******
      */

      /* Let's start DNS sevices. If it's running it won't start again. */
      setProvState(provBlockp, mtaProvState_waitingForConfigFile);

      rc = prfSnmpProv(provBlockp);
      if (rc)
      {
         /* Restart initialization from DHCP Provisioning (MTA1) */
         provBlockp->prxDhcpActive = 0;
         continue;
      }

      prfGetTlvConfigFile(provBlockp);
      if (!provBlockp->prxErrCode)  /* If successful we are done! */
      {
         break;
      }
      else
      {
         /*
            If the error happened due to the incorrectness of the Config File then we have to finish
            trying and inform SNMP Entity and (possible) SYSLOG about thye Prov Results.
         */
         if (provBlockp->prxErrCode >= PROV_TASK_ERR(PROV_ERROR_CRYT))
         {
            break;
         }

         /* Restart initialization from DHCP Provisioning (MTA1 flow) */
         provBlockp->prxDhcpActive = 0;

      }

   } /* End for (failCounter) provisioning repetition loop */


   /*
   ** If we get here and the error code is indicated then the Provisioning Process
   ** has failed after all "retries", and we need to stop everything related to it.
   */
   if (provBlockp->prxErrCode)
   {
      setProvisioningState(provBlockp, mtaProvisioningState_Fail);
      snmpSendProvisStatTrap(provBlockp);

      if (provBlockp->prxProvTimer == 0)
      {
         /* If the provTimer is not enabled, give the semaphore immediately */
         bcmOsSemGive(prgsemaProvstart); /* Let the user task continue running */
      }
      /* If the provisioning timer is enabled, do not give semaphore
      ** it will time out when the provtimer expires eventually */
   }
   else
   {
      /* MTA is provisioned and operational! */
      setProvisioningState(provBlockp, mtaProvisioningState_Success);
      setProvState(provBlockp, mtaProvState_operational);
      snmpSendProvisStatTrap(provBlockp);
      bcmOsSemGive(prgsemaProvstart); /* Let the user task continue running */
   }

   /*
   ** Exit from main and let the task de-init itself
   */
   return;

} /* end of prfProvTaskMain() */


/*
*****************************************************************************
** FUNCTION:   prfProvStart
**
** PURPOSE:    This function starts all Provisioning related tasks.
**
** PARAMETERS: Pointer to the "PRZPROVIS" structure.
**
** RETURNS:    TaskID if OK,  0 - otherwise.
**
** NOTE:
*****************************************************************************
*/
UINT32 prfProvStart(PRZPROVIS* provBlockp)
{
   UINT32         provPrio = provBlockp->prxprioProvTask;
   UINT32         dhcpPrio = provBlockp->prxprioDhcpTask;
   UINT32         rc;
   OS_TASKID      curTaskId;
   OS_STATUS      status;
   OS_TASK_CLASS  tskPrio;
   int            tmr;
   BOOL           bProvDone = FALSE;


   if (!provBlockp) /* Sanity check */
   {
      return 0; /* Too bad! */
   }

   g_pProvisDescBlock = provBlockp;

   /*
   ** If priority is 0, retrieve the priority for calling task.
   */
   if (!provPrio || !dhcpPrio)
   {
      status = bcmOsTaskIdentSelf(&curTaskId); /* What is current task ID? */
      if (status != OS_STATUS_OK)
      {
         provBlockp->prxErrCode = PROV_TASK_ERR(0x1001);
         return 0; /* Too bad! */
      }

      status = bcmOsTaskGetPriority(curTaskId,  &tskPrio);    /* What is current task priority? */
      if (status != OS_STATUS_OK)
      {
         provBlockp->prxErrCode = PROV_TASK_ERR(0x1002);
         return 0;
      }
      if (!provPrio)
      {
         provPrio = tskPrio;
         provBlockp->prxprioProvTask = tskPrio;
      }
      if (!dhcpPrio)
      {
         provBlockp->prxprioDhcpTask = tskPrio;
      }
   }

   CheckStandAlone();

   /*
   ** If we are working stand alone than pSOS timer has not been set yet.
   ** We need this timer to set the securiy related stuff (nonce, CorrID, etc).
   */
   if (g_pProvisDescBlock->prxMtaStandalone)
   {
      mtaSetSystemTime(2000, 02, 21, 10, 22, 33, 0);
   }

   /* Let's grab the data assumed to be prepared for us by the CM */
   cmNetGetDhcpSuboptions(&g_PktcServerInfo);

   rc = mergeServersInfo(&g_PktcServerInfo, provBlockp);

   /* Let's Initialize PKTC MIBs */
   Init_pktcMtaMib();
   Init_pktcNcsMib();
   Init_pktcDevEvent();

   /* Let's tell the MIB that we are in progress of Provisioning */
   setProvisioningState(provBlockp, mtaProvisioningState_inProgress);

   /*
   ** Make a semaphore to wait on until Prov Task is finished.
   */
   prgsemaProvstart = 0;
   bcmOsSemCreate(PROV_TASK_SEMA_NAME, &prgsemaProvstart);
   if (!prgsemaProvstart)
   {
      provBlockp->prxErrCode = PROV_TASK_ERR(0x1003);
      LOG(("ERROR: Failure to create HSPR Sema. Err=0x%x", rc));
      return 0; /* Too bad! */
   }

   /*Initialize the provisioning counter to zero*/
   provBlockp->prxProvCntr = 0;

   /*
   ** Repeat until provisioning succeeds
   */
   while ( !bProvDone )
   {
      if ( appOsTaskIsResetPending() )
      {
         /*
         ** A reset has occurred during provisioning.
         ** Since we are a module startup routine, we return control to the
         ** caller to allow application reset to proceed.
         */
         LOG(( "prfProvStart() interrupted by a reset!" ));
         return ( 0 );
      }

      /* Let's tell the MIB that we are in progress of Provisioning */
      setProvisioningState(provBlockp, mtaProvisioningState_inProgress);

      /*
      ** Let's create a main Provisioning task - "HPRV" ("Hothaus PRoVisioning").
      ** This task will be the main task for all other tasks: DHCP, SNMP, etc.
      ** involved into the Provisioning spec.
      */
      appOsTaskCreate(  PROV_TASK_NAME, 2*PROV_TASK_STACK_SIZE, provPrio,
                        prfProvTaskInitCB,
                        (void*) prfProvTaskMain,
                        prfProvTaskDeinitCB,
                        (UINT32) provBlockp, &provTaskId);
      if (!provTaskId)
      {
         provBlockp->prxErrCode = PROV_TASK_ERR(0x1004);
         bcmOsSemDelete(prgsemaProvstart);
         LOG(("ERROR: Failure to create HPRV task. Err=0x%x", rc));
         return 0; /* Too bad! */
      }

      /*
      ** Let's wait until we execute all provisioning procedures
      */

      /* Get Provisioning timer in minutes */
      provBlockp->prxMibAccessFunc(MIBITEM(PRV, ProvisioningTimer), NULL, NULL, NULL, (void*)&tmr, MIB_ITEM_GET);
      provBlockp->prxProvTimer = tmr;
      LOG(("Provisioning In Progress... tmr=%d", tmr));
      if ( tmr )
      {
         rc = bcmOsSemTake(prgsemaProvstart, SEC2TICKS(tmr*60));
      }
      else
      {
         rc = bcmOsSemTake(prgsemaProvstart, OS_WAIT_FOREVER );
      }

      /*
      ** Four possible outcomes:
      **  1) Semaphore timed out (ProvTimer expired)
      **  2) Semaphore return reset (An application reset has occurred)
      **  3) Semaphore taken successfully
      **     a) Provisioning completed successfully
      **     b) Provisioning finished with an error
      */
      switch ( rc )
      {
         case OS_STATUS_OK:
         {
            /*
            ** Sempahore was taken successfully.
            ** Check if provisioning failed or was successful.
            */
            if ( !provBlockp->prxErrCode )
            {
               /* Let's tell the MIB that Provisioning PASS-ed */
               setProvisioningState(provBlockp, mtaProvisioningState_Success);
               setProvState(provBlockp, mtaProvState_operational);

         /*Boolean for determining if krb tasks is started*/
            g_pProvisDescBlock->bKrbStarted = FALSE;
         #if (PROV_CFG_SECURITY == 1)
               /*
               ** Let's perform the Kerberized Authentication and IPSEC Key Negotiation
               ** with the Telephony KDC.
               */

               /*
               ** For each endpoint, find the CMS fqdn, and do a AS and AP exchange
               ** Internaly AS exchange is omitted if there is a valid ticket corresponding
               */
               LOG(("Start Provisiong with CMS.\n"));
               rc = prfCmsKrbClient(provBlockp);
               if (rc)
               {
                  LOG(("WARNING: Some Endpoints not authenticated to any CMS. Err = 0x%x.",
                           provBlockp->prxErrCode));
               }

               /*
               ** Start tasks for Security associations and timeouts
               ** Also to process Wake-up and rekey messages
               */
            if ( !provBlockp->prxErrCode )
               {
                  rc = krbStartTasks();
                  if (rc)
                  {
                     LOG(("*** ERROR *** Could not create kerberos tasks.\n"));
                  }
              g_pProvisDescBlock->bKrbStarted = TRUE;
               }
            #endif /* PROV_CFG_SECURITY */

               /* We are done provisioning, exit the main loop */
               bProvDone = TRUE;
               setProvState(provBlockp, mtaProvState_operational);
               LOG(("PROVISIONING OK."));
               break;

            } /* End if ( prov succeeded ) */

            else
            {
               /* Provisioning has failed restart it */
               LOG(("ERROR: Provisioning task failed. Err=0x%x",provBlockp->prxErrCode));
               setProvisioningState(provBlockp, mtaProvisioningState_Fail);

               /* Fall-through to normal reset provtask handling */
            }
            /* break; */

         }

         case OS_STATUS_TIMEOUT:
            /*
            ** ProvTimer expired. Fall-through to reset provtask
            */
            if ( rc == OS_STATUS_TIMEOUT )
            {
               LOG(( "ProvTimer expired." ));
            }

         default:
         {
            /*
            ** Error condition. Reset provtask.
            */
            LOG(( "Resetting provTask..." ));

            /* Notify task of reset and wait for ack */
            appOsTaskResetNotify( provTaskId );

            /* Tell the task to proceed with its shutdown */
            appOsTaskProceedToShutdown( provTaskId );

            /* Wait for the task to complete its shutdown */
            appOsTaskWaitForDone( provTaskId );

            provBlockp->prxErrCode = PROV_TASK_ERR(0x1006);
            LOG(("ERROR: Provisioning task has timed out."));
         }
         break;

         case AOS_STATUS_RESET:
         {
            /* Reset has occurred during provisioning. Do nothing.
            ** Return to top of while to handle application reset */
         }
         break;

      } /* End switch ( rc ) */

   } /* End while ( !bProvDone ) */

   return provTaskId;

} /* end of prfProvStart() */


/*
*****************************************************************************
** FUNCTION:   prfProvShutdown
**
** PURPOSE:    This function shuts down all Provisioning related tasks.
**
** PARAMETERS:
**
** RETURNS:
**
** NOTE:
*****************************************************************************
*/
void prfProvShutdown( void )
{
   /*
   ** Deinitialize PKTC MIBs
   */
   Deinit_pktcMtaMib();
   Deinit_pktcNcsMib();
   Deinit_pktcDevEvent();

   /*
   ** Shutdown KRB task
   */
   if (g_pProvisDescBlock->bKrbStarted)
   {
       krbShutDownTasks();
   }

   g_pProvisDescBlock->bKrbStarted = FALSE;

   /*
   ** Shutdown provtask
   */

   /* Tell the task to proceed with its shutdown */
   appOsTaskProceedToShutdown( provTaskId );

   /* Wait for the task to complete its shutdown */
   appOsTaskWaitForDone( provTaskId );

   /*
   ** Cleanup any module resources
   */
   if ( prgsemaProvstart )
   {
      bcmOsSemDelete( prgsemaProvstart );
      prgsemaProvstart = 0;
   }
} /* end of prfProvShutdown() */

#define FQDNMAXLABELSIZE 63
/*
*****************************************************************************
** FUNCTION:   prfVerifyFQDNSyntax
**
** PURPOSE:    This function checks the syntax of an FQDN
**
** PARAMETERS:  instr - FQDN
 *              length - length of FQDN
**
** RETURNS:
 *              1 - if syntax is good
 *              0 - if bad syntax
**
** NOTE:
*****************************************************************************
*/
int prfVerifyFQDNSyntax(char * instr, int length)
{
   int i;
   int labelctr;
   int PortLength;
   char ch;
   char *pPort, *pinstr;

   /*Rules for FQDN syntax (from RFC-1034)

   The following syntax will result in fewer problems with many
   applications that use domain names (e.g., mail, TELNET).

   <domain> ::= <subdomain> | " "

   <subdomain> ::= <label> | <subdomain> "." <label>

   <label> ::= <letter> [ [ <ldh-str> ] <let-dig> ]

   <ldh-str> ::= <let-dig-hyp> | <let-dig-hyp> <ldh-str>

   <let-dig-hyp> ::= <let-dig> | "-"

   <let-dig> ::= <letter> | <digit>

   <letter> ::= any one of the 52 alphabetic characters A through Z in
   upper case and a through z in lower case

   <digit> ::= any one of the ten digits 0 through 9

   Note that while upper and lower case letters are allowed in domain
   names, no significance is attached to the case.  That is, two names with
   the same spelling but different case are to be treated as if identical.

   The labels must follow the rules for ARPANET host names.  They must
   start with a letter, end with a letter or digit, and have as interior
   characters only letters, digits, and hyphen.  There are also some
   restrictions on the length.  Labels must be 63 characters or less.
  */

   pinstr = instr;

   /*Check non-vol to see if this is enabled*/
   if (cmNvGetMtaFQDNCheckDisabled())
   {
       return 1;
   }
   /* look for Port */
   pPort = strpbrk(pinstr, ":");
   if (pPort)
   {
      *pPort = '\0';
      pPort ++;
   }
    /* check port format */
   if (pPort)
   {
      if (isdigit(*pPort))
      {
         if ( (atoi(pPort) < 1025) || (atoi(pPort) > 65535) )
         {
            return 0;
         }
      }
      else
      {
         return 0;
      }

      /* ensure the entire port string does not contain characters */
      PortLength = strlen(pPort);
      for (i=0; i<PortLength; i++, pPort++)
      {
         if (!isdigit(*pPort))
         {
            return 0;
         }
      }
   }

   /*Check the rest of the FQDN*/
   /*We are at the beginning of a label*/
   labelctr = 0;
   for (i=0; i<length; ++i)
   {
       /*statements for verifying FQDN syntax*/
       ch = *(pinstr+i);

      /*ch is a letter*/
      if (((ch >= 'a')&&(ch <= 'z'))||((ch >= 'A')&&(ch <= 'Z')))
      {
         /*Check that the label is not larger than the maximun label size (63)*/
         if (labelctr < FQDNMAXLABELSIZE)
         {
            labelctr++;
            continue;
         }
         else
            return 0;
      }
      else if (((ch >= '0')&&(ch <= '9'))||(ch == '-')||(ch=='_'))
      {
         /*FQDN is faulty if the first character in a label is not a letter
          * or if we've gone over the maximum label size*/
         if ((labelctr==0)||(labelctr>=FQDNMAXLABELSIZE))
            return 0;
         else
            labelctr++;

         continue;
      }
      else if (ch == '.')
      {
        /*When the string contains a "." we know we've reached the end of a label. CHeck the previous character to
         * make sure it was a letter or number */
         ch = *(pinstr+(i-1));
         if (( labelctr == 0)||(i == length-1))
         {
           /*First character must be a letter, and label (and FQDN) must end with a letter or number*/
            return 0;
         }
         else if ((((ch >= 'a')&&(ch <= 'z'))||((ch >= 'A')&&(ch <= 'Z'))) ||
                   ((ch >= '0')&&(ch <= '9')))
         {
            /*Character before the "." was a letter or number.  Label is OK, reset counter*/
            labelctr = 0;
            continue;
         }
         else
            return 0;
      }
      else
      {
         return 0;
      }
   }
   return 1;

}


