/***************************************************************************
*    Copyright 2000  Broadcom Corporation
*    All Rights Reserved
*    No portions of this material may be reproduced in any form without the
*    written permission of:
*             Broadcom Corporation
*             16251 Laguna Canyon Road
*             Irvine, California  92618
*    All information contained in this document is Broadcom Corporation
*    company private, proprietary, and trade secret.
*
****************************************************************************
*
*    Filename: apreq.c
*    Creation Date: 26 Jan 2001 (v0.00)
*    VSS Info:
*        $Revision: 1.11 $
*        $Date: 2005/08/15 15:01:04 $
*
****************************************************************************
*    Description:
*
*       Implementation of packetCable AP_REQ and AP_REP
*
****************************************************************************/
 
#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include <string.h>
#include <time.h>
#ifdef TARGETOS_eCos
#include <hausenv.h>
#endif
#include <haus_cpu.h>
#include "seckrb.h"
#ifdef TARGETOS_eCos
///#include <xbcmos.h>
#endif
#include <SecLib.h>
#include <krb5client.h>
#ifdef PSOS
    #include <MibAccessDef.h>
    #include <MibAccessSec.h>
#endif
#include "krb5_err.h"
#include "krbbuf.h"
#include "pc_apreq.h"
#include "subkeyStore.h"
#include "serverAuthKey.h"
#include <bcmosHelper.h>
#include <SecUtils.h>
#ifdef DEMO
int ap_fopened_once1 = 0;
int ap_fopened_once2 = 0;
int ap_fopened_once3 = 0;
int ap_fopened_once4 = 0;
int ap_fopened_once5 = 0;
int ap_fopened_once6 = 0;
#endif
extern int gMaxRetries;
extern int gTimeout;

krb5_error_code decode_krb5_ap_rep_len(
                                      const krb5_data * code,
                                      int *rep);

/*
*****************************************************************************
** FUNCTION:     aprepParseProvASD
**
** PURPOSE:      parse the ASD pointer for AS Rep from Provisioning
**
** PARAMETERS:   pProvAsd - out - result ASD pointer
**               msg - in - AS Rep message pointer
**
** RETURNS:      int - length of ASD
**
** NOTE:
**
**
*****************************************************************************
*/
int aprepParseProvASD(PROV_ASD_APREP *pProvAsd, char *msg)
{
    int i=0;
#if (CPU_LITTLE_ENDIAN != 0)
    UINT32 temp;
#endif

    pProvAsd->engIdLen = msg[i++];
    memcpy(&(pProvAsd->engId), &msg[i], pProvAsd->engIdLen);
    i += pProvAsd->engIdLen;

#if (CPU_LITTLE_ENDIAN != 0)
    memcpy(&temp, &msg[i], 4);
    pProvAsd->engBoots = ntohl(temp);
    i += 4;
    memcpy(&temp, &msg[i], 4);
    pProvAsd->engTime = ntohl(temp);
    i+=4;
#else
    memcpy(&(pProvAsd->engBoots), &msg[i], 4);
    i += 4;
    memcpy(&(pProvAsd->engTime), &msg[i], 4);
    i += 4;
#endif

    pProvAsd->userNameLen = msg[i++];
    memcpy(&(pProvAsd->userName), &msg[i], pProvAsd->userNameLen);
    i += pProvAsd->userNameLen;

    return i;
}

/*
*****************************************************************************
** FUNCTION:     aprepParseCmsASD
**
** PURPOSE:      parse the ASD pointer for AP Rep from Callagent
**
** PARAMETERS:   pCmsAsd - out - result ASD pointer
**               msg - in - AS Rep message pointer
**
** RETURNS:      none
**
** NOTE:
**
**
*****************************************************************************
*/
int aprepParseCmsASD(CMS_ASD_APREP *pCmsAsd, char *msg)
{
    int i = 4;
#if (CPU_LITTLE_ENDIAN != 0)    
    UINT32 temp;
    memcpy(&temp, msg, i);
    pCmsAsd->inbound_SPI = ntohl(temp);
#else
    memcpy(&(pCmsAsd->inbound_SPI), &msg[0], i);
#endif

    return i;
}



/*
 *****************************************************************************
 ** FUNCTION:    process_ap_rep
 **
 ** PURPOSE:     process an AP_REQ
 **
 ** PARAMETERS:   packet - the received asn1 encoded AP_REQ
 **               request - the decoded request
 **
 ** RETURNS:      success or error code
 **
 ** NOTE:
 **
 **
 *****************************************************************************
 */
krb5_error_code process_ap_rep(
                              krb5_data *reply,
                              krb5_keyblock *auth_keyblock,
                              krb5_keyblock  *subkey)
{
    krb5_ap_rep             *ap_reply;
    krb5_error_code         retval;
    krb5_data               scratch2;
    krb5_ap_rep_enc_part    **repl;
    int                     i;

    if ( !krb5_is_ap_rep(reply) )
    {
        printf("Not an AP_REP message.\n");
        return KRB5KRB_AP_ERR_MSG_TYPE;
    }

    /* It must be a KRB_AS_REP message, or an bad returned packet */
    if ( (retval = decode_krb5_ap_rep(reply, &ap_reply)) )
    {
        /* some other error code ??? */
        printf("AP_reply bad error code %u\n", (unsigned int)retval);
        return 1;
    }

    if ( ap_reply->enc_part.enctype != ENCTYPE_DES3_CBC_MD5 )
    {
        printf("AP Reply:enc_part:type (=%u) not DES3_CBC_MD5.\n",
               (unsigned int)ap_reply->enc_part.enctype);
        return 1;
    }

    scratch2.length = ap_reply->enc_part.ciphertext.length;
    if ( !(scratch2.data = (unsigned char *)krbAllocBuf(scratch2.length)) )
    {
        krb5_free_ap_rep(ap_reply);
        return(ENOMEM);
    }

    if ( (retval = krb5_c_decrypt(auth_keyblock,
                                  KRB5_KEYUSAGE_AP_REP_ENCPART, 0,
                                  &ap_reply->enc_part, &scratch2)) )
    {
        LOG(("AP Reply Decryption failed."));
        goto clean_scratch2;
    }

    /* now decode the decrypted stuff */
    repl = (krb5_ap_rep_enc_part **)krbAllocBuf(sizeof(krb5_ap_rep_enc_part *));
    retval = decode_krb5_ap_rep_enc_part(&scratch2, repl);


    if ( retval )
    {
        LOG(("ERROR: Can not decode AP Reply enc-part."));
        goto clean_scratch2;
    }

    if ( (*repl)->subkey->enctype != -1 )
    {
        printf("AP Reply Subkey type (=%d) is not -1.\n",
               (*repl)->subkey->enctype);
        retval = 1;
        goto clean_scratch2;
    }

    printf("AP_REP decrypted successfully.\n");
    printf("AP_REP ctime = %d, susec=%d, seq_number=%d.\n",
           (*repl)->ctime, (*repl)->cusec, (*repl)->seq_number);

    /* copy and return subkey */
    subkey->length = (*repl)->subkey->length;
    subkey->contents = (unsigned char *)krbAllocBuf(subkey->length);
    memcpy(subkey->contents, (*repl)->subkey->contents, subkey->length);
    subkey->enctype = (*repl)->subkey->enctype;
    printf("subkey is (%d bytes): ", subkey->length);
    for ( i=0; i<subkey->length; i++ )
        printf("%x ", subkey->contents[i]);
    printf("\n");

    clean_scratch2:
    memset(scratch2.data, 0, scratch2.length);
    errout:
    krb5_free_ap_rep(ap_reply);
    krbFreeBuf(scratch2.data);
    return retval;
}

/*
*****************************************************************************
** FUNCTION:     setApreqClientPrincipal
**
** PURPOSE:      set the client principal name and realm in a given AP_REQ
**
** PARAMETERS:   request - request structure
**               krbrealm - kerberos realm
**               mtafqdn - FQDN for the MTA
**
** RETURNS:      success or failure
**
** NOTE:
**
**
*****************************************************************************
*/
extern char gChPrincipalName[32];

int setApreqClientPrincipal(
                           krb5_authenticator *request,
                           char *krbrealm,
                           char *mtafqdn)
{
    //char comp1[] = "PSElement"; //"mta";
    char *comp1 = (char*)gChPrincipalName;

    /* set the realm */
    request->client->magic=KV5M_PRINCIPAL;
    request->client->realm.magic=KV5M_DATA;
    request->client->realm.data = (char *)krbAllocBuf(strlen(krbrealm)+1);
    if ( request->client->realm.data == NULL )
    {
        return ENOMEM;
    }
    memcpy(request->client->realm.data, krbrealm, strlen(krbrealm));
    request->client->realm.length=strlen(krbrealm);
    request->client->realm.data[request->client->realm.length] = '\0';

    /* set the principal names comp1=mta, comp2=fqdn */
    request->client->data = (krb5_data *)krbAllocBuf(2*sizeof(krb5_data));
    if ( request->client->data == NULL )
    {
        return ENOMEM;
    }

    /* set the first component "mta" */
    request->client->data[0].magic=KV5M_DATA;
    request->client->data[0].data = (char *)krbAllocBuf(strlen(comp1)+1);
    if ( request->client->data[0].data == NULL )
    {
        return ENOMEM;
    }
    memcpy(request->client->data[0].data, comp1, strlen(comp1));
    request->client->data[0].length=strlen(comp1);
    request->client->data[0].data[request->client->data[0].length] = '\0';

    /* set the second component <fqdn> */
    request->client->data[1].magic=KV5M_DATA;
    request->client->data[1].data = (char *)krbAllocBuf(strlen(mtafqdn)+1);
    if ( request->client->data[1].data == NULL )
    {
        return ENOMEM;
    }
    memcpy(request->client->data[1].data, mtafqdn, strlen(mtafqdn));
    request->client->data[1].length=strlen(mtafqdn);
    request->client->data[1].data[request->client->data[1].length] = '\0';

    request->client->length= 2;
    //request->client->type = KRB5_NT_SRV_HST;    /* NT-SRV-HST */
    request->client->type = KRB5_NT_SRV_INST;  //trc- per ECR
    return 0;
}

/*
 *****************************************************************************
 ** FUNCTION:    validateSNMPv3Ciphersuites
 **
 ** PURPOSE:     validate AP cipher suites for SNMPv3
 **
 ** PARAMETERS:  enc - list of encryption algs
 **              auth - list of auth algs
 **              len - length of array
 **
 ** RETURNS:      success or error code
 **
 ** NOTE:
 **
 **
 *****************************************************************************
 */
int validateSNMPv3Ciphersuites(int *enc, int *auth, int len)
{
    int i;

    for ( i=0;i<len;i++ )
    {
#ifndef DEMO

        //LOG(("enc is: 0x%.2X, auth is: 0x%.2X, i is: %d, len is: %d", enc[i], auth[i], i, len));
        /* Do not accept noAuth (0x20) if SNMPCompliancy is not disabled */
        if ( !cmNvGetMtaSnmpCompDisable() && auth[i] == SNMPV3_NULL_TID )
        {
            LOG(("ERROR: Authentication must be MD5 or SHA1.  To allow noAuth, turn on non-vol flag SNMPCompDisable."));
            return 1;
        }
#endif
        if ( (enc[i] != SNMPV3_DES_TID) && (enc[i] != SNMPV3_NULL_TID) )
            return 1;
        if ( (auth[i] != SNMPV3_HMAC_MD5_TID) && (auth[i] != SNMPV3_HMAC_SHA1_TID) && 
             (auth[i] != SNMPV3_NULL_TID) )
            return 1;
    }
    return 0;
}


/*
 *****************************************************************************
 ** FUNCTION:    validateIPSECCiphersuites
 **
 ** PURPOSE:     validate AP cipher suites for IPSEC
 **
 ** PARAMETERS:  enc - list of encryption algs
 **              auth - list of auth algs
 **
 ** RETURNS:      success or error code
 **
 ** NOTE:
 **
 **
 *****************************************************************************
 */
int validateIPSECCiphersuites(int *enc, int *auth)
{
    int i;

    for ( i=0;i<MAX_AP_CSUITES_LENGTH;i++ )
    {
        if ( (enc[i]) && (enc[i] != IPSec_ESP_3DES_TID) &&
             (enc[i] != IPSec_ESP_NULL_TID) )
            return 1;
        if ( (auth[i]) && (auth[i] != IPSec_ESP_HMAC_MD5_TID) &&
             (auth[i] != IPSec_ESP_HMAC_SHA1_TID) )
            return 1;
    }
    return 0;
}

/*
 *****************************************************************************
 ** FUNCTION:    sendSARecovered 
 **
 ** PURPOSE:     generate and send SA recoverd message
 **
 ** PARAMETERS:   flag - IPSEC or SNMPv3
 **               ap_rep_msg - AP reply used to send SA recovered
 **               srvc_ip_addr - ip address of server
 **               subkey - 
 **
 ** RETURNS:      
 **
 ** NOTE:
 **
 *****************************************************************************
 */
void sendSARecovered(int flag, 
                     krb5_data          *ap_rep_msg, 
                     unsigned char      *srvc_ip_addr,
                     krb5_keyblock      *subkey)
{
    unsigned char       sa_buf[255];
    krb5_data           sa_packet;
    krb5_data           reply;
    int                 ix=0;
    unsigned long       tmp_len;
    unsigned char       sa_key[SHA1_SIZE];
    unsigned char       sa_hash[SHA1_SIZE];
    krb5_error_code     retval;

    /* key management message ID */
    sa_buf[ix++] = 0x04;

    /* DOI */
    sa_buf[ix++] = flag;

    /* protocol version */
    sa_buf[ix++] = 0x10;

    /* HMAC */
    /* Take SHA-1 of subkey */
    SecLibGetCryptoIface()->secCryptoCalcHash(HASH_SHA1_TYPE, subkey->contents,
                                              subkey->length,
                                              NULL, 0, sa_key, &tmp_len);

    /* Sign entire AP reply */
    SecLibGetCryptoIface()->secCryptoCalcHash(HASH_SHA1_TYPE, ap_rep_msg->data,
                                              ap_rep_msg->length,
                                              sa_key, SHA1_SIZE, sa_hash, &tmp_len);
    memcpy(sa_buf+ix, sa_hash, SHA1_SIZE);
    ix+=SHA1_SIZE;

    sa_packet.data = sa_buf;
    sa_packet.length = ix;
    /* send to call agent or prov server */
    if ( (retval = krb5_sendto_server(&sa_packet, srvc_ip_addr,
                                      AP_UDP_PORT_NUMBER, 30, &reply, KRB_SARECOVERED)) != 0 )
    {
        LOG(("SA recovered not sent. Failed. retval = %d.", retval));
    }
    else
    {
        LOG(("SA Recovered message sent successfully."));
    }
}


/*
 *****************************************************************************
 ** FUNCTION:    process_ap_msg
 **
 ** PURPOSE:     process an AP_REP
 **
 ** PARAMETERS:   packet - the received asn1 encoded AP_REQ
 **               srvc_ip_addr - ip address of server
 **               pAsd - output: ASD pointer
 **               auth_keyblock - the key to be used
 **               app_subkey -
 **               app_flag - type of application: IPSec or SNMPv3
 **
 ** RETURNS:      success or error code
 **
 ** NOTE:
 **
 **
 *****************************************************************************
 */
krb5_error_code process_ap_msg(
                              krb5_data         *ap_msg,
                              char              *srvc_ip_addr,
                              void              *pAsd,
                              krb5_keyblock     *auth_keyblock,
                              krb5_keyblock     *app_subkey,
                              const int         app_flag,
                              int               *authTID,
                              int               *encTID,
                              unsigned long     *splt,
                              unsigned long     *gp)
{
    unsigned char       *msg = ap_msg->data;
    int                 msg_len = ap_msg->length;
    int                 ix = 0;
    krb5_data           ap_packet;
    int                 ap_len = 0;
    krb5_error_code     retval;
    int                 enc[MAX_AP_CSUITES_LENGTH] = {0};
    int                 auth[MAX_AP_CSUITES_LENGTH] = {0};
    int                 i;
    int                 temp;
    int                 cs_len;
    int                 tmp_msg_len;
    unsigned char       sig[SHA1_SIZE];
    unsigned char       sig1[SHA1_SIZE];
    int                 ack_flag;
    unsigned long       tmp_len;
    CMS_ASD_APREP       *pCmsAsd   = NULL;
    PROV_ASD_APREP      *pProvAsd = NULL;
    unsigned char       ap_key[SHA1_SIZE];
    unsigned long       re_establish_flag;

    SecLibGetCryptoIface()->secCryptoCalcHash(HASH_SHA1_TYPE, auth_keyblock->contents,
                                              auth_keyblock->length,
                                              NULL, 0,ap_key,&tmp_len);


    /* Sign the message */
    /* call SecLib for HMAC_SHA1 */
    SecLibGetCryptoIface()->secCryptoCalcHash(HASH_SHA1_TYPE, msg, msg_len-SHA1_SIZE,
                                              ap_key, SHA1_SIZE, sig1, &tmp_len);



    tmp_msg_len = msg_len - SHA1_SIZE;

    memcpy(sig, &msg[tmp_msg_len],SHA1_SIZE);
    if ( memcmp(sig,sig1,SHA1_SIZE) )
    {
        printf("SIGNATURE Did not Verify.\n");
        return 1;
    }

    printf("AP_REP HMAC SIGANTURE VERIFIED.\n");

    if ( msg[ix++] != 0x03 )
        printf("Bad msg ID.\n");

    if ( msg[ix++] != app_flag )
        printf("Bad DOI.\n");

    if ( msg[ix++] != 0x10 )
        printf("Bad protocol version.\n");

    if ( msg[ix++] != 0x6F )
        printf("Bad AP_REQ.\n");

    /* Copy the first 10 bytes or so, just to decode the length */
    ap_packet.length = 10;
    ap_packet.data = krbAllocBuf(10);
    memcpy(ap_packet.data, &(msg[ix-1]), 10);
    if ( retval = decode_krb5_ap_rep_len(&ap_packet, &ap_len) )
        return retval;
    printf(".k5  process_ap_msg: after decode_krb5_ap_rep_len\n");

    /* Add the length bytes to this,
     * if length < 127 it is 6F XX
     * if  between 127 and 255 it is 6F 81 XX
     * if greater than 255 it is 6F 82 XX XX
     */
    if ( ap_len <= 127 )
        ap_len += 2;
    else if ( (ap_len > 127) && (ap_len <= 255 ) )
        ap_len += 3;
    else if ( ap_len > 255 )
        ap_len += 4;
    else
    {
        printf("Bad AP_REP part length %d.\n", ap_len);
        return 1;
    }

    ap_packet.length = ap_len;
    ap_packet.data = krbReallocBuf(ap_packet.data, 10, ap_len);
    memcpy(ap_packet.data, &(msg[ix-1]), ap_len);
    printf(".k5  process_ap_msg: calling process_ap_rep\n");

    retval = process_ap_rep(&ap_packet, auth_keyblock, app_subkey);
    if ( retval )
        return 1;
    printf(".k5  process_ap_msg: BACK FROM calling process_ap_rep\n");

    krbFreeBuf(ap_packet.data);

    ix += ap_len-1;

    /* ASD (SPI for IPSEC) */
    if ( app_flag == APPLICATION_PROV )
    {
        pProvAsd = (PROV_ASD_APREP*)pAsd;
        ix += aprepParseProvASD(pProvAsd, &msg[ix]);
    }
    else
    {
        pCmsAsd = (CMS_ASD_APREP*)pAsd;
        ix += aprepParseCmsASD(pCmsAsd, &msg[ix]);
    }

    /* length of ciphersuit list */
    cs_len = msg[ix++];

    /* list of ciphersuites */
    for ( i=0;i<cs_len;i++ )
    {
        auth[i] = msg[ix++];
        enc[i] = msg[ix++];
    }
    printf(".k5  process_ap_msg: calling validateSNMPv3Ciphersuites\n");

    /* for now, just return the first ciphersuites in the list */
    if ( app_flag == APPLICATION_PROV )
    {
        if ( validateSNMPv3Ciphersuites(enc, auth, cs_len) != 0 )
        {
            printf("SNMPV3 AP ciphersuites not valid.\n");
            return 1;
        }
    }
    else
    {
        if ( validateIPSECCiphersuites(enc, auth) != 0 )
        {
            printf("IPSEC AP ciphersuites not valid.\n");
            return 1;
        }
    }

    *authTID = auth[0];
    *encTID = enc[0];

#if (CPU_LITTLE_ENDIAN != 0)
    temp = 0;
    memcpy(&temp, &msg[ix], 4);
    *splt = ntohl(temp);
    ix += 4;
    memcpy(&temp, &msg[ix], 4);
    *gp = ntohl(temp);
    ix += 4;
#else
    memcpy(splt, &msg[ix], 4);
    ix += 4;
    memcpy(gp, &msg[ix], 4);
    ix += 4;
#endif

    printf("Security Parameter Lifetime = %d seconds.\n", *splt);
    printf("Grace Period = %d seconds.\n", *gp);

    re_establish_flag = msg[ix++];
    printf("Re-establish Flag 0x%x.\n", re_establish_flag);

    /* ACK Required Flag */
    ack_flag = msg[ix++];
    if ( ack_flag == 1 )
    {
        printf("ACK Required flag is set, send SA Recovered.\n");
        sendSARecovered(app_flag, ap_msg, srvc_ip_addr, app_subkey);
    }

    printf("\nAP exchange successful. Key Received.\n");

    return 0;
}

/*
 *****************************************************************************
 ** FUNCTION:    process_error_reply
 **
 ** PURPOSE:     process an KRB_ERROR
 **
 ** PARAMETERS:   err_rep - the received asn1 encoded KRB_ERROR message
 **               session_key - session key from the ticket for checksum verification
 **               seq_num - sequence number from authenticator
 **               app_flag - type of application: IPSec or SNMPv3
 **               srvcfqdn - fqdn of service
 **
 ** RETURNS:      success or error code
 **
 ** NOTE:
 **
 **
 *****************************************************************************
 */
krb5_error_code process_error_reply(
                                   krb5_data      *err_rep,
                                   krb5_keyblock  *session_key,
                                   krb5_int32     seq_num,
                                   const int      app_flag,
                                   char           *srvcfqdn)
{
    int                retval;
    int                ix = 0;
    krb5_error         err_reply;
    krb5_error         *perr_reply = &err_reply;
    unsigned char      *msg = err_rep->data;
    krb5_data          err_packet;
    int                err_len = 0;
    int                valid;
    krb5_data          *err_ptr;
    krb5_checksum      cksum;
    krb5_checksum      *pcksum;
    krb5_typed_data    **e_data;
    int                in_seq;
    int                clock_skew;
    int                orig_clock_skew;
    int                max_clock_skew_from_mib;
    int                tmp;
    krb5_data          scratch;


    /* parse Packet Cable fields */
    if ( msg[ix++] != 0x06 )
        printf("Bad msg ID.\n");

    if ( msg[ix++] != 0x10 )
        printf("Bad protocol version.\n");

    if ( msg[ix++] != app_flag )
        printf("Bad DOI.\n");

    if ( msg[ix++] != 0x7E )
        printf("Bad KRB_ERROR.\n");

    /* Copy the first 10 bytes or so, just to decode the length */
    err_packet.length = 10;
    err_packet.data = krbAllocBuf(10);
    memcpy(err_packet.data, &(msg[ix-1]), 10);
    if ( retval = decode_krb5_ap_rep_len(&err_packet, &err_len) )
    {
        goto cleanup;
    }

    /* Add the length bytes to this,
     * if length < 127 it is 6F XX
     * if between 127 and 255 it is 6F 81 XX
     * if greater than 255 it is 6F 82 XX XX
     */
    if ( err_len <= 127 )
        err_len += 2;
    else if ( (err_len > 127) && (err_len <= 255 ) )
        err_len += 3;
    else if ( err_len > 255 )
        err_len += 4;
    else
    {
        printf("Bad KRB_ERR part length %d.\n", err_len);
        return 1;
    }

    err_packet.length = err_len;
    err_packet.data = krbReallocBuf(err_packet.data, 10, err_len);
    memcpy(err_packet.data, &(msg[ix-1]), err_len);

    if ( !krb5_is_krb_error(&err_packet) )
    {
        printf("Not an AP_REP message.\n");
        retval = KRB5KRB_AP_ERR_MSG_TYPE;
        goto cleanup;
    }

    if ( (retval = decode_krb5_error(&err_packet, &perr_reply)) )
    {
        /* some other error code--??? */
        printf("err_reply bad error code %d\n", perr_reply->error);
        goto cleanup;
    }

    /* verify error message is of type REQ-SEQ and sequence number matches */
    if ( ( retval =
           decode_krb5_typeddata_sequence(&(perr_reply->e_data), &e_data)) )
    {
        printf("KRB-ERROR typeddata can not be decoded.\n");
        goto cleanup;
    }

    scratch.data = (*e_data)->contents;
    scratch.length = (*e_data)->length;
    /* Now decode the req-seq inside e-data */
    if ( ( retval = decode_krb5_typeddata_req_seq(&scratch, &in_seq)) )
    {
        printf("KRB-ERROR REQ-SEQ can not be decoded.\n");
        goto cleanup;
    }

    if ( ((*e_data)->data_type != KRB5_TD_REQ_SEQ) &&
         ((*e_data)->data_type != KRB5_TD_APP_DEFINED_ERROR) ||
         (in_seq != seq_num) )
    {
        printf("invalid KRB_ERROR reply.\n");
        retval =  KRB5KRB_AP_ERR_BAD_INTEGRITY;
        goto cleanup;
    }

    /* verify checksum */
    cksum.length = perr_reply->e_cksum.length;
    cksum.checksum_type = perr_reply->e_cksum.checksum_type;
    cksum.magic = perr_reply->e_cksum.magic;
    cksum.contents = krbAllocBuf(cksum.length);
    if ( cksum.contents )
    {
        memcpy(cksum.contents, perr_reply->e_cksum.contents, cksum.length);
    }
    pcksum = &cksum;

    /* first re-encode the error packet to pass into the checksum routine */
    perr_reply->e_cksum.length = 0;
    err_ptr = &err_packet;
    if ( retval = encode_krb5_error(perr_reply, &err_ptr) )
    {
        goto cleanup;
    }

    if ( retval = krb5_c_verify_checksum(session_key, 0, err_ptr,
                                         pcksum, &valid) )
    {
        printf("ERROR: Failed to verify checksum\n");
        goto cleanup;
    }

    if ( !valid )
    {
        printf("ERROR: krb-error checksum do not verify\n");
        goto cleanup;
    }

    if ( perr_reply )
    {
        retval = perr_reply->error - ERROR_TABLE_BASE_krb5;
        printf("\n*** Receive KRB-ERROR: AP_reply valid err_reply %d\n", retval);
        if ( perr_reply->text.length > 0 )
        {
            printf("Text: %s.\n", perr_reply->text.data);
        }

        /* If time skew, write NVRAM */
        if ( perr_reply->error == KRB5KRB_AP_ERR_SKEW )
        {
            clock_skew = perr_reply->ctime - perr_reply->stime;
            printf("Clock Skew error ctime=%d, stime=%d, skew=%d.\n",
                   perr_reply->ctime, perr_reply->stime, clock_skew);
#if 0

#ifndef DEMO
            if ( app_flag == APPLICATION_CMS )
            {
                if ( mibAccessSec(
                                 MIBITEM(SEC, cmsMaxClockSkew),
                                 NULL,
                                 srvcfqdn,
                                 NULL,
                                 &tmp,
                                 (void*)&max_clock_skew_from_mib,
                                 MIB_ITEM_GET) )
                {
                    LOG(("ERROR: MIB GET cmsMaxClockSkew failed for fqdn %s, setting to default %d seconds.", srvcfqdn, MAX_ALLOWED_CLOCK_SKEW_IN_SECONDS));  
                    max_clock_skew_from_mib = MAX_ALLOWED_CLOCK_SKEW_IN_SECONDS;
                }
                LOG(("CMS Clock skew from MIB for %s is %d.", 
                     srvcfqdn, max_clock_skew_from_mib));
            }
            else
            {
                /* For PROV set the max skew to 1 hr */
                max_clock_skew_from_mib = MAX_ALLOWED_CLOCK_SKEW_IN_SECONDS;
            }
#else
            /* for windows always set max skew to 1 hr */
            max_clock_skew_from_mib = MAX_ALLOWED_CLOCK_SKEW_IN_SECONDS;
#endif

#ifndef DEMO
            if ( ! cmNvIsStandaloneEnabled() )
            {
#endif

                /* Note: this is the real code, but for now ... */
                if ( labs(clock_skew) > max_clock_skew_from_mib )
                {
                    printf("AP SKEW GREATER THAN MAXIMUM ALLOWED (1 HR).\n");
                    retval = KRB5_KDCREP_SKEW;
                    goto cleanup;
                }
                else
                {
                    /* Write to NVRAM */
                    if ( app_flag == APPLICATION_CMS )
                    {
                        /* add to the original skew */
                        orig_clock_skew =  ch_secReadNVRAMSkew(AP_CMS_FLAG, 
                                                               srvcfqdn, 
                                                               strlen(srvcfqdn));
                        clock_skew += orig_clock_skew;

                        ch_secWriteNVRAMSkew(clock_skew, AP_CMS_FLAG, srvcfqdn, strlen(srvcfqdn));
                    }
                    else  /* bad flag has already been taken care of */
                    {
                        /* add to the original skew */
                        orig_clock_skew =  ch_secReadNVRAMSkew(AP_PROV_FLAG, 
                                                               srvcfqdn, 
                                                               strlen(srvcfqdn));
                        clock_skew += orig_clock_skew;

                        ch_secWriteNVRAMSkew(clock_skew, AP_PROV_FLAG, srvcfqdn, strlen(srvcfqdn));
                    }
                }
#endif //if 0

#ifndef DEMO
            }
            else       /* Standalone MTA */
            {
#endif            
                if ( labs(clock_skew) > max_clock_skew_from_mib )
                {
                    printf("AP SKEW GREATER THAN MAXIMUM ALLOWED (1 hr), Ignored for Standalone.\n");
                }
                /* Write to NVRAM */
                if ( app_flag == APPLICATION_CMS )
                {
                    /* add to the original skew */
                    orig_clock_skew =  ch_secReadNVRAMSkew(AP_CMS_FLAG,srvcfqdn,strlen(srvcfqdn));
                    clock_skew += orig_clock_skew;

                    ch_secWriteNVRAMSkew(clock_skew, AP_CMS_FLAG, srvcfqdn, strlen(srvcfqdn));
                }
                else  /* bad flag has already been taken care of */
                {
                    /* add to the original skew */
                    orig_clock_skew =  ch_secReadNVRAMSkew(AP_PROV_FLAG,srvcfqdn,strlen(srvcfqdn));
                    clock_skew += orig_clock_skew;

                    ch_secWriteNVRAMSkew(clock_skew, AP_PROV_FLAG, srvcfqdn, strlen(srvcfqdn));
                }
//         }
                goto cleanup;
//#endif
            }
        }

        cleanup:
        if ( err_packet.length )
            krbFreeBuf((char*) err_packet.data);
        if ( cksum.length )
            krbFreeBuf((char*) cksum.contents);
        if ( perr_reply->error != 0 )
            krbFreeBuf((char*) perr_reply);

        return retval;
    }


/*
*****************************************************************************
** FUNCTION:     getAPExpAlgInfo
**
** PURPOSE:      get necessary information for exp backoff algorithm
**
** PARAMETERS:   app_flag - ipsec or snmpv3 flag
**               srvcfqdn - fqdn of service 
**               max_retries - out - 
**               nom_timeout - out - 
**
** RETURNS:      success or failure
**
** NOTE:
**
**
*****************************************************************************
*/
    void getAPExpAlgInfo(
                        int              app_flag,
                        unsigned char    *fqdn,
                        int              *max_retries,
                        int              *nom_timeout)
    {
        int         tmp;
        int         mt;
        int         nt;

        switch ( app_flag )
        {
        case APPLICATION_PROV:

#ifdef DEMO    
            *max_retries = 2; //4
            *nom_timeout = 10;
#else
            if ( mibAccessSec(
                             MIBITEM(SEC, provUnsolicitedKeyMaxRetries),
                             NULL,
                             NULL,
                             NULL,
                             &tmp,
                             (void*)&mt,
                             MIB_ITEM_GET) )
            {
                LOG(("ERROR: MIB GET provUnsolicitedKeyMaxRetries Failed, set to default 4."));
            }
            if ( mibAccessSec(
                             MIBITEM(SEC, provUnsolicitedKeyNomTO),
                             NULL,
                             NULL,
                             NULL,
                             &tmp,
                             (void*)&nt,
                             MIB_ITEM_GET) )
            {
                LOG(("ERROR: MIB GET provUnsolicitedKeyNomTO Failed, set to default 10."));
            }
#endif
            break;
        case APPLICATION_CMS:
#ifdef DEMO
            *max_retries = 2; //4
            *nom_timeout = 10;
#else
            if ( mibAccessSec(
                             MIBITEM(SEC, cmsMaxUnsolicitedKeyMaxRetries),
                             NULL,
                             fqdn,
                             NULL,
                             &tmp,
                             (void*)&mt,
                             MIB_ITEM_GET) )
            {
                LOG(("ERROR: MIB GET cmsMaxUnsolicitedKeyMaxRetries for fqdn %s Failed, set to default 10.", fqdn));
            }
            if ( mibAccessSec(
                             MIBITEM(SEC, cmsMaxUnsolicitedKeyNomTimeOut),
                             NULL,
                             fqdn,
                             NULL,
                             &tmp,
                             (void*)&nt,
                             MIB_ITEM_GET) )
            {
                LOG(("ERROR: MIB GET cmsMaxUnsolicitedKeyNomTimeOut for fqdn %s Failed, set to default 10.", fqdn));
            }
#endif
            break;
        default:
            LOG(("getAPExpAlgInfo BAD FLAG %d.", app_flag));
            break;
        }

        //LOG(("MIB VALUES mt = %d, nt = %d.", mt, nt));

        *max_retries = mt;
        *nom_timeout = nt;

        /* defaults if MIBS are empty */
        if ( *max_retries <= 0 )
            *max_retries = 2;//4;
        if ( *nom_timeout <= 0 )
            *nom_timeout = 10;

        //LOG(("VALUES set to max retries = %d, nom timeout = %d.", 
        //     *max_retries, *nom_timeout));
    }


/*
*****************************************************************************
** FUNCTION:     send_ap_req
**
** PURPOSE:      sends AP_REQ, receive AP_REP, processes and stores subkey
**
** PARAMETERS:   mtafqdn - MTA FQDN
**               srvcfqdn - fqdn of service ( used for CMS cases only)
**               srvc_ip_addr - cms or prov server fqdn
**               krbrealm - kerberos realm
**               krbto - time out for udp
**               app_flag - ipsec or snmpv3 flag
**               server_nonce - in - nonce from wakeup or rekey (0 otherwise)
**               pAppData - in - ASD pointer AP Req
**               pAppRep  - out - ASD pointer for AP Rep
**               credentails - credentials from AS exchange
**               app_key - out - shared key information returned
**               authTID - out - transform ID for authentication
**               encTID - out - transform ID for encryption
**               reason - in - wakeup, rekey, time expiry, etc.
**
** RETURNS:      success or failure
**
** NOTE:
**
**
*****************************************************************************
*/
    krb5_error_code send_ap_req(
                               const char          *mtafqdn,
                               const char          *srvcfqdn,
                               const char          *srvc_ip_addr,
                               const char          *krbrealm,
                               const int           krbto,
                               const int           app_flag,
                               const int           server_nonce,
                               void                *pAppData,
                               void                *pAppRep,
                               MtaCredentials      *credentials,
                               MtaApplicationKey   *app_key,
                               int                 *authTID,
                               int                 *encTID,
                               unsigned long       reason)
    {
        krb5_data           tkt;
        krb5_ap_req         *request;
        krb5_authenticator  *auth;
        krb5_keyblock       auth_keyblock;
        krb5_data           *scratch;
        krb5_error_code     retval;
        krb5_data           *toutbuf;
        krb5_octet          ap_msg[1024];
        int                 ap_msg_len = 0;
        krb5_data           reply;
        //krb5_error          *err_reply=0;
        krb5_flags          ap_req_options;
        krb5_data             ap_packet;
        krb5_data           hmac_key;
        krb5_keyblock       app_subkey;
        //int                 tmp;
        int                 i;
        int                 clock_skew;
        time_t              tmp_time;
        unsigned long       splt;
        unsigned long       gp;
        int                 max_retries;
        int                 nom_timeout;
        int                 retries = 0;
        int                 tafter;
        int                 tbefore;
        int                 rtt;
        int                 utc_time_offset;
        int                 exp_flag;

        /* time related declarations */
        //OS_TMDATE   idate;
        //OS_TMTIME   itime;
        OS_TICKS    iticks;
        //OS_STATUS   istat;

//#ifdef DEMO
//#if 1
//   krb5_octet   treply[5120];
//   krb5_octet   treq[5120];
//   int          flen;
//   FILE *alopa_ap_rep_bin;
//   FILE *alopa_session_key_bin;
//   FILE *cisco_ap_rep_bin;
//   FILE *my_ap_req_bin;
//   FILE *brcm_ap_req_bin, csco_ap_req_bin;
//#endif
//printf(".k  send_ap_req() -start but we are in DEMO\n");

/*#if 1      -trc-
   if (!ap_fopened_once1)
   {
      brcm_ap_req_bin = fopen("ap_req.bin", "wb");
   }
#endif

#if 0
   if (!ap_fopened_once2)
   {
      alopa_ap_rep_bin = fopen("alopa_ap_rep.bin","rb");
   }
#endif

#if 0
   if (!ap_fopened_once3)
   {
      cisco_ap_rep_bin = fopen("aprep.bin","rb");
   }
#endif
#if 0
   if (!ap_fopened_once4)
   {
      my_ap_req_bin = fopen("my_apreq.bin","rb");
   }
#endif
#if 0
   if (!ap_fopened_once5)
   {
      alopa_session_key_bin = fopen("session_key.out","rb");
   }
#endif
    */
//#endif   /* DEMO */
//   printf(".k  send_ap_req() -start now out of in DEMO\n");

        hmac_key.data = 0;
        ap_packet.data = 0;
        app_subkey.length = 0;

        request = krbAllocBuf(sizeof(krb5_ap_req));
        request->ap_options = ap_req_options & AP_OPTS_WIRE_MASK;

        /* get the ticket out of cerdentials and decode */
        if ( credentials->tkt.tkt_len == 0 )
            return(KRB5_NO_TKT_SUPPLIED);

        tkt.length = credentials->tkt.tkt_len;
        tkt.data = krbAllocBuf(tkt.length);
        memcpy(tkt.data, credentials->tkt.tkt_val, tkt.length);
        //printf("STOP FOR DEBUG 10 -ok here must be after! ! !\n");


        if ( (retval = decode_krb5_ticket(&tkt, &(request->ticket))) )
        {
            printf("decode_krb5_ticket ret %d\n",retval);
            krb5_free_ap_req(request);
            return(retval);
        }
        KDEBUG("okok decode_krb5_ticket ret %d\n",retval);

        /* verify that the ticket is not expired */
        /* if ((retval = krb5_validate_times(&in_creds->times)) != 0)
           goto cleanup; */
//printf("STOP FOR DEBUG 7 - goto cleanup here crashes us! ! !\n");

//printf("STOP FOR DEBUG 9 - crashes here! ! !\n");


        /* construct an authenticator */
        auth = (krb5_authenticator *)krbAllocBuf(sizeof(krb5_authenticator));
        auth->client = (krb5_principal)krbAllocBuf(sizeof(krb5_principal_data));
        auth->magic = KV5M_AUTHENTICATOR;
        setApreqClientPrincipal(auth, krbrealm, mtafqdn);

        /* if this is in response to a rekey message, generate the subkey and
         * put in AP-REQ
         */
        if ( reason == KRB_REKEY )
        {
            auth->subkey = (krb5_keyblock *)krbAllocBuf(sizeof(krb5_keyblock));
            auth->subkey->enctype = -1;
            auth->subkey->length = SUBKEY_LENGTH;
            auth->subkey->contents = (unsigned char *)krbAllocBuf(SUBKEY_LENGTH);
            for ( i=0;i<SUBKEY_LENGTH;i++ )
                auth->subkey->contents[i] = rand();

            printf("SubKey: ");
            for ( i=0;i<SUBKEY_LENGTH;i++ )
                printf("%x ", auth->subkey->contents[i]);
            printf("\n");

            /* also no need to set the MUTUAL_REQUIRED flag */
            ap_req_options = 0;
        }
        else
        {
            auth->subkey = 0;
            ap_req_options = AP_OPTS_MUTUAL_REQUIRED;
        }
//printf("STOP FOR DEBUG 8 - no crash but no command bad here! ! !\n");


        /* get current time */
        //istat = bcmOsGetTimeEx(&idate, &itime, &iticks);
        //if ( OS_STATUS_OK != istat )
        //{
        //    //printf("Could not get the time. error code is %d\n", istat);
        //    retval = istat;
        //    //trc-    goto cleanup;
        //    if ( ktracelevel > KTRACE_INFO )
        //        printf(".k trc- setting iticks to zero for AP\n");
        //    iticks=0;
        //}

         iticks=0;
         ///////////////

         

        if ( app_flag == APPLICATION_CMS )
        {
            clock_skew =  ch_secReadNVRAMSkew(AP_CMS_FLAG, srvcfqdn, strlen(srvcfqdn));
        }
        else
        {
            clock_skew =  ch_secReadNVRAMSkew(AP_PROV_FLAG, srvcfqdn, strlen(srvcfqdn));
        }
        LIBDEBUG("AP Clock Skew for fqdn %s = %d.\n", srvcfqdn, clock_skew);

        auth->cusec = (OS_MSEC_PER_TICK*iticks*1000)%1000;
        time(&tmp_time);

#ifdef DEMO
        utc_time_offset = 0;
#else
        utc_time_offset = cmMiscGetUTCTimeOffset();
#endif
        LOG(("CM UTC Time Offset is %d.", utc_time_offset));
        tmp_time -= utc_time_offset;

        auth->ctime = tmp_time - clock_skew;
        //printf("AP Authenticator time is %d.\n", auth->ctime);
        auth->checksum = 0;
        auth->seq_number = rand();
        auth->authorization_data = 0;

        printf("\nSending AP_REQ to ProvServer %s.\n", srvc_ip_addr);
        LIBDEBUG("AP_REQ ctime = %d, susec=%d, seq_number=%d.\n",auth->ctime, auth->cusec, auth->seq_number);

        /* encode the authenticator */
        if ( (retval = encode_krb5_authenticator(auth, &scratch)) )
        {
            LIBDEBUG("failed to encode AP authenticator.\n");
            goto cleanup;
        }
        LIBDEBUG(".k back from encode_krb5_authenticator\n");
        /* set the key to encrypt the authenticator */
        auth_keyblock.enctype = ENCTYPE_DES3_CBC_MD5;
        auth_keyblock.length = credentials->session_key.session_key_len;

        auth_keyblock.contents = (unsigned char *)krbAllocBuf(auth_keyblock.length);
        memcpy(auth_keyblock.contents, credentials->session_key.session_key_val,
               auth_keyblock.length);

        //printf(".k calling krb5_encrypt_helper\n");

        /* call the encryption routine */
        if ( (retval = krb5_encrypt_helper(&auth_keyblock,
                                           KRB5_KEYUSAGE_AP_REQ_AUTH,
                                           scratch, &(request->authenticator))) )
        {
            printf("Failed to encrypt AP authenticator retval = %d.\n", retval);
            goto cleanup;
        }
        LIBDEBUG(".k calling encode_krb5_ap_req\n");

        if ( (retval = encode_krb5_ap_req(request, &toutbuf)) )
        {
            printf("Failed to encode AP request.\n");
            goto cleanup;
        }

        /* Add packet cable fields */
        hmac_key.length = auth_keyblock.length;
        hmac_key.data = (unsigned char *)krbAllocBuf(auth_keyblock.length);
        memcpy(hmac_key.data, auth_keyblock.contents, hmac_key.length);

        if ( reason == !KRB_REKEY )
        {
            make_pc_ap_req(toutbuf, &hmac_key, pAppData, ap_msg,
                           &ap_msg_len, app_flag, server_nonce, 0, 0);
        }
        else
        {
            make_pc_ap_req(toutbuf, &hmac_key, pAppData, ap_msg,
                           &ap_msg_len, app_flag, server_nonce, *authTID, *encTID);
        }

        krbFreeBuf(toutbuf->data);
        krbFreeBuf(toutbuf);

        krbFreeBuf(hmac_key.data);

        ap_packet.length = ap_msg_len;
        ap_packet.data = (krb5_octet *)krbAllocBuf(ap_msg_len);
        memcpy(ap_packet.data, ap_msg, ap_msg_len);
        //printf(".k whew did we get here?\n");


#if 0 //DEMO
#if 1

        /* Write our ap req to a file */
        if ( !ap_fopened_once1 )
        {
            fwrite(ap_packet.data, sizeof(char), ap_msg_len, brcm_ap_req_bin);
            fclose(brcm_ap_req_bin);
            ap_fopened_once1 = 1;
        }
#endif
#endif

#if 0// DEMO
#if 0
        /* read ap req from a file and send */
        if ( !ap_fopened_once4 )
        {
            flen = fread(treq, 1, 5120, my_ap_req_bin);
            fclose(my_ap_req_bin);
            ap_packet.length = flen;
            ap_packet.data = treq;
            ap_fopened_once4 = 1;
        }
#endif
#endif


#if 1
        //LOG(("Getting Max Retries and Nom Timeout from MIB."));
        getAPExpAlgInfo(app_flag, srvcfqdn, &max_retries, &nom_timeout);
        if ( app_flag == APPLICATION_CMS )
        {
            exp_flag = AP_CMS_FLAG;
        }
        else
        {
            exp_flag = AP_PROV_FLAG;
        }

        //trc- hack
        //max_retries = 0;
        //printf("Set max_retries to 0\n");
        max_retries = gMaxRetries;
        nom_timeout = gTimeout;
        //nom_timeout = 20;
        //printf("max_retries = <%d>, nom_timeout = <%d>\n",max_retries, nom_timeout);

        while ( retries <= max_retries )
        {
            time(&tbefore);           

            /* send to call agent or prov server */

            LIBDEBUG("SENDING AP REQUEST", (unsigned char *) ap_packet.data, ap_packet.length, 'x');  //just show a few


            if ( (retval = krb5_sendto_server(&ap_packet, srvc_ip_addr,
                                              AP_UDP_PORT_NUMBER, nom_timeout, &reply, reason)) != 0 )
            {
                time(&tafter);
                rtt = tafter - tbefore;
                retries++;
                LOG(("Elapsed time is %d.", rtt));
                printf("AP REQ timed out, retval=%d, retries=%d\n", retval, retries);
                if ( retries > max_retries )
                {
                    LOG(("Reached max number of retries."));
                    secUpdateExpBackoffRetryTable(
                                                 exp_flag,
                                                 rtt,
                                                 1,
                                                 0,   /* endpoint ? */
                                                 srvcfqdn);
                    goto cleanup;
                }
                else
                {
                    secUpdateExpBackoffRetryTable(
                                                 exp_flag,
                                                 rtt,
                                                 1,
                                                 0,   /* endpoint ? */
                                                 srvcfqdn);
                }
            }
            else
            {
                break;
            }
            nom_timeout = secGetRTT(exp_flag, srvcfqdn);
        }
#endif
        time(&tafter);
        rtt = tafter - tbefore;
        secUpdateExpBackoffRetryTable(
                                     exp_flag,
                                     rtt,
                                     0,
                                     0,   /* endpoint ? */
                                     srvcfqdn);



        krbFreeBuf(ap_packet.data);

        /* if this is a response to a rekey, there is no reply */
        if ( reason == KRB_REKEY )
        {
            printf("REKEY AP-REQ SENT.\n");
            retval = 0;
            /* return the subkey for key derivation */
            app_key->sub_key.sub_key_type = auth->subkey->enctype;
            app_key->sub_key.sub_key_len = auth->subkey->length;
            app_key->sub_key.sub_key_val = malloc(auth->subkey->length);
            memcpy(app_key->sub_key.sub_key_val, auth->subkey->contents,
                   app_key->sub_key.sub_key_len);

            krbFreeBuf(auth->subkey->contents);
            krbFreeBuf(auth->subkey);
            goto cleanup;
        }

#ifdef DEMO
#if 0
        if ( !ap_fopened_once2 )
        {
            flen = fread(treply, 1, 5120, alopa_ap_rep_bin);
            fclose(alopa_ap_rep_bin);
            reply.length = flen;
            reply.data = treply;
            ap_fopened_once2 = 1;
        }
#endif
#endif


#ifdef DEMO
#if 0
        if ( !ap_fopened_once3 )
        {
            flen = fread(treply, 1, 5120, cisco_ap_rep_bin);
            fclose(cisco_ap_rep_bin);
            reply.length = flen;
            reply.data = treply;
            ap_fopened_once3 = 1;
        }
#endif
#endif

#ifdef DEMO
#if 0
        if ( !ap_fopened_once5 )
        {
            auth_keyblock.length = fread(auth_keyblock.contents, 1, 5120,
                                         alopa_session_key_bin);
            fclose(alopa_session_key_bin);
            ap_fopened_once5 = 1;
        }
#endif
#endif




        /* now decode the reply...could be error or as_rep */
        if ( *reply.data == 0x06 )
        {
            retval = process_error_reply(&reply, &auth_keyblock, auth->seq_number, app_flag, srvcfqdn);
            goto cleanup;
        }

        if ( (retval = process_ap_msg(&reply, srvc_ip_addr, pAppRep, &auth_keyblock,
                                      &app_subkey, app_flag, authTID, encTID, &splt, &gp)) )
        {
            printf("ERROR AP_REPLY Falied.\n");
            goto cleanup;
        }

        app_key->sub_key.sub_key_type = app_subkey.enctype;
        app_key->sub_key.sub_key_len = app_subkey.length;
        app_key->sub_key.sub_key_val = malloc(app_subkey.length);
        memcpy(app_key->sub_key.sub_key_val, app_subkey.contents, app_subkey.length);
        app_key->sec_params_lifetime = splt;
        app_key->grace_period = gp;

        cleanup:
        /* clean up stuff here */
        LIBDEBUG("IN CLEANUP - \n");
        //return(retval);
        krbFreeBuf(tkt.data);
        krbFreeBuf(auth->client->realm.data);
        krbFreeBuf(auth->client->data[0].data);
        krbFreeBuf(auth->client->data[1].data);
        krbFreeBuf(auth->client->data);
        krbFreeBuf(auth->client);
        krbFreeBuf(auth);
        krbFreeBuf(scratch->data);
        krbFreeBuf(scratch);
        krb5_free_ap_req(request);
        if ( reply.length != 0 )
            krbFreeBuf(reply.data);
        if ( app_subkey.length != 0 )
            krbFreeBuf(app_subkey.contents);
        krbFreeBuf(auth_keyblock.contents);
        return(retval);
    }


/*
******************************************************************************
** FUNCTION:   secGetKrbApplicationKey
**
** PURPOSE:    Send a kerberos AP_REQ,
**             recv a reply and process, and provide shared key or error code
**
**   PARAMETERS:
**    1. mtafqdn - in - MTA fqdn
**    2. srvcfqdn - in - fqdn of PS or CA
**    3. srvc_ip_addr - in - Call Agent or provisioning server fqdn
**    4. krbrealm - kerberos realm
**    5. krbto - maximum timeout for completion
**    6. app_flag - in - flag (CMS or PROV)
**    7. server_nonce - in - 0 or value if response to a wake-up
**    8. pAppData - in - snmp information to be place in the message
**    9. pAppRep - out - snmp information be returned
**    10. credentials - in - credentials (ticket, session key, etc.)
**    11. application_key - struct containing Shared key between MTA and
**          application server (call agent or prov server), times, etc.
**    12. authTID - out - authentication transform ID result of ciphersuite
**          negotiation
**    13. encTID - out - encryption transform ID result of ciphersuite
**          negotiation
**    14. reason - in - bootup, wakeup, rekey, etc.
**
**
** RETURNS:   Success or error code - krb5_error_code
**
** NOTE:
**     1. Caller responsible for freeing application_key
**
******************************************************************************
*/
    UINT32 secGetKrbApplicationKey(
                                  const char             *mtafqdn,
                                  const char             *srvcfqdn,
                                  const char             *srvc_ip_addr,
                                  const char             *krbrealm,
                                  const int              krbto,
                                  App_flag               app_flag,
                                  const int              server_nonce,
                                  void                   *pAppData,
                                  void                   *pAppRep,
                                  MtaCredentials         *credentials,
                                  MtaApplicationKey      *application_key,
                                  int                    *authTID,
                                  int                    *encTID,
                                  unsigned long          reason)
    {
        krb5_error_code retval;
        krb5_error_code    rvalue;
        unsigned char   skey[SUBKEY_LENGTH];
        unsigned long   splt;
        unsigned long   gp;
        int             stat;

        LIBDEBUG(".k  secGetKrbApplicationKey(): -start\n");

        retval = secRetreiveSubkeyInfo(skey, srvcfqdn, &splt, &gp, &stat);
        if ( reason != KRB_SUBKEY_EXPIRED )
        {
            /* If server_nonce is non-zero, this is a message in response to
             * Wakeup or Rekey. Even if there is valid subkey, we need to
             * get a new one
             */
            if ( ! server_nonce )
            {
                if ( retval )
                {
                    LIBDEBUG("Found Valid Subkey for service %s.\n", srvcfqdn);
                    return retval;
                }
            }
        }
        LIBDEBUG(".k  secGetKrbApplicationKey(): -calling send_ap_req\n");
        

        retval = send_ap_req(
                            mtafqdn,
                            srvcfqdn,
                            srvc_ip_addr,
                            krbrealm,
                            krbto,
                            app_flag,
                            server_nonce,
                            pAppData,
                            pAppRep,
                            credentials,
                            application_key,
                            authTID,
                            encTID,
                            reason);

#if 0
        krbPrintBufCounts();
#endif

        
        LIBDEBUG(".k  secGetKrbApplicationKey(): -retval from send_ap_req -<%d>, ERROR_TABLE_BASE_krb5 = <%d>\n",retval, ERROR_TABLE_BASE_krb5);
        /* analyse error returned */
        rvalue = retval + ERROR_TABLE_BASE_krb5;
        switch ( rvalue )
        {
        case KRB5KRB_AP_ERR_SKEW:
            LOG(("Time skew written to NVRAM, sending AP_REQ with adjusted time."));
            retval = send_ap_req(
                                mtafqdn,
                                srvcfqdn,
                                srvc_ip_addr,
                                krbrealm,
                                krbto,
                                app_flag,
                                server_nonce,
                                pAppData,
                                pAppRep,
                                credentials,
                                application_key,
                                authTID,
                                encTID,
                                reason);
            break;
        case KRB5_KDCREP_SKEW:
            LOG(("CAN'T TRUST APPLICATION SERVER, SKEW GREATER THAN 1 HR."));
            return retval;
        case KRB5KRB_AP_ERR_BADKEYVER:
            LOG(("KRB5KRB_AP_ERR_BADKEYVER Error Received."));
            return retval;
        default:
            LOG(("default switch rvalue = <%d>\n",rvalue));
        }


#if 0
        krbPrintBufCounts();
#endif

        if ( !retval )
        {
            /* store the subkey localy for endpoints sharing the same CMS */
            if ( app_flag == APPLICATION_CMS )
            {
                secStoreSubkeyInfo(application_key->sub_key.sub_key_val,
                                   srvcfqdn,
                                   application_key->sec_params_lifetime,
                                   application_key->grace_period);
                secUpdateServerAuthKey(credentials->session_key.session_key_val,
                                       credentials->session_key.session_key_len,
                                       srvcfqdn);
            }
            return 0;
        }
        else
            return retval;
    }





    
