/***************************************************************************
*    Copyright 2000  Broadcom Corporation
*    All Rights Reserved
*    No portions of this material may be reproduced in any form without the
*    written permission of:
*             Broadcom Corporation
*             16251 Laguna Canyon Road
*             Irvine, California  92618
*    All information contained in this document is Broadcom Corporation
*    company private, proprietary, and trade secret.
*
****************************************************************************
*
*    Filename: dns.c
*    Creation Date: 28 June 2001 (v0.00)
*    VSS Info:
*        $Revision: 1.1 $
*        $Date: 2002/08/21 17:54:37 $
*
****************************************************************************
*    Description:
*
*     Broadcom Generic Operating System Functions. This file contains the VxWorks
*     specific implementation of the domain name service (DNS).
*
****************************************************************************/

#include <hausenv.h>
#include <string.h>
//#include <bcmos.h>
#include "dns.h"
#include "socket.h"
//#include "bcmosHelper.h"
/* Operating system specific includes */
#include <vxWorks.h>
#include <netdb.h>
#include <in.h>
#include <hostLib.h>
#include <u_Lib.h>


/*
*****************************************************************************
** TYPE DEFINITIONS and CONSTANTS
*****************************************************************************
*/

/* Domain Name Service (DNS) state structure */
typedef struct DNS_STATE
{
   BOOL        bActive;    /* Indicates if DNS has been started */
   OS_SEMID    mutex;      /* Concurrency protection */
} DNS_STATE;



/*
*****************************************************************************
** LOCAL VARIABLES
*****************************************************************************
*/

/* Domain name service (DNS) state */
static DNS_STATE  dns;


/*****************************************************************************
** FUNCTION:   dnsInit
**
** PURPOSE:    Initialize the domain name server (DNS)
**
** PARAMETERS: None
**
** RETURNS:    OS_STATUS
*****************************************************************************/
OS_STATUS dnsInit( void )
{
   /* Initialize DNS state */
   dns.bActive = FALSE;
   dns.mutex = 0;

   return( OS_STATUS_OK );
}


/*****************************************************************************
** FUNCTION:   bcmOsDnsGetHostName
**
** PURPOSE:    Given an IP address, this function finds the corresponding host
**             name and returns it using the domain name service (DNS).
**
**
** PARAMETERS: ipAddr   [IN] - Specifies the address to be resolved to a name (in
**                               (host-byte order).
**             hostName [OUT] - Specifies the pointer to a buffer that is filled
**                               with the resolved name. The caller should
**                               preallocate the buffer to store the resolved name
**                               (which can be upto 255 characters).
**
** RETURNS:    OS_STATUS
**
*****************************************************************************/
OS_STATUS bcmOsDnsGetHostName( IPADDR ipAddr, char *hostName )
{
   NATIVE_SOCKET_STATUS  status = ERROR;

   /* Error checking */
   return ( checkSocketStatus( status, "bcmOsGetHostByAddr" ));
}


/***************************************************************************
** FUNCTION:   bcmOsDnsGetHostAddr
**
** PURPOSE:    Given the hostname, this function finds the corresponding IP
**             address and returns it using the domain name service (DNS).
**
** PARAMETERS: hostName [IN] - host name to resolve into an IP address
**             ipAddr   [OUT] - resolved IP address (in host-byte order)
**
** RETURNS:    OS_STATUS
**
*****************************************************************************/
OS_STATUS bcmOsDnsGetHostAddr( char *hostName, IPADDR *ipAddr )
{
   
#if 1
    NATIVE_SOCKET_STATUS status = ERROR;
    
   /* Error checking */
   return ( checkSocketStatus( status, "bcmOsGetHostByName" ));
#endif
    
#if 0
      NATIVE_SOCKET_STATUS status;
      struct hostent host;
   
      /* Prime the host entity structure */
      memset((char *) &host, 0, sizeof( struct hostent ));
      host.h_addr_list[0] = (char *) ipAddr;
   
      /* Resolve the IP address */
      status = gethostbyname( hostName, &host );
   
      /* Convert to host byte-order */
      *ipAddr = ntohl( *ipAddr );
   
      /* Error checking */
      return ( checkSocketStatus( status, "bcmOsGetHostByName" ));
#endif
}

/****************************************************************************
** FUNCTION:   bcmOsDnsGetSRVName
**
** PURPOSE:    Retrieves the SRV record and returns target FQDN Name of server
**             providing requested service on requested domain.
**
** PARAMETERS: hostSRVName[IN]- host name and service with format
**                               "_kerberos._udp.broadcom.com\0"
**
**             srvtarget[OUT] - target servers providing service with format:
**                              "kerbserver1.broadcom.com.\0". If more than
**                              one server present, servers listed in order,
**                              "kerbserv1.broadcom.com.\0kerb2.broadcom.com.\0\0"
**
** RETURNS:    OS_STATUS
**
*****************************************************************************/
OS_STATUS bcmOsDnsGetSRVName( char *hostSRVName, char *srvtarget )
{
   NATIVE_SOCKET_STATUS status = ERROR;

   /* Error checking */
   return ( checkSocketStatus( status, "bcmOsDnsGetSRVName" ));
}


/***************************************************************************
** FUNCTION:   bcmOsDnsGetHostAddrs
**
** PURPOSE:    Given the hostname, this function finds corresponding IP
**             addresses and returns them.
**
** PARAMETERS: hostName [IN]        - host name to resolve into IP addresses
**             ipAddr   [OUT]       - array of resolved IP addresses (in
**                                     host byte order). Memory must be
**                                     allocated by caller.
**             numIpAddrs [IN/OUT]  - On input, points to the array length
**                                     of 'ipAddr'. On output, returns the
**                                     number of resolved IP addresses contained
**                                     in the 'ipAddr' array.
**             useCache [IN]        - indicates whether to use the DNS cache
**
** RETURNS:    OS_STATUS
**
*****************************************************************************/
OS_STATUS bcmOsDnsGetHostAddrs( char *hostName, IPADDR *ipAddr, int *numIpAddrs, BOOL useCache )
{
   NATIVE_SOCKET_STATUS status = ERROR;

   /* Error checking */
   return ( checkSocketStatus( status, "bcmOsGetHostByName" ));
}


/****************************************************************************
** FUNCTION:   bcmOsStartDns
**
** PURPOSE:    Launches the domain name service (DNS).
**
** PARAMETERS: dnsAddr  [IN] -  array of DNS server IP addresses (in host-byte
**                               order)
**             nServers [IN] - the number of DNS server IP addresses given in the
**                         list 'dnsAddr'. Must be less than OS_DNS_MAX_SERVERS.
**
** RETURNS:    OS_STATUS
**
*****************************************************************************/
OS_STATUS bcmOsStartDns( IPADDR *dnsAddr, int nServers, int ipstacknum )
{
   NATIVE_SOCKET_STATUS status = ERROR;

   /* Error checking */
   return ( checkSocketStatus( status, "bcmOsStartDns" ));
}
