/***************************************************************************
*    Copyright 2000  Broadcom Corporation
*    All Rights Reserved
*    No portions of this material may be reproduced in any form without the
*    written permission of:
*             Broadcom Corporation
*             16251 Laguna Canyon Road
*             Irvine, California  92618
*    All information contained in this document is Broadcom Corporation
*    company private, proprietary, and trade secret.
*
****************************************************************************
*
*    Filename: mutex.c
*    Creation Date: 28 June 2001 (v0.00)
*    VSS Info:
*        $Revision: 1.1 $
*        $Date: 2002/08/21 17:54:40 $
*
****************************************************************************
*    Description:
*
*     Broadcom Generic Operating System Functions. This file contains a VxWorks
*     specific implementation of mutual exclusion functions.
*
****************************************************************************/

#include <hausenv.h>
#include <string.h>
#include <bcmos.h>

/* Operating system specific includes */
#include <vxWorks.h>
#include <intLib.h>

#include <bcmosHelper.h>

/*****************************************************************************
** FUNCTION:   bcmOsLock
**
** PURPOSE:    Prevent the OS from switching tasks for concurrency protection
**
** PARAMETERS: lockKey - pointer to OS_LOCKKEY
**
** RETURNS:    OS_STATUS_OK
**
** NOTES:      The lockKey value is passed back to bcmOsUnlock to re-enable
**             the OS's scheduler
*****************************************************************************/
OS_STATUS bcmOsLock(OS_LOCKKEY *lockKey)
{
   ASSERT (lockKey);

   *lockKey = (OS_LOCKKEY) intLock();

   return ( OS_STATUS_OK );
}


/*****************************************************************************
** FUNCTION:   bcmOsUnlock
**
** PURPOSE:    Re-enable the OS sheduler
**
** PARAMETERS: lockKey - input - value set by bcmOsLock
**
** RETURNS:    OS_STATUS_OK
**
** NOTES:
*****************************************************************************/
OS_STATUS bcmOsUnlock( OS_LOCKKEY lockKey )
{
   intUnlock((int) lockKey );

   return ( OS_STATUS_OK );
}

///////////////////////////////////////////

/*****************************************************************************
** FUNCTION:   bcmOsMutexCreate
**
** PURPOSE:    Create a mutex
**
** PARAMETERS: mutexName   - mutex name
**             mutexId     - pointer to the mutex handle
**
** RETURNS:    OS_STATUS_OK - success
**             error code   - failure
**
** NOTE:
******************************************************************************/
OS_STATUS bcmOsMutexCreate(char *mutexName, OS_MUTEXID *mutexId)
{
   return BcmMuCreate(mutexId);
   //NATIVE_STATUS  status;
   //
   //ASSERT( mutexName && mutexId );
   //
   ///* Params: mutex name, flags, ceiling, id */
   //status = mu_create(
   //            mutexName,
   //            MU_LOCAL | MU_PRIOR | MU_RECURSIVE | MU_PRIO_INHERIT,
   //            0, /* ceiling only used for PRIO_PROTECT */
   //            mutexId);
   //
   ///* Error checking */
   //return( checkStatus( status, "bcmOsMutexCreate" ));
}


/*****************************************************************************
** FUNCTION:   bcmOsMutexLock
**
** PURPOSE:    Lock a mutex
**
** PARAMETERS: mutexId        - Handle of the mutex
**             timeout_ticks  - timeout in ticks
**
** RETURNS:    OS_STATUS_OK   - success
**             error code     - failure
**
** NOTE:       this function ASSERTs that the operation was successful
******************************************************************************/
OS_STATUS bcmOsMutexLock(OS_MUTEXID mutexId, OS_TICKS timeout)
{
    return BcmMuLock(mutexId);
    
   //NATIVE_STATUS status;
   //
   //ASSERT( mutexId );
    //
   ///* Convert to OS-specific tick value */
   //if ( timeout != OS_WAIT_FOREVER )
   //{
   //   timeout = timeout / OS_TICK_SCALE_FACTOR;
   //}
   //
   //
   //if ( timeout == OS_WAIT_FOREVER )
   //{
   //   /* Wait forever */
   //   status = mu_lock( mutexId, MU_WAIT, 0 );
   //}
   //else if ( timeout )
   //{
   //   /* Wait until either mutex is given, or timeout expires */
   //   status = mu_lock( mutexId, MU_WAIT, timeout );
   //}
   //else
   //{
   //   /* Non-blocking mutex take */
   //   status = mu_lock( mutexId, MU_NOWAIT, 0 );
   //}
   //
   ///* Error checking */
   //return( checkStatus( status, "bcmOsMutexLock" ));

}


/*****************************************************************************
** FUNCTION:   bcmOsMutexUnlock
**
** PURPOSE:    Unlock a mutex
**
** PARAMETERS: mutexId        - Handle of the mutex
**
** RETURNS:    OS_STATUS_OK   - success
**             error code     - failure
**
** NOTE:
******************************************************************************/
OS_STATUS bcmOsMutexUnlock( OS_MUTEXID mutexId )
{
    return BcmMuUnlock(mutexId);
   //NATIVE_STATUS  status;
   //
   //ASSERT( mutexId );
   //
   //status = mu_unlock( mutexId );
   //
   ///* Error checking */
   //return( checkStatus( status, "bcmOsMutexUnlock" ));
}


/*****************************************************************************
** FUNCTION:   bcmOsMutexDelete
**
** PURPOSE:    deletes a mutex
**
** PARAMETERS: mutexID - mutex identification
**
** RETURNS:    OS_STATUS_OK - success
**             error code   - failure
**
** NOTE:
******************************************************************************/
OS_STATUS bcmOsMutexDelete( OS_MUTEXID mutexId )
{
   return BcmMuDelete(mutexId);
   //NATIVE_STATUS  status;
   //
   ///* Delete the task */
   //status = mu_delete(mutexId);
   //
   /* Error checking */
   //return( checkStatus( status, "bcmOsMutexDelete" ));
}

