/*
 * Copyright (C) 1998 by the FundsXpress, INC.
 * 
 * All rights reserved.
 * 
 * Export of this software from the United States of America may require
 * a specific license from the United States Government.  It is the
 * responsibility of any person or organization contemplating export to
 * obtain such a license before exporting.
 * 
 * WITHIN THAT CONSTRAINT, permission to use, copy, modify, and
 * distribute this software and its documentation for any purpose and
 * without fee is hereby granted, provided that the above copyright
 * notice appear in all copies and that both that copyright notice and
 * this permission notice appear in supporting documentation, and that
 * the name of FundsXpress. not be used in advertising or publicity pertaining
 * to distribution of the software without specific, written prior
 * permission.  FundsXpress makes no representations about the suitability of
 * this software for any purpose.  It is provided "as is" without express
 * or implied warranty.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <stdio.h>
#include <errno.h>
//#include <memory.h>
#include <string.h>
#include "seckrb.h"
#include "old.h"

#ifndef HAVE_MEMMOVE
#ifdef HAVE_BCOPY
#define memmove(dst,src,size) bcopy(src,dst,size)
#endif
#endif


/* Added by Broadcom: Determine the length to get rid of random pads,
   MIT assumed pads are always 0
*/
void get_enc_part_length(krb5_data *length_bytes, int *enc_part_length)
{
   unsigned char tmp;
   int num;
   int len=0;
   int i=0;

   tmp = length_bytes->data[i];
   i++;
   if((tmp&0x80) == 0)
   {
      *enc_part_length = (int)(tmp&0x7F);
   }
   else
   {
      for(num = (int)(tmp&0x7F); (num>0)&&(num<8); num--)
      {
         tmp = length_bytes->data[i];
         i++;
         len = (len<<8) + (int)tmp;
      }
      *enc_part_length = len;
   }
}

   
krb5_error_code krb5_old_decrypt(enc, hash, key, usage, ivec, input, arg_output)
   const  krb5_enc_provider *enc;
   const  krb5_hash_provider *hash;
   const krb5_keyblock *key;
   krb5_keyusage usage;
   const krb5_data *ivec;
   const krb5_data *input;
   krb5_data *arg_output;
{
   krb5_error_code ret;
   size_t blocksize, hashsize, plainsize;
   unsigned char *cksumdata;
   krb5_data output, cksum, crcivec;
   int alloced;
   krb5_data length_bytes;
   int enc_part_length;


   (*(enc->block_size))(&blocksize);
   (*(hash->hash_size))(&hashsize);

   plainsize = input->length - blocksize - hashsize;

   if (arg_output->length < (int)plainsize)
      return(KRB5_BAD_MSIZE);

   /* if there's enough space to work in the app buffer, use it,
      otherwise allocate our own */

   if ((cksumdata = (unsigned char *) krbAllocBuf(hashsize)) == NULL)
   {
      printf("Nomem #1019.\n");
      return(ENOMEM);
   }

   if (arg_output->length < input->length) {
      output.length = input->length;

      if ((output.data = (krb5_octet *) krbAllocBuf(output.length)) == NULL) {
         krbFreeBuf(cksumdata);
         printf("nomem #1020.\n");
         return(ENOMEM);
      }

      alloced = 1;
   } else {
      output.length = input->length;

      output.data = arg_output->data;

      alloced = 0;
   }

   /* decrypt it */

   /* XXX this is gross, but I don't have much choice */
   if ((key->enctype == ENCTYPE_DES_CBC_CRC) && (ivec == 0)) {
      crcivec.length = key->length;
      crcivec.data = key->contents;
      ivec = &crcivec;
   }

   if (ret = ((*(enc->decrypt))(key, ivec, input, &output)))
      goto cleanup;

   /* For packet-cable Get rid of random pads */
   /* copy 8 bytes or so to be decoded for length */
   length_bytes.data = krbAllocBuf(8);
   memcpy(length_bytes.data, &(output.data[blocksize+hashsize+1]),
         blocksize);
   length_bytes.length = 8;
   get_enc_part_length(&length_bytes, &enc_part_length);

   if ( enc_part_length <= 127 )       /* xx length itself */
      enc_part_length += 2;
   else if ( (enc_part_length > 127) && (enc_part_length <= 255 ) )
      enc_part_length +=  3;   /* 0x81 xx one byte length */
   else if (enc_part_length > 255 )
      enc_part_length += 4;  /*0x82 xx xx 2 bytes length */
   else
   {
      printf("Bad as-rep enc-part length %d.\n", enc_part_length);
      return 1;
   }

   /* set the length for hash not to include random pads */
   output.length = enc_part_length+blocksize+hashsize;

   /* verify the checksum */

   memcpy(cksumdata, output.data+blocksize, hashsize);
   memset(output.data+blocksize, 0, hashsize);

   cksum.length = hashsize;
   cksum.data = output.data+blocksize;

   if (ret = ((*(hash->hash))(1, &output, &cksum)))
      goto cleanup;

   if (memcmp(cksum.data, cksumdata, cksum.length) != 0) {
      ret = KRB5KRB_AP_ERR_BAD_INTEGRITY;
      goto cleanup;
   }

   /* copy the plaintext around */

   if (alloced) {
      memcpy(arg_output->data, output.data+blocksize+hashsize,
            plainsize);
   } else {
      memmove(arg_output->data, arg_output->data+blocksize+hashsize,
            plainsize);
   }
   arg_output->length = plainsize;

   ret = 0;

cleanup:
   if (alloced) {
      memset(output.data, 0, output.length);
      krbFreeBuf(output.data);
   }

   memset(cksumdata, 0, hashsize);
   krbFreeBuf(cksumdata);
   return(ret);
}

