/***************************************************************************
 *     Copyright 2000  Broadcom Corporation
 *     All Rights Reserved
 *     No portions of this material may be reproduced in any form without the
 *     written permission of:
 *              Broadcom Corporation
 *              16251 Laguna Canyon Road
 *              Irvine, California  92618
 *     All information contained in this document is Broadcom Corporation
 *     company private, proprietary, and trade secret.
 *
 ****************************************************************************
 *
 *     Filename:  SecLib.c
 *     Creation Date: 06 September 2000 (v0.00)
 *     VSS Info:
 *         $Revision: 1.1 $
 *         $Date: 2002/08/21 17:53:44 $
 *
 ****************************************************************************
 *     Description:
 *
 *       This file contains implementations of methods required by
 *       CablexChange Security purposes
 *
 *     Notes:
 *        OpenSSL package is used widely underneath.
 *
 ****************************************************************************/
 /*
 6-6-02
 Added OpenSSL includes
 removed hausevn.h
 */
/*
*****************************************************************************
* INCLUDE FILES
*****************************************************************************
*/

#ifdef WIN32
#define SECLIB_DLL
#define LINKDLL __declspec( dllexport)
#endif

/* system includes */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
//#include <hausenv.h>

/* OpenSSL includes */
#include "crypto.h"
#include "dh.h"
#include "rand.h"
#include "asn1.h"
#include "x509.h"
#include "bio.h"
#include "x509v3.h"
#include "pem.h"
#include "rsa.h"
#include "bn.h"
#include "objects.h"
///#include "apps.h"
#include "err.h"
#include "pem.h"
#include "sha.h"
#include "hmac.h"
#include <evp.h>

/* library header file */
#include "SecLib.h"

#define KEYSIZ	8

const signed char enctab[64] = {     /* radix64 encoding table */
  'A','B','C','D','E','F','G','H','I','J','K','L','M','N','O','P',
  'Q','R','S','T','U','V','W','X','Y','Z','a','b','c','d','e','f',
  'g','h','i','j','k','l','m','n','o','p','q','r','s','t','u','v',
  'w','x','y','z','0','1','2','3','4','5','6','7','8','9','+','/'
} ; /* enctab[] */

const signed char dectab[256] = {        /* radix64 decoding table */
   -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
   -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
   -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 62, -1, -1, -1, 63,
   52, 53, 54, 55, 56, 57, 58, 59, 60, 61, -1, -1, -1, 64, -1, -1,
   -1,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14,
   15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, -1, -1, -1, -1, -1,
   -1, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40,
   41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, -1, -1, -1, -1, -1,
   -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
   -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
   -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
   -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
   -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
   -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
   -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
   -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1
} ; /* dectab[] */

/*
*****************************************************************************
* STATIC DECLARATIONS
*****************************************************************************
*/
static void brcm_des_string_to_key(const char *inkey, des_cblock *key);
static void dec64(unsigned char *outbuff, int out_lth, int *polth, unsigned char *inbuff);
static void enc64(unsigned char *outbuff, int out_lth, int *polth, unsigned char *inbuff, int inb_lth, int line_lth, int n_space);

/* random seed */
static const char rnd_seed[] = "string to make the random number generator think it has entropy";
static int randLatch;

static unsigned int secCryptoDhGener( 
      const unsigned char           *pPrime,
      unsigned long                 primeLen,
      unsigned long                 generator,
      unsigned char                 **pPubKey,
      unsigned long                 *pubKeyLen,
      unsigned char                 **pPrivKey,
      unsigned long                 *privKeyLen );
static unsigned int secCryptoDhComputeSecret( 
      unsigned char            **pSecretKey,
      unsigned long            *secretKeyLen,
      unsigned char            *pPrivKey,
      unsigned long            privKeyLen,
      unsigned char            *pPubKeyOther,
      unsigned long            pubKeyOtherLen,
      const unsigned char      *pPrime,
      unsigned long            primeLen,
      unsigned long            generator );    
static unsigned int secCryptoBitStringEnc( 
      unsigned char           *pKeyBuffer,
      unsigned long           keyBufferLen,
      unsigned char           **ppBitStrBuf,
      unsigned long           *bitStrBufLen);
static unsigned long secCryptoSign( 
      unsigned char            *inBuffer,
      unsigned long            inBufferLen,
      RSA_key                  *key,
      unsigned char            **outSignature,
      unsigned int             *outSignatureLen );
static unsigned long secCryptoSignVerify( 
      unsigned char               *inData,
      unsigned long               inDataLen,
      unsigned char               *pSignature,
      unsigned long               pSigLen, 
      RSA_key                     *key );
static int secCryptoPkcs5KeyGener( 
      const char          *pPassword,
      int                 passwordLen,
      unsigned char       *pSalt,
      int                 saltLen,
      int                 iteration,
      int                 keyBufferLen,
      unsigned char       *pKeyBuffer );
static unsigned long secCryptoSign( 
      unsigned char            *inBuffer,
      unsigned long            inBufferLen,
      RSA_key                  *key,
      unsigned char            **outSignature,
      unsigned int             *outSignatureLen );
static unsigned long secCryptoSignVerify( 
      unsigned char               *inData,
      unsigned long               inDataLen,
      unsigned char               *pSignature,
      unsigned long               pSigLen, 
      RSA_key                     *key );
static void secCryptoPHash(
      EVP_MD                   *md,
      unsigned char            *sec,
      unsigned long            sec_len,
      const unsigned char      *seed,
      unsigned long            seed_len,
      unsigned char            *out,
      unsigned long            olen);
static unsigned long secCryptoDeriveSNMPv3Key(
      unsigned char         *secret,
      unsigned long         secret_len,
      const unsigned char   *seed,
      unsigned long         seed_len,
      unsigned char         *auth_key,
      int                   auth_tid,
      unsigned char         *enc_key,
      int                   enc_tid);
static unsigned long secCryptoDeriveIPSecKey(
      unsigned char        *secret,
      unsigned long        secret_len,
      const unsigned char  *seed,
      unsigned long        seed_len,
      unsigned char        *outbound_auth_key,
      unsigned char        *outbound_enc_key,
      unsigned char        *inbound_auth_key,
      unsigned char        *inbound_enc_key,
      int                  auth_tid,
      int                  enc_tid);
static unsigned long secCryptoCalcHash(
      int               hashType,
      unsigned char     *inBuf,
      unsigned long     inBufLen,
      unsigned char     *hashKey,
      unsigned long     hashKeyLen,
      unsigned char     *outHash,
      unsigned long     *outHashLen);
static int secCryptoDes(
      char        *inBuf, 
      char        *outBuf, 
      int         *bufLen, 
      char*       key, 
      int         enc);
static int secCryptoDeriveRTPKeys(
      unsigned char      *seed,
      unsigned long      seedLen,
      unsigned char      *secret,
      unsigned long      secretLen,
      unsigned char      *pad,
      unsigned long      padLen,
      unsigned long      rtpPrivKeyLen,
      unsigned long      rtpInitKeyLen,
      unsigned long      rtpMacKeyLen,
      unsigned char      *rtpPrivKey,
      unsigned long      *rtpInitTS,
      unsigned char      *rtpInitKey,
      unsigned char      *rtpMacKey);
static int secCryptoDeriveRTCPKeys(
      unsigned char      *seed,
      unsigned long      seedLen,
      unsigned char      *secret,
      unsigned long      secretLen,
      unsigned char      *pad,
      unsigned long      padLen,
      unsigned long      rtcpEncKeyLen,
      unsigned long      rtcpAuthKeyLen,
      unsigned char      *rtcpEncKey,
      unsigned char      *rtcpAuthKey);
static void secCryptoBase64(
      unsigned char   *inBuf, 
      int             inLen, 
      unsigned char   *outBuf, 
      int             *outLen, 
      int             encode, 
      int             maxLength);
static int secCryptoGetCurrUTCTime(
      int                  ptime, 
      unsigned char        *utc_time_str,
      unsigned long        *utc_str_len);
static int secCryptoConvertUTCToTime(
      unsigned char        *utc_time_str,
      unsigned long        utc_str_len,
      unsigned long        *ptime);


/* assign functions to pointers */
static SEC_CRYPTO_IFACE   secCryptoIface =
{
   secCryptoDhGener,
   secCryptoDhComputeSecret,
   secCryptoBitStringEnc,
   secCryptoSign,
   secCryptoSignVerify,
   secCryptoPkcs5KeyGener,
   secCryptoDeriveSNMPv3Key,
   secCryptoDeriveIPSecKey,
   secCryptoCalcHash,
   secCryptoDes,
   secCryptoDeriveRTPKeys,
   secCryptoDeriveRTCPKeys,
   secCryptoBase64,
   secCryptoGetCurrUTCTime,
   secCryptoConvertUTCToTime
};


void Init_secCrypto(void)
{
   randLatch = 0;
}

void Deinit_secCrypto(void)
{
}

/*
*****************************************************************************
**
**  Function Name: SecLibGetCryptoIface
**
**  PURPOSE:   return iface struct containing pointers to seclib functions
**
**  PARAMETERS:
**
**  RETURNS:
**
**  NOTES:
*****************************************************************************
*/
SEC_CRYPTO_IFACE*   SecLibGetCryptoIface(void)
{
   return &secCryptoIface;
}


/*
******************************************************************************
** FUNCTION:   secCryptoDhGener
**
** PURPOSE:    To retrieve Diffie-Hellman private and public key given the parameters
**
** PARAMETERS:
**    1. pPrime - pointer to prime number buffer
**    2. primeLen - length of prime number buffer in bytes
**    3. generator - DH generator
**    4. pPubKey - public key buffer, used an output parameter
**    5. pPubKeyLen - length of public key buffer in bytes
**    6. pPrivKey - private key buffer, used as an ouput
**    7. privKeyLen - length of private key buffer in bytes
**
**
** RETURNS:   Success or error code - unsigned long
**
** NOTE:
**
*****************************************************************************
*/
static unsigned int secCryptoDhGener( 
      const unsigned char           *pPrime,
      unsigned long                 primeLen,
      unsigned long                 generator,
      unsigned char                 **pPubKey,
      unsigned long                 *pubKeyLen,
      unsigned char                 **pPrivKey,
      unsigned long                 *privKeyLen )

{

   DH *dhKeyStruc;               /* Diffie-Hellman key data structure */
   unsigned int returnCode;      /* code to be returned */
   unsigned long size;           /* local varible to hold key sizes */


   if ( (!pPrime) || (primeLen < 1) || ((generator != 2) && (generator != 5)) )
   {
      return BAD_DATA;
   }

   /* seed the random number generator only once */
   if ( !randLatch )
   {
      RAND_seed( rnd_seed, sizeof rnd_seed );
      randLatch = 1;
   }

   /* allocate the space for DH key */
   dhKeyStruc = DH_new();

   /* assign the prime number field in DH key to prime input buffer */
   dhKeyStruc->p = BN_bin2bn( pPrime, primeLen, NULL );

   /* set the generator and check for error */
   dhKeyStruc->g=BN_new();

   if (!BN_set_word(dhKeyStruc->g,generator))
   {
      /* free memory */
      DH_free( dhKeyStruc );

      return GENERATOR_ERROR;
   }


   /* generate private and public key and check for error */
   returnCode = DH_generate_key( dhKeyStruc );

   if ( !returnCode )
   {
      /* free memory */
      DH_free( dhKeyStruc );

      return KEY_GEN_ERROR;
   }


   /* allocate memory for public key */
   size = BN_num_bytes( dhKeyStruc->pub_key );
   *pPubKey = (unsigned char*)malloc(size);

   /* put public key in output buffer */
   *pubKeyLen = BN_bn2bin( dhKeyStruc->pub_key, *pPubKey );

   /* allocate memory for private key */
   size = BN_num_bytes( dhKeyStruc->priv_key );
   *pPrivKey = (unsigned char*)malloc(size);

   /* put private key in ouput buffer */
   *privKeyLen = BN_bn2bin( dhKeyStruc->priv_key, *pPrivKey );

   /* free memory */
   DH_free( dhKeyStruc );

   /* return success */
   return OK;

}

/*
******************************************************************************
** FUNCTION:   secCryptoDhComputeSecret
**
** PURPOSE:    To compute and return Diffie-Hellman shared secret
**
** PARAMETERS:
**       1. pSecretKey - pointer to secret key buffer, output parameter
**       2. secretKeyLen - lenght of secret key buffer in bytes, output parameter
**       3. pPrivKey - private key buffer, used as an ouput
**       4. privKeyLen - length of private key buffer in bytes
**       5. pPubKeyOther - other party public key buffer
**       6. pubKeyOtherLen - length of other party public key in bytes
**       7. pPrime - pointer to prime number buffer
**       8. primeLen - length of prime number buffer in bytes
**       9. generator - DH generator
**
**
** RETURNS:   Success or error code - unsigned long
**
** NOTE:
**
*****************************************************************************
*/
static unsigned int secCryptoDhComputeSecret( 
      unsigned char            **pSecretKey,
      unsigned long            *secretKeyLen,
      unsigned char            *pPrivKey,
      unsigned long            privKeyLen,
      unsigned char            *pPubKeyOther,
      unsigned long            pubKeyOtherLen,
      const unsigned char      *pPrime,
      unsigned long            primeLen,
      unsigned long            generator )

{
   DH *dhKeyStruc;				/* Diffie-Hellman key structure */
   BIGNUM* pubKeyOtherBN;		/* big number representation of other party public key */
   unsigned int size;			/* local variable */


   if ( !pPrivKey || !pPubKeyOther || !pPrime || ( (generator != 2) && (generator != 5) ) )
   {
      return BAD_DATA;
   }

   /* seed the random number generator only once */
   if ( !randLatch )
   {
      RAND_seed( rnd_seed, sizeof rnd_seed );
      randLatch = 1;
   }


   /* allocate the space for DH key */
   dhKeyStruc = DH_new();

   /* assign the prime number field in DH key to prime input buffer */
   dhKeyStruc->p = BN_bin2bn( pPrime, primeLen, NULL );

   /* set the generator and check for error */
   dhKeyStruc->g=BN_new();

   if (!BN_set_word(dhKeyStruc->g,generator))
   {
      /* free memory */
      DH_free( dhKeyStruc );

      return GENERATOR_ERROR;
   }

   /* assign the private key field in DH key to private key input buffer */
   dhKeyStruc->priv_key = BN_bin2bn( pPrivKey, privKeyLen, NULL );

   /* convert otherPublicKey to big number representation */
   pubKeyOtherBN = BN_bin2bn( pPubKeyOther, pubKeyOtherLen, NULL );

   /* allocate memory for secret key */
   size = DH_size( dhKeyStruc );

   *pSecretKey = (unsigned char*)malloc(size);

   /* compute the secret key and check for error */
   if ( !(*secretKeyLen = DH_compute_key( *pSecretKey, pubKeyOtherBN, dhKeyStruc ) ) )
   {
      /* free memory */
      DH_free( dhKeyStruc );

      return KEY_GEN_ERROR;
   }

    /* free memory */
   DH_free( dhKeyStruc );
   BN_free( pubKeyOtherBN );
   /* return success */
   return OK;
}

/*
*****************************************************************************
** FUNCTION:   secCryptoBitStringEnc
**
** PURPOSE:    To get ASN1 bitstring representation of input buffer with given leength
**
** PARAMETERS:
**    1. pKeyBuffer - pointer to the input buffer, usually it is a key
**    2. KeyBufferLen - length of input buffer in bytes
**    3. ppBitStrBuf - output buffer, output parameter
**    4. bitStrBufLen - length of the ouput buffer, used as an output parameter
**
** RETURNS:   Success or error code - unsigned long
**
** NOTE:
**
******************************************************************************
*/


static unsigned int secCryptoBitStringEnc( 
      unsigned char           *pKeyBuffer,
      unsigned long           keyBufferLen,
      unsigned char           **ppBitStrBuf,
      unsigned long           *bitStrBufLen)
{
   ASN1_BIT_STRING *bitString;             /* OpenSSL ASN1 type */
   unsigned int bytesNumberForSize;    /* local variable, number of bytes needed for size */


   /* check input parameters */
   if ( !pKeyBuffer || !keyBufferLen )
   {
      return BAD_DATA;
   }

   /* create new bit string and check for errors */
   if ( !(bitString = ASN1_STRING_new()) )
   {
      return MEM_ALLOC_ERROR;
   }


   /* assign value to bit string and check for erorrs */
   if ( !(ASN1_STRING_set( bitString, pKeyBuffer, keyBufferLen )) )
   {
      /* free the ASN1 type */
      ASN1_STRING_free( bitString );

      return STRING_SET_ERROR;
   }

   /*
    * Get number of bytes for buffer size,
    * if it needs more than 3 bytes, consider
    * it as an error.
   */
   if ( keyBufferLen <= 0xff )
   {
      bytesNumberForSize = 1;
   }
   else if ( keyBufferLen <= 0xffff )
   {
      bytesNumberForSize = 2;
   }
   else if ( keyBufferLen <= 0xffffff)
   {
      bytesNumberForSize = 3;
   }
   else
   {
      /* free the ASN1 type */
      ASN1_STRING_free( bitString );

      return BAD_DATA;
   }

   /*
    * allocate space for output buffer: 1 for BIT STRING specifier,
    * 1 for padding byte, bytesNumberForSize for buffer size length,
    * and keyBufferLen-1 for size of input buffer
   */
   *ppBitStrBuf = (unsigned char*)malloc( 1 + 1 + bytesNumberForSize + keyBufferLen-1 );

   /* encode and put bit string into output buffer */
   *bitStrBufLen = i2d_ASN1_BIT_STRING( bitString, ppBitStrBuf );

   /* update pointer to the beggining of the buffer */
   *ppBitStrBuf = *ppBitStrBuf - *bitStrBufLen;

   /* free the ASN1 type */
   ASN1_STRING_free( bitString );
   /* if here return success */
   return OK;
}



/*
******************************************************************************
** FUNCTION:   secCryptoPkcs5KeyGener
**
** PURPOSE:    To compute key according to the PKCS5 Password-Based Cryptography Standard
**
** PARAMETERS:
**    1. pPassword - pointer to password buffer
**    2. passwordLen - length of password buffer in bytes
**    3. pSalt - salt buffer
**    4. saltLen - length of salt buffer
**    5. iteration - number of iterations
**    6. keyBufferLen - length of output key buffer, input parameter
**    7. pKeyBuffer - output key buffer
**
**
** RETURNS:   Success or error code - unsigned long
**
** NOTE:
**
******************************************************************************
*/
static int secCryptoPkcs5KeyGener( 
      const char          *pPassword,
      int                 passwordLen,
      unsigned char       *pSalt,
      int                 saltLen,
      int                 iteration,
      int                 keyBufferLen,
      unsigned char       *pKeyBuffer )
{

   /* only keyLengths in interval [1,20] are accepted - SHA1 digest length is 20 */
   if ( !( passwordLen && saltLen) || ( keyBufferLen > SHA_DIGEST_LENGTH ) ||
         ( keyBufferLen < 0 ) || (pKeyBuffer == NULL) )
   {
      return BAD_DATA;
   }

   /* call to corresponding method in OpenSSL */
   PKCS5_PBKDF2_HMAC_SHA1( pPassword, passwordLen, pSalt, saltLen, iteration,
                              keyBufferLen, pKeyBuffer );
   return OK;
}



/*
*****************************************************************************
**  Function Name: secCryptoSign
**
**  PURPOSE:
**     To sign the data in input buffer using the input private key buffer
**     Key used is RSA key and the digest method is SHA1
**
**  PARAMETERS:
**      1. inBuffer - input buffer.
**      2. inBufferLen - length of the input buffer
**      3. key - iternal representation of RSA private key buffer
**      4. outSignature - signature buffer, used as an output parameter
**      5. outSignatureLen - length of the signature buffer, an output paramete
**
**  RETURNS:
**     succes or error code - unsigned long
**
**  NOTES:
**
******************************************************************************
*/
static unsigned long secCryptoSign( 
      unsigned char            *inBuffer,
      unsigned long            inBufferLen,
      RSA_key                  *key,
      unsigned char            **outSignature,
      unsigned int             *outSignatureLen )
{
   RSA *keyLocal;           /* OpenSSL RSA key structure */
   int cypherLen;           /* RSA key length */
   int status;
   unsigned char SHA1digest[SHA_DIGEST_LENGTH]; /* holds digests */


   /* allocate RSA key */
   keyLocal = RSA_new();

   /* put key modulos and private exponent in big number repr in newly allocated RSA key */
   keyLocal->n = BN_bin2bn( key->mod, key->modBufLen, NULL );
   keyLocal->d = BN_bin2bn( key->exp, key->expBufLen, NULL );

   SHA1( inBuffer, inBufferLen, SHA1digest );

   /* key length */
   cypherLen = RSA_size( keyLocal );

   *outSignature = (unsigned char*)malloc( cypherLen );

   /* sign the digest and check for success */
   status = RSA_sign( NID_sha1, SHA1digest, SHA_DIGEST_LENGTH, *outSignature, outSignatureLen, keyLocal);

   /* check for success */
   if ( status )
   {
      RSA_free( keyLocal );
      return OK;
   }
   else
   {
      RSA_free( keyLocal );
      return SIGN_ERROR;
   }
}

/*
******************************************************************************
**  Function Name: secCryptoSignVerify
**
**  PURPOSE:
**     To verify the signed data stored in input buffer using the input public key buffer.
**     The original data is stored in input buffer inData.
**     Key used is RSA key and the digest method is SHA1.
**
**  PARAMETERS:
**      1. inData - input buffer, original data - to verify against
**      2. inDataLen - length of the input buffer
**      3. pSignature - signature buffer to be verified
**      4. sigLen - length of the signature buffer
**      5. key - iternal representation of RSA public key buffer
**
**  RETURNS:
**     succes or error code - unsigned long
**
**  NOTES:
**
******************************************************************************
*/
static unsigned long secCryptoSignVerify( 
      unsigned char               *inData,
      unsigned long               inDataLen,
      unsigned char               *pSignature,
      unsigned long               pSigLen, 
      RSA_key                     *key )
{

   RSA *keyLocal;                               /* OpenSSL RSA key structure */
   int cypherLen;                               /* RSA key length */
   unsigned char SHA1digest[SHA_DIGEST_LENGTH]; /* holds digests */
   int status;                                  /* status code for RSA_sign function, returned value */

   /* allocate the RSA key */
   keyLocal = RSA_new();

   /* put the modulos and the public exponent in the key */
   keyLocal->n = BN_bin2bn( key->mod, key->modBufLen, NULL );
   keyLocal->e = BN_bin2bn( key->exp, key->expBufLen, NULL );


   /* get the key length */
   cypherLen = RSA_size( keyLocal );

   /* get the digest of message */
   SHA1( inData, inDataLen, SHA1digest );

   /* verify the digest against the signature */
   status = RSA_verify( NID_sha1, SHA1digest, SHA_DIGEST_LENGTH, pSignature, cypherLen, keyLocal );

   /* check for errror and return */
   if ( status )
   {
      RSA_free( keyLocal );
      return OK;
   }
   else
   {
      RSA_free( keyLocal );
      return VERIFY_ERROR;
   }
}


/*
******************************************************************************
** FUNCTION:   secCryptoPHash
**
** PURPOSE:    run F(secret, seed) according to sec spec section 9.7
**
** PARAMETERS:
**    1. md - in - hash (md5 or sha1)
**    1. sec - in - pointer to secret buffer
**    2. sec_len  - in - length of secret buffer in bytes
**    3. seed - in - seed buffer
**    4. seed_len - in - length of seed buffer
**    5. out -  out - pointer to key buffer
**    6. olen - in - output key buffer length
**
**
** RETURNS:   void
**
** NOTE:
**
******************************************************************************
*/
static void secCryptoPHash(
      EVP_MD                   *md,
      unsigned char            *sec,
      unsigned long            sec_len,
      const unsigned char      *seed,
      unsigned long            seed_len,
      unsigned char            *out,
      unsigned long            olen)
{
   unsigned long chunk;
   unsigned int j;
   HMAC_CTX ctx;
   HMAC_CTX ctx_tmp;
   unsigned char A1[HMAC_MAX_MD_CBLOCK];
   unsigned int A1_len;

   /* Note: this code is cut and paste from openssl t1_enc.c:tls1_P_hash
    * it was not worth pulling in the whole ssl because of this one fcn
    */
   chunk=EVP_MD_size(md);

   HMAC_Init(&ctx,sec,sec_len,md);
   HMAC_Update(&ctx,seed,seed_len);
   HMAC_Final(&ctx,A1,&A1_len);

   for (;;)
   {
      HMAC_Init(&ctx,NULL,0,NULL); /* re-init */
      HMAC_Update(&ctx,A1,A1_len);
      memcpy(&ctx_tmp,&ctx,sizeof(ctx)); /* Copy for A2 */ /* not needed for last one */
      HMAC_Update(&ctx,seed,seed_len);

      if (olen > chunk)
      {
         HMAC_Final(&ctx,out,&j);
         out+=j;
         olen-=j;
         HMAC_Final(&ctx_tmp,A1,&A1_len); /* calc the next A1 value */
      }
      else	/* last one */
      {
         HMAC_Final(&ctx,A1,&A1_len);
         memcpy(out,A1,olen);
         break;
      }
   }
   HMAC_cleanup(&ctx);
   HMAC_cleanup(&ctx_tmp);
   memset(A1,0,sizeof(A1));
}


/*
******************************************************************************
** FUNCTION:  secCryptoDeriveSNMPv3Key
**
** PURPOSE:   Derive keys from the kerberos sub-key (calls p_hash)
**
** PARAMETERS:
**    1. secret - in - pointer to secret buffer
**    2. secret_len  - in - length of secret buffer in bytes
**    3. seed - in - seed buffer
**    4. seed_len - in - length of seed buffer
**    5. auth_key -  out - pointer to key buffer
**    6. auth_tid - in - authentication transform ID
**    7. enc_key - out - encryption key
**    8. enc_tid - in - encryption trnasform ID
**
**
** RETURNS:   1 if invalid key length
**
** NOTE:
**
******************************************************************************
*/
static unsigned long secCryptoDeriveSNMPv3Key(
      unsigned char         *secret,
      unsigned long         secret_len,
      const unsigned char   *seed,
      unsigned long         seed_len,
      unsigned char         *auth_key,
      int                   auth_tid,
      unsigned char         *enc_key,
      int                   enc_tid)
{
   unsigned char *full_key;
   unsigned long full_key_len;
   int auth_key_len;
   int enc_key_len;
   EVP_MD *md;

   /* algorithm into p-hash is always SHA1 */
   md = EVP_sha1();

   /* determine algorithms based on transform IDs */
   if ( auth_tid == SNMPV3_HMAC_SHA1_TID )
   {
      auth_key_len = SNMPV3_HMAC_SHA1_SIZE;
   }
   else if ( auth_tid == SNMPV3_HMAC_MD5_TID )
   {
      auth_key_len = SNMPV3_HMAC_MD5_SIZE;
   }
   else if ( auth_tid == SNMPV3_NULL_TID )
   {
      auth_key_len = SNMPV3_NULL_KEY_SIZE;
   }
   else
   {
      printf("secCryptoDeriveSNMPv3Key Error: wrong AP authentication tranform ID %d.\n", auth_tid);
      return 1;
   }

   if ( enc_tid == SNMPV3_DES_TID )
   {
      enc_key_len = SNMPV3_DES_KEY_SIZE;
   }
   else if ( enc_tid == SNMPV3_NULL_TID )
   {
      enc_key_len = SNMPV3_NULL_KEY_SIZE;
   }
   else
   {
      printf("secCryptoDeriveSNMPv3Key Error: wrong AP encryption tranform ID %d.\n", enc_tid);
      return 1;
   }

   full_key_len = auth_key_len + enc_key_len;
   full_key = (unsigned char *)malloc(full_key_len);

   /* Call p_hash to get the bytes necessary */
   secCryptoPHash(md, secret, secret_len, seed, seed_len,
         full_key, full_key_len);

   /* copy the bytes we need into the keys */
   if (auth_key_len) 
      memcpy(auth_key, full_key, auth_key_len);
   if ( enc_key_len )
      memcpy(enc_key, full_key+auth_key_len, enc_key_len);

   free(full_key);
   return 0;

}


/*
******************************************************************************
** FUNCTION:  secCryptoDeriveIPsecKey
**
** PURPOSE:   Derive keys from the kerberos sub-key (calls p_hash)
**
** PARAMETERS:
**    1. secret - in - pointer to secret buffer
**    2. secret_len  - in - length of secret buffer in bytes
**    3. seed - in - seed buffer
**    4. seed_len - in - length of seed buffer
**    5. outbound_auth_key -  out - pointer to key buffer
**    6. outbound_enc_key - out - encryption key
**    7. inbound_auth_key -  out - pointer to key buffer
**    8. inbound_enc_key - out - encryption key
**    9. auth_tid - in - authentication transform ID
**    10. enc_tid - in - encryption trnasform ID
**
**
** RETURNS:   1 if invalid key length
**
** NOTE:
**
******************************************************************************
*/
static unsigned long secCryptoDeriveIPSecKey(
      unsigned char        *secret,
      unsigned long        secret_len,
      const unsigned char  *seed,
      unsigned long        seed_len,
      unsigned char        *outbound_auth_key,
      unsigned char        *outbound_enc_key,
      unsigned char        *inbound_auth_key,
      unsigned char        *inbound_enc_key,
      int                  auth_tid,
      int                  enc_tid)
{
   unsigned char *full_key;
   unsigned long full_key_len;
   int auth_key_len;
   int enc_key_len;
   EVP_MD *md;

   /* algorithm into p-hash is always SHA1 */
   md = EVP_sha1();

   /* determine algorithms based on transform IDs */
   if ( auth_tid == IPSec_ESP_HMAC_SHA1_TID )
   {
      auth_key_len = IPSec_ESP_HMAC_SHA1_SIZE;
   }
   else if ( auth_tid == IPSec_ESP_HMAC_MD5_TID )
   {
      auth_key_len = IPSec_ESP_HMAC_MD5_SIZE;
   }
   else
   {
      printf("secCryptoDeriveIPSecKey Error: wrong AP authentication tranform ID %d.\n", auth_tid);
      return 1;
   }

   if ( enc_tid == IPSec_ESP_3DES_TID )
   {
      enc_key_len = IPSec_ESP_3DES_KEY_SIZE;
   }
   else if ( enc_tid == IPSec_ESP_NULL_TID )
   {
      enc_key_len = IPSec_ESP_NULL_KEY_SIZE;
   }
   else
   {
      printf("secCryptoDeriveIPSecKey Error: wrong AP encryption tranform ID %d.\n", enc_tid);
      return 1;
   }

   full_key_len = auth_key_len + enc_key_len + auth_key_len +
      enc_key_len;
   full_key = (unsigned char *)malloc(full_key_len);

   /* Call p_hash to get the bytes necessary */
   secCryptoPHash(md, secret, secret_len, seed, seed_len,
         full_key, full_key_len);

   /* copy the bytes we need into the keys */
   memcpy(outbound_auth_key, full_key, auth_key_len);
   if ( enc_key_len )
   {
      memcpy(outbound_enc_key, full_key+auth_key_len,
            enc_key_len);
   }
   memcpy(inbound_auth_key, full_key+auth_key_len+enc_key_len,
         auth_key_len);
   if ( enc_key_len )
   {
      memcpy(inbound_enc_key,
      full_key+auth_key_len+enc_key_len+auth_key_len,
            enc_key_len);
   }

   free(full_key);
   return 0;
}

/*
******************************************************************************
** FUNCTION:  calcSimpleHash
**
** PURPOSE:   calculate  a non-keyed hash (sha-1 or md5)
**
** PARAMETERS:
**    1. hashType - in -
**    2. inBuf  - in -
**    3. inBufLen - in -
**    4. outHash - out -
**    5. outHasLen - out -
**
**
** RETURNS:   success or error code
**
** NOTE:
**
******************************************************************************
*/
static unsigned long calcSimpleHash(
      int             hashType,
      unsigned char   *inBuf,
      unsigned long   inBufLen,
      unsigned char   *outHash,
      unsigned long   *outHashKeyLen)
{
   /* Determine hash based on key length */
   if ( hashType == 2 )
   {
      SHA1( inBuf, inBufLen, outHash );
      *outHashKeyLen = SHA1_SIZE;
   }
   else if ( hashType == 1 )
   {
      MD5( inBuf, inBufLen, outHash );
      *outHashKeyLen = MD5_SIZE;
   }
   else
   {
      printf("secLibCalcHash Error: Bad hash type.\n");
      return 1;
   }
   return 0;
}

/*
******************************************************************************
** FUNCTION:  calcKeyedHash
**
** PURPOSE:   calculate  a non-keyed hash (hmac sha-1 or md5)
**
** PARAMETERS:
**    1. hashType - in -
**    2. inBuf  - in -
**    3. inBufLen - in -
**    4. hashKey - in -
**    5. hashKeyLen - in -
**    6. outHash - out -
**    7. outHasLen - out -
**
**
** RETURNS:   success or error code
**
** NOTE:
**    Caller responsible for allocating and freeing outHash
**
******************************************************************************
*/
static unsigned long calcKeyedHash(
      int             hashType,
      unsigned char   *inBuf,
      unsigned long   inBufLen,
      unsigned char   *hashKey,
      unsigned long   hashKeyLen,
      unsigned char   *outHash,
      unsigned long   *outHashLen)
{
   if ( hashType == 2 )
   {
      HMAC(EVP_sha1(), hashKey, hashKeyLen,inBuf, inBufLen,
             outHash, (unsigned int *)outHashLen);
   }
   else if ( hashType == 1 )
   {
      HMAC(EVP_md5(), hashKey, hashKeyLen, inBuf, inBufLen,
            outHash, (unsigned int *)outHashLen);
   }
   else
   {
      printf("secLibCalcHash Error: Bad hash type.\n");
      return 1;
   }
   return 0;
}
/*
******************************************************************************
** FUNCTION:  secLibCalcHash
**
** PURPOSE:   calculate hash (sha-1 or md5)
**
** PARAMETERS:
**    1. hashType - in -
**    2. inBuf  - in -
**    3. inBufLen - in -
**    4. hashKey - in -
**    5. hashKeyLen - in -
**    6. outHash - out -
**    7. outHasLen - out -
**
**
** RETURNS:   success or error code
**
** NOTE:
**     hashType - 1 (MD5) or 2 (SHA-1)
**     if hashKey== NULL or hashKeyLen == 0, then simple hash is calculated,
**     othewise - keyed HMAC hash
**     Caller responsible for allocating and freeing outHash parameter
**
******************************************************************************
*/
static unsigned long secCryptoCalcHash(
      int               hashType,
      unsigned char     *inBuf,
      unsigned long     inBufLen,
      unsigned char     *hashKey,
      unsigned long     hashKeyLen,
      unsigned char     *outHash,
      unsigned long     *outHashLen)
{
   if ( (hashKey == NULL) || (hashKeyLen == 0) )
   {
      return (calcSimpleHash(hashType,inBuf, inBufLen,
               outHash, outHashLen));
   }
   else
   {
      return (calcKeyedHash(hashType, inBuf, inBufLen, hashKey, hashKeyLen,
               outHash, outHashLen));
   }
}


/*
******************************************************************************
** FUNCTION:   secCryptoDes
**
** PURPOSE:    To encrypt or decrypt a file with a key
**
** PARAMETERS:
**    1. inBuf - in: input buffer to be encrypted/decrypted
**    2. outBuf - out: output buffer after encryption/decryption
**    3. bufLen - in/out: input buffer length and output bytes processed
**    4. key - in: 8 byte key
**    5. enc - in: 1-encrypt, 0-decrypt
**
** RETURNS: Success - 1
**          Failure - 0
**
** NOTE:    inFile MUST allocate memory of (size of input file + 16) bytes.
**          (8 bytes for prepending, 8 bytes for appending)
**          outFile MUST allocate memory of multiple of 8 byte.
**
*****************************************************************************
*/

static int secCryptoDes(
      char        *inBuf, 
      char        *outBuf, 
      int         *bufLen, 
      char*       key, 
      int         enc)
{
	register int i;
	des_key_schedule ks;
	des_cblock iv;
	int rem;
	des_cblock kk;
	int Exit=0;
   int foundEOF=0;

   if ((inBuf == NULL) || (outBuf == NULL) || (*bufLen <= 0) || (key == NULL))
      return 0;

	memcpy(&kk, key, 8);
   des_set_key_unchecked(&kk,ks);
	memset(key,0,sizeof(key));
	memset(kk,0,sizeof(kk));
	memset(iv,0,sizeof(iv));

	/* first read */
	if (enc)
	{
      /* prepend 8 random bytes */
      memmove((char *)(inBuf+8), inBuf, *bufLen);
      (*bufLen)+=8;
      for (i=0; i<8; i++)
      {
         RAND_pseudo_bytes((unsigned char *)inBuf+i, 1);
      }

      rem=*bufLen%8;
      if (rem>0)
      {
         for (i=0; i<8-rem; i++)
         {
            RAND_pseudo_bytes((unsigned char *)(inBuf + ((*bufLen)++)), 1);
         }
      }

#if 0 /*  Does not comply with Cisco's provisioning server */
		for (i=7-rem; i>0; i--)
      {
         RAND_pseudo_bytes(inBuf + ((*bufLen)++), 1);
      }
      inBuf[(*bufLen)++]=rem;
#endif

		des_cbc_encrypt(inBuf,outBuf,(long)*bufLen,ks,&iv,DES_ENCRYPT);

      /* Don't understand, might be for security reason */
		if (rem)
         memcpy(inBuf,&(inBuf[*bufLen]),(unsigned int)rem);
	}
   /* decrypt */
	else
	{
      if (*bufLen%8)
      {
         Exit = 1;
         goto problem;
      }

		des_cbc_encrypt(inBuf,outBuf,(long)*bufLen,ks,&iv,DES_DECRYPT);

#if 0 /* does not comply with Cisco's provisioning server */
		last=outBuf[*bufLen-1];
		if ((last > 7) || (last < 0))
		{
         /* The file was not decrypted correctly. */
			Exit=2;
			last=0;
		}
      *bufLen=*bufLen-8+last;
#endif

      for (i=(*bufLen-1); i>(*bufLen-9); i--)
      {
         if (((unsigned char)outBuf[i]==255) && ((unsigned char)outBuf[i-1]==1) && ((unsigned char)outBuf[i-2]==254))
         {
            foundEOF = i;
         }
      }
      if (!foundEOF)
      {
         Exit = 3;
         goto problem;
      }
      *bufLen=foundEOF+1;

      /* Remove first 8 random bytes prepended to the original file */
      memmove(outBuf, (char *)(outBuf+8), (*bufLen)-8);
      *bufLen=*bufLen-8;

	}

problem:
	memset(ks,0,sizeof(ks));
	memset(iv,0,sizeof(iv));
	memset(kk,0,sizeof(kk));
	if (Exit)
      return 0;
   return 1;
}


/*
******************************************************************************
** FUNCTION:   secCryptoDeriveRTPKeys
**
** PURPOSE:    Derive all the rtp keys from secret and pad
**
** PARAMETERS:
**    1. seed - in - constant string
**    2. seedLen - in - length of seed
**    3. secret - in - (46 bytes) rtp secret
**    4. secretLen - in -
**    5. pad - in - (46 bytes) pad
**    6. padLen - in -
**    7. rtpPrivKeyLen - in -
**    8. rtpIntitKeyLen - in -
**    9. rtpMacKeyLen - in -
**    10. rtpPrivKey - out - rtp privacy key
**    11. initTS - out - initial time stamp, always 4 bytes
**    12. rtpInitKey - out - rtp initial key
**    13. rtpMacKey - out - MAC key, 0,2 or 4 bytes
**
** RETURNS: Success - 1
**          Failure - 0
**
** NOTE:
**    1. all memory must be allocated before calling this function, freed after
**    2. pad is optional, if padLen is 0, pad will not be part of derivation
**    3. pad and secret must be 46 bytes (not checked by this function)
**
*****************************************************************************
*/
static int secCryptoDeriveRTPKeys(
      unsigned char      *seed,
      unsigned long      seedLen,
      unsigned char      *secret,
      unsigned long      secretLen,
      unsigned char      *pad,
      unsigned long      padLen,
      unsigned long      rtpPrivKeyLen,
      unsigned long      rtpInitKeyLen,
      unsigned long      rtpMacKeyLen,
      unsigned char      *rtpPrivKey,
      unsigned long      *rtpInitTS,
      unsigned char      *rtpInitKey,
      unsigned char      *rtpMacKey)
{
   EVP_MD        *md;
   unsigned char *full_key;
   int           full_key_len;
   unsigned char *full_secret;
   int           full_secret_len;

   md = EVP_sha1();

   /* calculate the total number of bytes we need,
    * initTS is always 4 bytes
    */
   full_key_len = 4 + rtpPrivKeyLen + rtpInitKeyLen + rtpMacKeyLen;
   full_key = (unsigned char *)malloc(full_key_len);

   /* concatenate stuff for full_secret
    * ssrc is 4 bytes, nRekey is 2 bytes
    * */
   full_secret_len = secretLen + padLen;
   full_secret = malloc(full_secret_len);

   memset(full_secret, 0, full_secret_len);

   memcpy(full_secret, secret, secretLen);

   if ( padLen != 0 )
   {
      memcpy(full_secret+secretLen, pad, padLen);
   }

   /* Call p_hash to get the bytes necessary */
   secCryptoPHash(md, full_secret, full_secret_len, seed, seedLen,
         full_key, full_key_len);

   /* copy the bytes we need into the keys */
   memcpy(rtpPrivKey, full_key, rtpPrivKeyLen);
   memcpy(rtpInitTS, full_key+rtpPrivKeyLen, 4);
   memcpy(rtpInitKey, full_key+rtpPrivKeyLen+4, rtpInitKeyLen);
   if ( rtpMacKeyLen != 0 )
   {
      memcpy(rtpMacKey, full_key+rtpPrivKeyLen+4+rtpInitKeyLen, rtpMacKeyLen);
   }

   free(full_key);
   free(full_secret);

   return 1;
}

/*
******************************************************************************
** FUNCTION:   secCryptoDeriveRTCPKeys
**
** PURPOSE:    Derive all the rtCp keys from secret and pad
**
** PARAMETERS:
**    1. seed - in - constant string
**    2. seedLen - in - length of seed
**    3. secret - in - (46 bytes) rtp secret
**    4. secretLen - in -
**    5. pad - in - (46 bytes) pad
**    6. padLen - in -
**    7. rtcpEncKeyLen - in -
**    8. rtcpAuthKeyLen - in -
**    9. rtcpEncKey - out - rtcp encryption key
**    10. rtcpAuthKey - out - rtcp authentication key
**
** RETURNS: Success - 1
**          Failure - 0
**
** NOTE:
**    1. all memory must be allocated before calling this function, freed after
**    2. pad is optional, if padLen is 0, pad will not be part of derivation
**    3. pad and secret must be 46 bytes (not checked by this function)
**
*****************************************************************************
*/
static int secCryptoDeriveRTCPKeys(
      unsigned char      *seed,
      unsigned long      seedLen,
      unsigned char      *secret,
      unsigned long      secretLen,
      unsigned char      *pad,
      unsigned long      padLen,
      unsigned long      rtcpEncKeyLen,
      unsigned long      rtcpAuthKeyLen,
      unsigned char      *rtcpEncKey,
      unsigned char      *rtcpAuthKey)
{
   EVP_MD           *md;
   unsigned char    *full_key;
   int              full_key_len;
   unsigned char    *full_secret;
   int              full_secret_len;

   md = EVP_sha1();

   /* calculate the total number of bytes we need, */
   full_key_len = rtcpEncKeyLen + rtcpAuthKeyLen;
   full_key = (unsigned char *)malloc(full_key_len);

   /* concatenate stuff for full_secret
    * ssrc is 4 bytes, nRekey is 2 bytes
    * */
   full_secret_len = secretLen + padLen;
   full_secret = malloc(full_secret_len);


   memcpy(full_secret, secret, secretLen);

   if ( padLen != 0 )
   {
      memcpy(full_secret+secretLen, pad, padLen);
   }

   /* Call p_hash to get the bytes necessary */
   secCryptoPHash(md, full_secret, full_secret_len, seed, seedLen,
         full_key, full_key_len);

   /* copy the bytes we need into the keys.  Generate the Auth key first, then the Privacy Key*/
   memcpy(rtcpAuthKey, full_key, rtcpAuthKeyLen);
   memcpy(rtcpEncKey, full_key+rtcpAuthKeyLen, rtcpEncKeyLen);
   
   free(full_key);
   free(full_secret);

   return 1;
}


/*
******************************************************************************
** FUNCTION:   secCryptoBase64
**
** PURPOSE:    To encode/decode using Base64
**
** PARAMETERS:
**    1. inBuf - IN: pointer to buffer being encoded/decoded
**    2. inLen - IN: length of input buffer
**    3. outBuf - OUT: pointer to buffer after encoding/decoding
**    4. outLen - OUT: length of buffer being output
**    5. encode - IN: 1->encode, 0->decode
**    6. maxLength - IN: maximum length of encoded output before carriage
**                       return is inserted
**
** RETURNS:   none
**
** NOTE: During decoding caller must allocate 1 extra byte to inBuf because
**       this function need to attach '\0' at the end of the buffer.
**
*****************************************************************************
*/
static void secCryptoBase64(
      unsigned char   *inBuf, 
      int             inLen, 
      unsigned char   *outBuf, 
      int             *outLen, 
      int             encode, 
      int             maxLength)
{
   if (encode)
   {
      enc64(outBuf, (2*inLen), outLen, inBuf, inLen, maxLength, 0);
   }
   else
   {
      inBuf[inLen] = '\0';
      dec64(outBuf, inLen, outLen, inBuf);
   }
   return;
}

/*
******************************************************************************
** FUNCTION:   enc64
**
** PURPOSE:    encode inbuff to outbuff, forcing \n every line_lth chars and
**             indenting continuation lines.
**
** PARAMETERS:
**    1. outbuff - output buffer
**    2. out_lth - allocated length of the output buffer
**    3. polth - actual length of output
**    4. inbuff - input (binary) buffer
**    5. inb_lth - length of inbuff
**    6. line_lth - maximum line lth
**    7. n_space - # spaces at start of each text line return is inserted
**
** RETURNS:   none
**
** NOTE: Caller needs to allocate more outbuff than needed by 4 or 5 bytes.
**
*****************************************************************************
*/
static void enc64(
      unsigned char            *outbuff, 
      int                      out_lth, 
      int                      *polth, 
      unsigned char            *inbuff, 
      int                      inb_lth, 
      int                      line_lth, 
      int                      n_space)
{
   long nl;			   /* # chars left in this line */
   unsigned char *b, *c;		/* walking pointers */

   nl = line_lth;
   b = inbuff;
   c = outbuff;

   while ((inb_lth>0) && (out_lth>5))
   {
    /* encoding - this may read past the end of the inbuf by 1 or
       2 chars, but the results from this  will be overiden by = chars
    */
      c[0]=enctab[(b[0]>>2)&0x3f] ;
      c[1]=enctab[((b[0]&0x3)<<4)|((b[1]>>4)&0xf)] ;
      c[2]=enctab[((b[1]&0xf)<<2)|((b[2]>>6)&0x3)] ;
      c[3]=enctab[b[2]&0x3f] ;
      out_lth -= 4 ;		   /* count the code bytes */

      switch (inb_lth)
      {		/* take care of the final bytes */
      case 1:
         c[2]='=' ;		   /* only 1, so == */
      case 2:
         c[3]='=' ;		   /* 2, so = */
         inb_lth = 0 ;		/* either way, we're done */
         c += 4 ;			   /* but no spaces */
         *(c++) = '\n' ;	/* and there's an end of line */
         break ;

      default:
         inb_lth -= 3;
         b += 3 ;
         c += 4 ;
         nl -= 4 ;
         if (nl <= 0)
         {
	         long i ;
	         *(c++) = '\n' ;
	         nl = line_lth ;
	         for (i=0;i<n_space;i++)
            {
	            *(c++) = ' ';
            }
	         out_lth -= 1 + n_space ;
         }
         break ;
      }
   }
   *polth = c - outbuff ;
}

/*
******************************************************************************
** FUNCTION:   dec64
**
** PURPOSE:    decode radix64 from 0-terminated inbuff to outbuff
**
** PARAMETERS:
**    1. outbuff - output binary buffer
**    2. out_lth - allocated lth of outbuff
**    3. polth - actual length of output
**    4. inbuff - input text buffer
**
** RETURNS:   none
**
** NOTE: Caller needs to allocate 2 or 3 bytes more than needed in outbuff.
**
*****************************************************************************
*/
static void dec64(
      unsigned char    *outbuff, 
      int              out_lth, 
      int              *polth,
      unsigned char    *inbuff)
{
   unsigned char d[4] ;			   /* a buffer of 4 coded characters */
   unsigned char *cp = inbuff ;	/* walk inbuff */
   unsigned char *b = outbuff ;	/* walk outbuff */
   long di = 0 ;			/* index into d[] */
   unsigned char c ;			      /* temp char */
   unsigned char *e = outbuff+out_lth-3 ; /* end pointer */

   while (  (b < e)		/* room to do output? */
	   &&((c = *(cp++))!= 0) )
   { /* loop until end of string */
      if ((c = dectab[c]) >= 0)
      {	/* valid code character? */
         d[di++] = c ;		/* yes, valid char */
         if (di == 4)
         {		/* have a full set? */
            b[0] = (unsigned char)( ( d[0] << 2 ) | ( ( d[1] >> 4 ) & 0x3 ) );
            b[1] = (unsigned char)( ( (d[1] & 0xf) << 4 ) | ( ( d[2] >> 2 ) & 0xf ) );
            b[2] = (unsigned char)( ( (d[2] & 0x3) << 6 ) | d[3] );
            b += 3 ;		/* have 3 new bytes */
            if (d[3]==64)
            {
               b-- ;	/* if final '=', one less byte */
            }
	         if (d[2]==64)
            {
               b-- ;	/* if 2nd '=', another less */
            }
         	di = 0 ;		/* start over */
         } /* done with the full set */

      } /* if dectab[c] >= 0 */

   } /* while not end of string */

   *polth = b - outbuff ;
}

/*
******************************************************************************
** FUNCTION:   secCryptoGetCurrUTCTime
**
** PURPOSE:    get the current system time in UTC format
**
** PARAMETERS:
**    1.  ptime - time in seconds
**    2.  utc_time_str - UTC time returned
**    3.  utc_str_len - length of the string
**
** RETURNS:   success or failure
**
** NOTE:
**
*****************************************************************************
*/
static int secCryptoGetCurrUTCTime(
      int                  ptime, 
      unsigned char        *utc_time_str,
      unsigned long        *utc_str_len)
{
   time_t       curtime;
   struct tm    *tmptm;

   curtime = ptime;
   tmptm = gmtime(&curtime);

   if ( utc_time_str == NULL )
      return 1;

	sprintf(utc_time_str,"%02d%02d%02d%02d%02d%02dZ",tmptm->tm_year%100,
		tmptm->tm_mon+1,tmptm->tm_mday,tmptm->tm_hour,tmptm->tm_min,tmptm->tm_sec);

   *utc_str_len = 13;

   return 0;
}


/*
******************************************************************************
** FUNCTION:   secCryptoConvertUTCToTime
**
** PURPOSE:    convert given UTC time to time_t
**
** PARAMETERS:
**    1.  utc_time_str - UTC time
**    2.  utc_str_len - length of the string
**    3.  ptime - time in seconds
**
** RETURNS:   Success or failure
**
** NOTE:
**
*****************************************************************************
*/
static int secCryptoConvertUTCToTime(
      unsigned char        *utc_time_str,
      unsigned long        utc_str_len,
      unsigned long        *ptime)
{
   struct tm tm;
   int offset;
   unsigned char *s=utc_time_str;

   memset(&tm,'\0',sizeof tm);

#define g2(p) (((p)[0]-'0')*10+(p)[1]-'0')
   tm.tm_year=g2(s);
   if(tm.tm_year < 50)
      tm.tm_year+=100;
   tm.tm_mon=g2(s+2)-1;
   tm.tm_mday=g2(s+4);
   tm.tm_hour=g2(s+6);
   tm.tm_min=g2(s+8);
   tm.tm_sec=g2(s+10);
   if(s[12] == 'Z')
      offset=0;
   else
   {
      offset=g2(s+13)*60+g2(s+15);
      if(s[12] == '-')
         offset= -offset;
   }
   tm.tm_isdst = 0;
#undef g2

   *ptime = mktime(&tm)-offset*60;

   return 0;
}

