/***************************************************************************
*    Copyright 2000  Broadcom Corporation
*    All Rights Reserved
*    No portions of this material may be reproduced in any form without the
*    written permission of:
*             Broadcom Corporation
*             16251 Laguna Canyon Road
*             Irvine, California  92618
*    All information contained in this document is Broadcom Corporation
*    company private, proprietary, and trade secret.
*
****************************************************************************
*
*    Filename: provlib.h
*    Creation Date: 24 Aug 1999 (v0.00)
*    VSS Info:
*        $Revision: 1.1 $
*        $Date: 2002/08/21 17:54:31 $
*
****************************************************************************
*    Description:
*
*      PROVLIB Module: MTA Provisioining Library Module .H File.
*
****************************************************************************/

#ifndef PROVLIB_H
#define PROVLIB_H

#ifdef __cplusplus
extern "C"
{
#endif

#include <hausenv.h>          /* HH Debug definitions */
//#include <MibAccessDef.h>
#include <bcmos.h>

#ifndef  MAX_FQDN_SIZE
#define  MAX_FQDN_SIZE     256
#endif


#define  TFTP_TASK_NAME          "HTFT"   /* TFTP Task Name */
#define  TFTP_TASK_STACK_SIZE    (1024)   /* TFTP Stack Size */
#define  TFTP_TASK_TIME_OUT      100      /* TFTP task time out - 2 sec */

#define  PROV_TASK_NAME          "HPRV"   /* Main Provisioning Task Name */
#define  PROV_TASK_STACK_SIZE    (4 * 1024)
#define  PROV_TASK_SEMA_NAME     "HSPR"   /* Provis Start sema name */
#define  PROV_TASK_SEMA_GEN      "HSEM"   /* Provis general sema name */
#define  PROV_TASK_ERROR_BASE    0x10000  /* Prov task error code base */
#define  PROV_TASK_DHCP_TO       10       /* secs: DHCP task time out */
#define  PROV_DHCP_FAIL_CNTR_MX  3        /* MAX Failure Counter for DHCP Failure */
#define  PROV_WHOLE_FAIL_CNTR_MX 3        /* MAX Failure Counter for the rest of the provisioning */

#define  PROV_PKTC_USM_VIEW      "pktcManagerView"

#define  PROV_TASK_ERR(err)      (PROV_TASK_ERROR_BASE + err)

#define  CLEANUP_TASK \
   { \
      fclose(0); \
      close_f(0); \
      close(0); \
      free((void*)-1); \
   }



enum  /* Enum types for MTA Events */
{
   MTA_EVT_NONE,        /* Event "No Events" */
   MTA_EVT_DHCP_DONE    /* Event DHCP binding is done */
};

enum  /* Enum types for three kind of servers in the provisioning */
{
   SERVR_TEL_PROV,      /* Telephony Provisioning Server */
   SERVR_CALL_MGMT,     /* Call Management Server */
   SERVR_SECUR_AUTH     /* Security and Authentication Server */
};

typedef struct tagPRZSERVERDSC    /* SERVER Descriptor: Data from DHCP option 177 */
{
   UINT32    prxSrvrIpAddr;        /* Server IP Address */
   UINT32    prxSrvrPort;          /* Server Port */
   char     prxSrvrIpAddrStr[MAX_FQDN_SIZE];/* IP Address as a string (might be FQDN). Max size is stated by the Prov spec */
} PRZSERVERDSC;

typedef enum  /* Enum for MIB events that can be notified */
{
   MIBEVENT_MTARESETNOW,

   /* Insert more events here */
   MIBEVENT_MAX

} MIBEVENT;

/* Function type for MIB EVENT notify callbacks */
typedef void (*MIBEVT_FUNC)(MIBEVENT eventId);

/****************************************************************
  The following is defining all neccessary data defs for
  MIB Access Functionality in Provisioining
****************************************************************/
/* trc-
BEGIN_MIBITEM_LIST(PRV,0)
MIBITEM(PRV, ProvState),
MIBITEM(PRV, ProvisioningState),
MIBITEM(PRV, ConfigFile),
MIBITEM(PRV, ConfigHash),
MIBITEM(PRV, ConfigKey),
MIBITEM(PRV, HardwareVersion),
MIBITEM(PRV, SwCurrentVers),
MIBITEM(PRV, TypeIdentifier),
MIBITEM(PRV, MacAddress),
MIBITEM(PRV, CorrelationId),
MIBITEM(PRV, FQDN),
MIBITEM(PRV, DNS1),
MIBITEM(PRV, DNS2),
MIBITEM(PRV, Enabled),
MIBITEM(PRV, DHCP1),
MIBITEM(PRV, DHCP2),
MIBITEM(PRV, TimeServer),
MIBITEM(PRV, SNMP),
MIBITEM(PRV, CODEC_TAB),
MIBITEM(PRV, ProvisioningTimer),
MIBITEM(PRV, ProvisioningCounter),
MIBITEM(PRV, KerbRealmName),
MIBITEM(PRV, SysLog)
END_MIBITEM_LIST(PRV)

DCL_MIBACC_FUNC(PRV)*/ /* Declare the MIB Access Funciton Prototype */

/****************************************************************
  This structure defines all data which configures the
  Provisioning task.
****************************************************************/
#define  TFTP_FILE_LEN  128
#define  MAX_ROUTERS    4        /* Routers */
#define  MAX_TSERVRS    4        /* Time Servers */
#define  MAX_LSERVRS    4        /* SysLog Servers */

typedef struct tagPRZPROVIS
{
   UINT32    prxProvCntr;         /* Provision Counter: number of times provisioning was invoked */
   char      prxMtaClassId[64];   /* MTA Class name (actually size is 9 but make it 16 for alignement) */
   UINT32    prxMtaClassIdLen;    /* MTA Class name length */
   UINT32    prxMtaStandalone;    /* We are working in standalone mode - no CM part */
//   OS_TASK_CLASS  prxprioProvTask;     /* Provisioning Task Priority */
//   OS_TASK_CLASS  prxprioDhcpTask;     /* DHCP Task Priority */
   UINT32    prxInitType;         /* DHCP Init type: cold(0), warm(1) */
   UINT32    prxDhcpEvent;        /* Last DHCP event */
   UINT32    prxErrCode;          /* Error Code to be returned to the Provis task */
   UINT32    prxDhcpCfgSet;       /* Flag is set if DHIP parameter has been set in GWCONFIG Data Structure */
   PRZSERVERDSC   prxMtaDhcp;    /* MTA DHCP Server */
   PRZSERVERDSC   prxMtaDhcp2;   /* MTA Secondary DHCP Server*/
   PRZSERVERDSC   prxProvSnmp;   /* MTA Provis SNMP Entity */
   PRZSERVERDSC   prxMtaDns1;    /* MTA Provis Primary DNS */
   PRZSERVERDSC   prxMtaDns2;    /* MTA Provis Secondary DNS */
   PRZSERVERDSC   prxTlvTftp;    /* TFTP server IP Address for TLV Connig File */
   PRZSERVERDSC   prxKdcServer;   /* Provisioning KDC Server */
   UINT32    prxciAddr;           /* client IP address from client */
   UINT32    prxyiAddr;           /* client IP address from server */
   UINT32    prxsiAddr;           /* IP address of TFTP servr */
   UINT32    prxgiAddr;           /* relay agent IP address */
   unsigned char  prxMacAddrStr[32]; /* client hardware address */
   char     prxtftpFile[TFTP_FILE_LEN];/* File name to be TFTP-ed from the TFTP host */
   char     prxtftpTlvFile[TFTP_FILE_LEN];/* TLV Config File name to be TFTP-ed from the TLV TFTP host */
   UINT32    prxSnetMask;         /* Sub-net mask "Subnet Mask" opt (1) */
   UINT32    prxTimeOffset;       /* "Time Offset" opt(2) */

#if 0
   UINT32    prxRouters[MAX_ROUTERS];/* MAX number of Routers in "Router" Opt(3) */
   UINT32    prxTimServs[MAX_TSERVRS];/* MAX number of Time Servrs in "Time Server" Opt(4) */
   UINT32    prxLogServs[MAX_LSERVRS];/* MAX number of SysLog Servrs in "Time Server" Opt(7) */
#endif

   UINT32      prxDnsActive;        /* DNS Service is active and running */
   UINT32      prxDhcpActive;       /* DHCP Service is active and running */
   UINT32      prxProvKrbDone;      /* Provisioning Kerberos Key Negotiations are done */
   UINT32      prxTgtRequest;       /* TGT Kerberos Request Flag (suboption-7)*/
   UINT32      prxProvTimer;        /* pktcMtaDevProvisioningTimer (suboption-8)*/
   UINT32      prxProvDone;         /* MTA Provisioned */
   char        prxMtaFqdn[MAX_FQDN_SIZE]; /* MTA FQDN set from the DHCP (option-12) */
   unsigned char  prxMtaSysLogServer[MAX_FQDN_SIZE];  /* MTA Syslog server IP address (option-7)*/
   char        usmUserSecName[32];  /* SNMPv3 USM User Security Name: built automatically as: "MTA-Prov-xx:xx:xx:xx:xx:xx" */
   UINT32      prxCorrelationId;
//   MIBACC_FUNC(PRV)  prxMibAccessFunc;    /* MIB Access Function */
   MIBEVT_FUNC prxMibEventNotifyFunc;     /* MIB Event Notification function */
   BOOL        bKrbStarted;
   char        prxSnmpMgrUserName[MAX_FQDN_SIZE];
   int         prxSnmpMgrUserNameLen;
   char        prxSnmpMgrEngineID[64];
   int         prxSnmpMgrEngineIDLen;
   UINT32      prxSnmpMgrEngineBoot;
   UINT32      prxSnmpMgrEngineTime;
} PRZPROVIS;


/*
***************************************************************
  This part of the file defines the constants for
  MIB Data Access. Enumerated values are mostly coming from or
  correspoding to the MIB values generated by the MIB
  compiler and located in "mibleaf.h" file.
***************************************************************
*/
#define mtaProvState_operational                1
#define mtaProvState_disabled                   2
#define mtaProvState_other                      3
#define mtaProvState_unknown                    4
#define mtaProvState_waitingToStart             5
#define mtaProvState_waitingForDhcpOffer        6
#define mtaProvState_waitingForDhcpAckResponse  7
#define mtaProvState_waitingForProvRealmKdcNameResponse  8
#define mtaProvState_waitingForProvRealmKdcAddrResponse  9
#define mtaProvState_waitingForAsReply          10
#define mtaProvState_waitingForTgsReply         11
#define mtaProvState_waitingForApReply          12
#define mtaProvState_waitingForSnmpGetRequest   13
#define mtaProvState_waitingForSnmpSetInfo      14
#define mtaProvState_waitingForTftpAddrResponse 15
#define mtaProvState_waitingForConfigFile       16
#define mtaProvState_waitingForTelRealmKdcNameResponse   17
#define mtaProvState_waitingForTelRealmKdcAddrResponse   18
#define mtaProvState_waitingForPkinitAsReply    19
#define mtaProvState_waitingForCmsKerbTickTgsReply       20
#define mtaProvState_waitingForCmsKerbTickApReply        21

#define mtaProvisioningState_Success         1
#define mtaProvisioningState_inProgress      2
#define mtaProvisioningState_Fail            3

/*
*****************************************************************************
** FUNCTION:   prfProvStart
**
** PURPOSE:    Function initiating provisioning task
**
** PARAMETERS:
**
** RETURNS:
**
** NOTE:
**
*****************************************************************************
*/
UINT32 prfProvStart(PRZPROVIS*);

/*
*****************************************************************************
** FUNCTION:   prfProvShutdown
**
** PURPOSE:    Function to shutdown provisioning module/tasks
**
** PARAMETERS:
**
** RETURNS:
**
** NOTE:
**
*****************************************************************************
*/
void prfProvShutdown(void);

/*
*****************************************************************************
** FUNCTION:   prfVerifyFQDNSyntax
**
** PURPOSE:    This function checks the syntax of an FQDN
**
** PARAMETERS:  instr - FQDN
 *              length - length of FQDN
**
** RETURNS:
 *              1 - if syntax is good
 *              0 - if bad syntax
**
** NOTE:
*****************************************************************************
*/
int prfVerifyFQDNSyntax(char * instr, int length) ;



#ifdef __cplusplus
}
#endif

#endif /* PROVLIB_H macro */



